### IMPORTANT: THIS CODE IS NOT OPEN SOURCE ### - You can look and learn, but do not use or publish any of this work online.  It is all (C) Copyright, 2020, by me - Chris Drake.
                            ###

#Author-Chris Drake
#Description-A comprehensive selection of utilities for creating, managing, and optimising\nairfoil shapes in air, water, or other gasses/fluids.  Contains tools to create\nWing surfaces, Cowling shapes, Propellers and Turbines, and utilities to read\nand write airfoil datafiles, to simplify and optimise file formats, to run\nparticle swarm optimisation on 2D airfoil shapes to maximise performance in your\nchosen situations, comparison solutions, spline sketch insertion, and more.
#Copyright-(C)2020 Chris Drake.  All rights reserved.  Note: This is NOT "Open Source": contact the author if you want to re-use any of this code.

# Work URLs:
# https://forums.autodesk.com/t5/fusion-360-api-and-scripts/airfoiltools-add-in-current-feature-list-reply-to-this-topic-to/td-p/9888073
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-c5193a7a-6c41-4b47-bac3-f3e810ef9146
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-36E1730E-C241-40AE-941A-7C7F7BDDAECD
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-1728AD08-743F-4ABB-AA10-C2B2C53A0757
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-110F4701-53D7-4BC8-B2BE-7A7C18983D09
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-b2e604ff-8734-4f8a-8ed7-fcbde200a46c
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-6E88AA99-65F2-47FA-8059-C79DC27FDD00
# https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-a2ec6e82-20c5-47c7-a1bd-c545f7e76810
# https://forums.autodesk.com/t5/fusion-360-api-and-scripts/api-how-to-loft-between-a-closed-edge-from-a-solid-to-a-sketched/td-p/8139386?us_oa=forums-us&us_si=8efa28eb-0025-4a05-ab54-51d6a63aa8c2&us_st=get%20brepEdge%20from%20patch
# 

VERSION=1.20220220                                                              # Remember to update this regularly!

# The right shape can halve your drag and double your lift - that's 4 times better than guessing.  The right propellers make more thrust with less noise and less unwanted torque effect, the right turbines extract more energy from the wind or water, the right cowlings and struts and masts slice through the air and water more efficiently and quietly, the right sails drive your yacht faster, the right wings glide further for less fuel, the right hydrofoils go faster with less drag, and work with more weight at lower speeds... the benefits from harnessing cutting-edge Computational-Fluid-Dynamic (CFD) optimisations are immense!  All FOR FREE.  Get this now, before the Author changes his mind, or gets bought-out by a commercial operator.

## TODO ##
# De-Select Wrap-Axis on first open (users never want that)
# Fix meaning of temperature
# Use Mach to get best foils
# close UI upon errors, to prevent looping errors.
# code export normalize
# add nose-tail axis line (if none exists?)
# camber line?  export... flip?
#
# turbine: diameter, tip speed, mach, density etc

# vim: set fileencoding=Windows-1252 :

# We build engineering solutions that put CFD-optimised high-performance Airfoils and Hydrofoils at your fingertips, with just a few clicks.  For Free!
# Our Airfoil Tools add-in for AutoDesk Fusion 360 lets you insert optimised Wings, Struts, Propellers, Turbines, Cowlings, and Ducts, and includes utilities to import and export foils, run performance analysis, do flow calculations, tinker with airfoil data files (repair, normalize, scale, rotate, wrap, cap, constrain, etc), and use particle-swarm genetic optimisation to tailor foil shapes to your precise needs.
# You tell us what you want and how it will be used, and our system finds and inserts the best shape that suits your needs!
# This is the most powerful, full-featured Airfoil solution currently available.  It's based on 8 years research and hundreds of thousands of CPU hours of performance pre-computation and Computational-Fluid-Dynamics (CFD) pre-optimisations, and has proven to offer 20% to 200% performance improvements over every known existing shape in all operating conditions.
# Our product is free!  Please use our affiliate link when installing Fusion 360 (required to use our add-in) so that if you later chose to buy a subscription, we will benefit.  Your support helps us make our product better for you!

# VERSION HISTORY
# 1.20200714 - added minthick
# 1.20200801 - added turbine creation strategy
# 1.20201110 - too many to document; includes import and export and lots of new menus and bugfixes
# 1.20210201 - added patch and surface-loft to props and turbines
# 1.20211202 - incorporate Mach #s

# Submission notes: https://damassets.autodesk.net/content/dam/autodesk/www/adn/pdf/desktop-app-submission-process-overview.pdf

# TODO
#   - speed stuff up by wrapping between  sketch.isComputeDeferred = True / sketch.isComputeDeferred = False
#   - investigate adsk.doEvents() to speed stuff up?
#   - auto-change a wing chord-line to a construction line if the user forgot
#   - axis-chord
#   - cool ends for struts as well as foils - better shape for ends (x% struts?)
#   - get rid of hub edge idea - replace with "from tip to hub, stop when wrap >=180 degrees"
#   - preset save/load (next release)
#   - test things do not break when files are missing (settings, update, help)
#   - disc theory with units: https://web.mit.edu/16.unified/www/FALL/thermodynamics/notes/node86.html
#   - metadata display (billboards): https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-ED78D886-0B51-4FCA-98FA-A549EDEF7C04
#   - consider: https://forums.autodesk.com/t5/fusion-360-api-and-scripts/how-would-you-extract-expanded-points-from-a-spline/m-p/9567551#M10440
#   - investigate converting from cubic B-splines to fusion control-point-splines https://www.autodesk.com/products/fusion-360/blog/sketch-control-point-splines-faq/?_ga=2.61932460.1487693074.1589630537-965653787.1589428597
#   - use ConstructionPoint.entityToken Property with Design.findEntityByToken to make everything parametric
#   - store entityTokens and metadata using attributes: http://help.autodesk.com/view/fusion360/ENU/?guid=GUID-84864338-D55F-47DF-95C8-53C5B8813DA7
#   - expose my subs for coders to use themselves
#   - add/view attributes right-click command: https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-entries-to-context-menu-for-dimension/m-p/8270880/highlight/false#M6616
#   - CSV export (right click?) https://forums.autodesk.com/t5/fusion-360-api-and-scripts/export-sketchpoint-coordinates-in-csv-file/td-p/9018519
#   - improve auto-reload after update: https://forums.autodesk.com/t5/fusion-360-api-and-scripts/python-restart-add-in-after-self-updating/m-p/8597138
#   - Add denticles texture: https://royalsocietypublishing.org/doi/full/10.1098/rsif.2017.0828 / https://www.nature.com/articles/s41598-020-71434-8 (SLA 3D testing?)

'''

evaluator = spline.geometry.evaluator
start, end = evaluator.getParameterExtents()
step = (end-start)/100

points = []
param= start
while param < end:
    points.append(evaluator.getPointAtParameter(param))
    param += step

'''

# REFS
#   - How to insert performance data: https://forums.autodesk.com/t5/fusion-360-api-and-scripts/when-will-custom-graphics-support-text/td-p/8271988
#   - Getting perf data into python: https://docs.python.org/2/extending/extending.html
#   - Dude from 2012 ideas about prop design: https://etd.auburn.edu/bitstream/handle/10415/3158/David_Wall_Thesis.pdf?sequence=2
#   - Units in a diag: https://www.grc.nasa.gov/www/k-12/airplane/propth.html and https://www.grc.nasa.gov/www/k-12/airplane/lifteq.html
#   - Nurbs... for control-point spine and J-spline to fit-point translation etc: https://help.autodesk.com/view/fusion360/ENU/?caas=caas/discussion/t5/Fusion-360-API-and-Scripts/NURBS/td-p/5531965.html
#   - nomenclature for propellers: https://ittc.info/media/4034/75-01-02-01.pdf

# PR
#   Any thing or part that is exposed to moving gas or liquid can benefit from a streamlined shape.  In particular, wings, fins, ducts, propellers and turbines can all be vastly
#   optimised by using the optimal correct shape.  "Optimal" is complicated; it depends on size, speed, turbulence, altitude (or depth), the properties of your medium (temperature 
#   and density) and the surface finish you use (rough or smooth).  To further complicated things, many of those things are a range (e.g. speed, temperature, pressure, etc) and 
#   when they are, there's often one point in that range where you need to concentrate on optimising (e.g. cruising speed at cruising altitude).  program_name/Airfoil Tools does this optimisation for you.

# RELEASE
# cd '/Users/cnd/Library/Application Support/Autodesk/Autodesk Fusion 360/API/AddIns'
# dozip; export ZFN=Airfoil_Tools-`now`_rel.zip; mv AirfoilTools/.vscode .;echo -e "Any thing or part that is exposed to moving gas or liquid can benefit from a streamlined shape.  In particular, wings, fins, ducts, propellers and turbines can all be vastly optimised by using the optimal correct shape." | zip -r $ZFN AirfoilTools/ -c -x 'AirfoilTools/airperl.pl' -x 'AirfoilTools/.DS_Store' -x 'AirfoilTools/.vscode/' -x 'AirfoilTools/resources/old/*';mv .vscode AirfoilTools/; unzip -t $ZFN |egrep '(DS_Store|vscode|airperl|old\b)'

import adsk.core, adsk.fusion, adsk.cam, traceback, math, os, subprocess, json, sys, select, socket, io, datetime, re, urllib.parse, time
from .foildb2020 import foildb2020
import sqlite3 as lite # Works! 2021 - see sqlite_test_foildb.py
# import gettext, sqlite3, _thread
# from os import altsep



#################################################################################
####                                                                          ###
##                                Global defaults                              ##
####                                                                          ###
#################################################################################

(program_name,aft) = ('Airfoil Tools','_aft')                                   # Easy name-change (e.g. "Pro" version), and id suffix (e.g. '_aftpro') so (a) nothing clashes, and (b) we can run 2 at once, (c) can have different graphics (remember to update folder names!)

prog_folder= os.path.dirname(os.path.realpath(__file__))                        # We run companion programs from here, and some API calls require full path
settings_file = 'user_settings'+aft+'.json'                                     # Things like firstrun flag, units, history helpers, etc - see also home_file() for where this gets put
update_file = 'latest_version'+aft+'.json'                                      # Created async by check_update.py - contains JSON from website with current version data in it
resource_folder = '.' + os.sep + 'resources'                                    # Application.currentUser ?
abs_resource_folder = os.path.join(prog_folder,'resources')                     # clips cannot use relative folders :-()
userdata={}                                                                     # global var - loaded from settings_file. Keys* include:   debug, release, share_errors, announcements, email, displayName  ... * All with +aft appended
cloud_cache={}                                                                  # global var - cache cloud_foil web lookups

panelorder = 0                                                                  # auto-increment ID so our dict has an order later
handlers = []                                                                   # global set of event handlers to keep them referenced for the duration of commands
mru={'mru':'','mruno':0}                                                        # Most-recently-used counter and pointer so we know the order they changed UI control in
(pnl,cmdh)=({},{})                                                              # All our panel control objects in one easy-to-get-to place
debug=False                                                                     # Might get toggled from the userdata later
(textpalette,textpalette_ori_state)=(None,False)                                # Gets set if the user wants to watch our commands/progress
units={'mps':{'scale':100},'m':{'scale':100},'C':{'offset':273.15}}             # How to convert to what I need from fusion numbers
_ui=None
pnl['loaded']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())
pnl['openc']=0
nofiletxt='(no file selected yet)'
if sys.platform == 'win32':
    store_url='https://apps.autodesk.com/FUSION/en/Detail/Index?id=5447707798035545266&os=Win64&appLang=en'
else:
    store_url='https://apps.autodesk.com/FUSION/en/Detail/Index?id=5447707798035545266&os=Mac&appLang=en'
forum_url='https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985'
fivestar_url='https://forums.autodesk.com/t5/fusion-360-api-and-scripts/airfoiltools-add-in-current-feature-list-reply-to-this-topic-to/td-p/9888073'
infoa='<div style="color: #b5b5b5;">' # for showing derived values in the controls dialogue ui
infob='</div>'
radperstorpm = 0.10471975511966   # multiply degrees by this to get fusion360 units.  Divide fusion360 numbers (radians per second) by this to get RPM
green=norm='' # ANSI colours

# how we deal with critical program errors
def err(sharejson,emsg=None,dlg=None):
    if emsg is None: emsg=sharejson['Error']
    eprint(emsg); vlog(emsg); sharejson['Error']=emsg
    if userdata.get('share_errors'+aft,1):                                      # Don't send us anything about this crash if the user has said not to in their settings
        shareIt(sharejson,dlg=dlg)
    adsk.core.Application.get().userInterface.messageBox(emsg)
    return sharejson # for sharejson['Error']


def main_aft():
    global uiel, userdata, updatedata, uiel_order, cmdl, pnl, debug, NEXTVERSION, clipa, clipb, green, norm

    userdata=load_user_settings(settings_file)                                  # Get the users settings, defaults, preferences, presets, etc - empty if doesn't exist
    if not userdata:
        userdata={'debug':False,'release':True}                                 # Gets overwritten shortly by defaults in run() because 'firstrun' flag is missing here
    elif userdata.get("debug",0):
        debug=True
        green='\033[32;1m'
        norm='\033[0m'
    default_user_settings(userdata)                                             # Set up any more potential first-run settings

    updatedata=load_user_settings(update_file)                                  # Check any old previously-loaded update info. Gets re-populated when the user opens any dialogue
    if updatedata.get('fatal',0):                                               # Note that this is "stale" info - the update data was fetched on a previous run; not this one.
        app = adsk.core.Application.get()
        ui  = app.userInterface
        ui.messageBox(updatedata['fatal'])                                      # Unlikely - something so bad that continuing to run is a major problem, and no solution possible
        sys.exit(1)                                                             # quit to prevent catastrophe (e.g. data corruption or security issue)
    NEXTVERSION='?'                                                             # Gets updated to a next version string if an update is available
    if updatedata.get('loaded',0) and updatedata.get('current_version'+aft,0):
        NEXTVERSION='{:1.8f}'.format(updatedata['current_version'+aft])
    if userdata.get('applyupdate',0):                                           # Complete any file updates that might not have been applied earlier
        if updates(False):                                                      # Does the update
            userdata['applyupdate']=0
            save_user_settings(None,settings_file) 
    web_nbreq(None) # proccess any non-blocking operations

    #eprint("Hello {}".format(__name__)) # Hello __main__%2FUsers%2Fcnd%2FLibrary%2FApplication%20Support%2FAutodesk%2FAutodesk%20Fusion%20360%2FAPI%2FAddIns%2FAirfoil%20Tools%2FAirfoil%20Tools_py



    #################################################################################
    ####                                                                          ###
    ##              Button and DropDown (slide-out) Menu Definitions.              ##
    ####                                                                          ###
    #################################################################################

    also='\n\nAlso inserts associated construction points and user-parameters (moments,' \
            'lift, drag, max L/D info, name, etc) useful for working with your new foil or shape.' # frequent static text

    # We create all our tooltip files at runtime, so all the doc is in this file instead of scattered around.     
    clipa='<!DOCTYPE html>\n<!-- CAUTION: auto-generated file. Do not change. (gets overwritten.) -->\n' \
        '<html><body BGCOLOR="#515151" WIDTH=300 TEXT="#ffffff" LINK="#ffffff" VLINK="#ffffff" ALINK="#ffffff" style="width:300">' \
        '<BASEFONT SIZE=3 FACE="Roboto,Arial,"Segoe UI","Helvetica Neue",sans-serif"><div style="color:white; width:300px;">' \
        '<table border=0 cellpadding=0 cellspacing=0 style="color:white; width:300px;"><tr><td width=298><b>'
    clipb='</b></tr></td></table></div></body></html>'

    # Put icons and clips into ./resources/[parentid]_[commandid_aft] (e.g ./solid_airfoil_command_aft_create_wing_aft/create_wing_aft.html). N.B. Uses *existence* of ['id'].html or .png to decide about using this folder.
    # if above missing, default:  ./resources/[commandid_aft]
    # if both missing, default: ./resources/
    uiel = {'airfoil_toolbar_button'+aft:
            { 'name':'Airfoils', # Button name - Fusion360 makes this AIRFOILS
                'type':'Button', 'ordr':nextId(),
                'desc':'Create, manage, and optimise airfoil shapes\nfor air, water, or other gasses/fluids.' +    # This text never shows anyplace - see about_help_command for what does show
                        '\n\n' + #             'info':
                        'Contains tools to create Wing surfaces, Struts, ' +
                        'Cowling shapes, Ducts, Propellers and Turbines.\n\n ' +
                        'Also has utilities to read and write airfoil datafiles; ' +
                        'to simplify, clean, repair, and optimise file formats; to run ' +
                        'particle swarm genetic optimisation (CFD) for ' +
                        'maximising performance for your specific need, '+
                        'performance analysis with charting and comparison '+
                        'tools, spline sketch insertion, comprehensive air/hydro '+
                        'foil library with precomputed ideal shapes, foil input '+
                        'scanning from photographs, calculators for Reynolds '+
                        'Numbers, fluid viscosity, power, torque, etc; and more.'},

        'solid_airfoil_command'+aft: # key is also the ID                       # all our user interface elements are in here
            { 'name':'Airfoil', # program_name, # Airfoil Tools
                'type':'DropDown', 'ordr':nextId(),
                'after':['PrimitivePipe'],                                      # Where on the menu to place our button after
                #'html.SolidCreatePanel': "stuff",                              # Example of how to have different words depending on who called us
                'desc':'Creates a solid airfoil, strut, cowling, duct propeller, or turbine.\n\n' +                # This text never shows anyplace - see about_help_command for what does show
                        'Select points or a path to locate your shape, and then ' + # Fusion360 auto-wraps these OK
                        'specify your expected operating conditions to ' +
                        'insert an ideal stream-lined foil (lofted spline) that is ' +
                        'CFD-Optimised to your need.'+also },

        'sketch_airfoil_command'+aft:
            { 'name':'Airfoil', # program_name, # Airfoil Tools
                'type':'DropDown', 'ordr':nextId(),
                'after':['DrawPoint'],                                             # Where on the menu to place our button after
                'desc':'Inserts airfoil points with a spline into the active sketch.\n\n' +                        # This text never shows anyplace - see about_help_command for what does show
                        'Select a chord line or nose and tail points, and ' +
                        'then specify your expected operating conditions.\n\n '+
                        'Can insert 2D or 3D sketch geometry.'+also },


        # Above are top-level controls.  Below is what goes into the above
        'about_help_command'+aft:                                                # This is the first menu item, because it is the one which has it's icon and text promoted to the parent control (airfoil button etc)
            { 'name':'Use the drop-down menu to insert foils in a new sketch', # 'About '+program_name+' + help/videos',
                'name.SketchCreatePanel':'About '+program_name+' + help/videos',
                'name.SolidCreatePanel':'About '+program_name+' + help/videos',
                'type':'Command', 'ordr':nextId(),
                #'tip':'Use the drop-down to open the sub-menu',

                'desc':'<hr><p>This Add-on lets you create, manage, and optimise airfoil shapes for air, water, or other gasses/fluids.</p>' + 
                        '<p>It contains tools to create Foil sketch profiles, Wing surfaces, Struts, ' +
                        'Cowling shapes, Ducts, Propellers and Turbines.</p>' +
                        '<p>It also has utilities to read and write airfoil datafiles; ' +
                        'to simplify, clean, repair, and optimise file formats; to run ' +
                        'particle swarm genetic optimisation (CFD) for ' +
                        'maximising performance for your specific need, '+
                        'performance analysis with charting and comparison '+
                        'tools, spline sketch insertion, comprehensive air/hydro '+
                        'foil library with precomputed ideal shapes, foil input '+
                        'scanning from photographs, calculators for Reynolds '+
                        'Numbers, fluid viscosity, power, torque, etc; and more.</p>',

                'desc.SolidCreatePanel':'Creates a solid airfoil, strut, cowling, duct propeller, or turbine.\n\n' +                # This text never shows anyplace - see about_help_command for what does show
                        'Select points or a path to locate your shape, and then ' + # Fusion360 auto-wraps these OK
                        'specify your expected operating conditions to ' +
                        'insert an ideal stream-lined foil (lofted spline) that is ' +
                        'CFD-Optimised to your need.'+also,

                'desc.SketchCreatePanel':'Inserts airfoil points with a spline into the active sketch.\n\n' +                        # This text never shows anyplace - see about_help_command for what does show
                        'Select a chord line or nose and tail points, and ' +
                        'then specify your expected operating conditions.'+also,

                'html':'<br><img src=../foilsep.png><br><p>' + 
                        program_name+' is written and maintained by:</p><p>'+
                        '<center>Chris Drake</center></p><p>' +
                        'Run this command to access our list of help topics and view '+
                        'quick introductory videos.</p>' ,
                'controls':[ 'banner', 'update_now', 'helpbanner', 'rating_group', 'rate_me'  ] },


        'separator_1'+aft:{'ordr':nextId(), 'type':'Separator' },                # visually separate logically distinct commands

        'create_wing'+aft:
            { 'name':'Create Wing',
                'name.SketchCreatePanel':'Wingt',          # Alternate name to use if the parent is SketchCreatePanel
                'name.sketch_airfoil_command'+aft:'Wing',  # Alternate name to use if the parent is this
                'type':'Command', 'ordr':nextId(),
                #'desc.SketchCreatePanel':'Creates a Wing (lifting airfoil)\n\n' +
                'desc':'Creates a Wing (lifting airfoil)\n\n' +
                        'Select:-\n\n' +
                        '*  The Nose point (leading edge) of the foil\n' +
                        '*  Tail point (trailing edge)\n'+
                        '     these make the chord length\n' +
                        '+  Optional Join point - to insert foil\n'+
                        '     pre-rotated to best L/D AoA\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, and operating ' +
                        'conditions to insert the optimum foil shape for your purpose.\n\n'+
                        # '* Tip: See also the Mass and Span option below\n\n' +
                        # 'Tip: If you select only the Nose or Tail (not both), you can then specify ' +
                        # 'a span and load or weight instead, and ' + program_name + ' will work out ' +
                        # 'the best chord length for you.\n\n' + 
                        'Tip: A good Join Point is usually 25% in from the nose, which is very close ' +
                        'to the foil center of pressure.  Your foil will be auto-rotated by the best ' +
                        'angle of attack (AoA, or alpha) for the most efficient lift-to-drag (L/D) ' +
                        'condition matching your specifications.\n\n' +
                        'Tip: select the chord line to auto-rotate the foil about its center-of-pressure moment.',
                'desc.SolidCreatePanel':'Creates a:\n\n        *  Solid Wing (lifting airfoil)\nor    * Strut (low-drag symmetric foil).\n\n' + # placeholder reminder how to do different desc/images for different panels
                        'Select one of:-\n\n' +
                        '*  Chord line, forward direction, and width\n' +
                        '*  Nose and Tail points, and width\n' +
                        '*  Path and forward direction\n' +
                        '+ Optional moment center\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, and operating ' +
                        'conditions to insert the optimum foil shape for your purpose.',
                'res':['SketchCreatePanel','SolidCreatePanel'], # gets overwritten below - this tells our loader that this item has its own resource folder
                'html':'<center><img src=../create_wing_aft.png></center><br>'+also,
                'html.SolidCreatePanel':'<img src=../../foilsep.png><br>'+also,
                'html.SketchCreatePanel':'<center><img src=../../create_wing_aft.png></center><br>'+also,
                # 'controls':[ 'banner', 'update_now', 'nickname', 'wingorstrut', 'nose', 'tail', 'chord_info', 'join', 'preset', 
                'required':{'nose':1, 'tail':1, 'medium':1, 'turbulence':1, 'finish':1, 're':1}, # 'chord':1, 
                'controls':[ 'banner', 'update_now', 'nickname', 'nose', 'tail', 'chord_info', 'join',
                            'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', 're_info',

                            # 'velocity_group', 'targetspeed', 
                            # 'altitude_group', 'targetaltitude', 
                            # 'temperature_group', 'targettemperature', 

                            'velocity_group', 'minspeed', 'targetspeed', 'maxspeed',
                            'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                            'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature', 'mach_info',
                            # 'mass_and_span_group', 'min_mass', 'target_mass', 'max_mass', 'target_span',
                            'spline_group', # 'spline_type',
                            'num_points', 'nt_concentrate', 'cloud_foil', # 'num_rails',
                            'settings_group', 'minthick',
                            'wrap_axis', 'up', # put this below the preset, so folk know that preset includes this.
                            'constrain', 'camber', 'endcap', 'enddist',
                            'share', 'helpok', 'rating_group', 'rate_me'   ] }, # 'save_preset_group', 'save', 'log', 

        'create_strut'+aft:
            { 'name':'Create Strut',
                'name.SketchCreatePanel':'Strut', # Alternate name to use if the parent is SketchCreatePanel
                'name.sketch_airfoil_command'+aft:'Strut', # Alternate name to use if the parent is this
                'type':'Command', 'ordr':nextId(),
                'desc.SketchCreatePanel':'Creates a strut (low-drag symmetric foil).\n\n' +
                        'Select:-\n\n' +
                        '*  The Nose point (leading edge) of the foil\n' +
                        '*  Tail point (trailing edge)\n'+
                        '     these make the chord length\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, and operating ' +
                        'conditions to insert the optimum foil shape for your purpose.',
                'desc':'Creates a strut (low-drag symmetric foil).\n\n' +
                        'Select one of:-\n\n' +
                        '*  Chord line, forward direction, and width\n' +
                        '*  Nose and Tail points, and width\n' +
                        '*  Path and forward direction\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, and operating ' +
                        'conditions to insert the lowest-drag foil shape for your purpose.',
                'res':['SketchCreatePanel','SolidCreatePanel'], # gets overwritten below - this tells our loader that this item has its own resource folder
                'html':'<center><img src=../create_strut_aft.png></center><br>' + also,
                'html.SolidCreatePanel':'<center><img src=../../create_strut_aft.png></center><br>' +also,
                'html.SketchCreatePanel':'<center><img src=../../create_strut_aft.png></center><br>'+also,
                'required':{'nose':1, 'tail':1, 'medium':1, 'turbulence':1, 'finish':1, 're':1}, # 'chord':1, 
                'controls':[ 'banner', 'update_now', 'nickname', 'nose', 'tail', 'chord_info', 
                            'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', 're_info',
                            'velocity_group', 'minspeed', 'targetspeed', 'maxspeed', 
                            'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                            'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature', 'mach_info',
                            'spline_group', # 'spline_type',
                            'num_points', 'nt_concentrate', 'cloud_foil', # 'num_rails',
                            'settings_group', 'minthick', 'wrap_axis', 'constrain', 'camber', 'endcap', 'enddist',
                            'share', 'helpok', 'rating_group', 'rate_me'  ] }, # 'save_preset_group', 'save', 'log', 

        'create_cowling'+aft:
            { 'name':'Create Cowling',
                'name.sketch_airfoil_command'+aft:'Cowling',
                'type':'Command', 'ordr':nextId(),
                'desc':'Creates a solid cowling (low-drag streamline shape)\n\n' +
                        'Select one of:-\n\n' +
                        '*  Chord line, forward direction, and thickness\n' +
                        '*  Nose and Tail points, and thickness\n' +
                        '+ Optional width for non-circular shrouds\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, and operating '+
                        'conditions to insert an optimum low-drag 3D shape for your purpose.',
                'html':'<center><img src=../create_cowling_aft.png></center><br>'+also,
                'controls':[ 'banner', 'update_now', 'votebanner', 'votebox' ] },

        'create_duct'+aft:
            { 'name':'Create Airfoil Duct',
                'name.sketch_airfoil_command'+aft:'Duct',
                'type':'Command', 'ordr':nextId(),
                'desc':'Creates a Duct suitable for use as a propeller or turbine shroud.\n\n' +
                        'Select:-\n\n' +
                        '*  Center Chord line, forward direction,\n'+
                        '    and inner diameter\n' +
                        '*  Center Nose and Tail points, and\n'+
                        '    inner diameter\n' +
                        '+ Optional thickness constraint or outer\n'+
                        '    diameter\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, ' +
                        'propeller or turbine power, specifications, and operating '+
                        'conditions to insert an optimal duct for your purpose.',
                'html':'<center><img src=../create_duct_aft.png></center><br>'+also,
                'controls':[ 'banner', 'update_now', 'votebanner', 'votebox' ] },

        'create_propeller'+aft:
            { 'name':'Create Propeller',
                'name.sketch_airfoil_command'+aft:'Propeller',
                'type':'Command', 'ordr':nextId(),
                'desc':'Creates a Driven Propeller.\n\n' +
                        'Select:-\n\n' +
                        '       *  Center Chord line, forward direction\n' +
                        'or   *  Center Nose and Tail points\n\n' +
                        'Next specify one of:-\n\n' +
                        '*  Power input available or output required\n' +
                        '*  Blade count and length or diameter\n' +
                        '+ Optional RPM range\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, ' +
                        'material and medium specifications, and operating '+
                        'conditions to insert optimal blade shape for your purpose.',
                'html':'<center><img src=../create_propeller_aft.png></center><br>'+also,
                'required':{'centerline':1, 'hub_nose':1, 'tip_location':1, 'medium':1, 'turbulence':1, 'finish':1,
                            'targetspeed':1, 'targettemperature':1, 'turbine_strategy':1, 'rpm':1, 'target_thrust':1, 'chord_formula':1 },
                'controls':[ 'banner', 'update_now', 
                             'nickname',
                             #'beta_group', 'beta_info',
                             'location_group', 'select_outline', 'centerline', 'root_chord_info', 'hub_nose', 'tip_location', 'tip_chord_info', 'diameter_info',

                             'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', 
                             'prop_velocity_group', 'minspeed', 'targetspeed', 'maxspeed', 
                             'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                             'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature',

                             'strategy_group', 'turbine_strategy', 'chord_formula', 'formula_instructions',
                             'constraints_group', 'rotate_direction', 'rpm', 'tip_speed', 'tip_mach_info', # 're_info',  # 'blade_len',
                             # 'blade_count',
                             # 'vary_pitch',
                             'num_profiles', 'minthick', 'camber',
                             # 'power_group', 'powercruise', 'powermax',
                             # 'torque_group', 'torquecruise', 'torquecruiserpm', 'torquemax', 'torquemaxrpm', 'reduction',
                             'thrust_group', 'target_thrust', 

                             'spline_group', # 'spline_type',
                             'num_points', 'nt_concentrate', 'num_rails', 'patchme', 'cloud_foil',

                             #'save_preset_group', 'save', 'remove',
                             'settings_group', 'share', 'helpok', 'rating_group', 'rate_me'  ] },

        'create_turbine'+aft:
            { 'name':'Create Turbine',
                'name.sketch_airfoil_command'+aft:'Turbine',
                'type':'Command', 'ordr':nextId(),
                'desc':'Creates a Driving Turbine.\n\n' +
                        'Select:-\n' +
                        '       *  Center Chord line\n' +
                        '       *  Nose point (flow entry point = direction)\n' +
                        '       *  Hub edge join point\n\n' +
                        'Next specify one of:-\n\n' +
                        '       * Power input available or output required\n' +
                        '       * Blade count and length or diameter or tip location\n' +
                        '       +  Optional RPM range\n\n' +
                        'Then specify your medium (air, water, etc), velocity range, ' +
                        'material and medium specifications, and operating '+
                        'conditions to insert the optimal blade shape for your purpose.',
                'html':'<center><img src=../create_propeller_aft.png></center><br>'+also,
                #'required':{'nickname':1, 'hub_radius':1, 'turbulence':1, 'finish':1, 'medium':1, 'targetspeed':1, 'targettemperature':1, 'turbine_strategy':1, 'root_chord_info':1, 'hub_edge':1, 'hub_nose':1, 'tip_location':1, 'rpm':1},
                'required':{'centerline':1, 'hub_nose':1, 'tip_location':1, 'medium':1, 'turbulence':1, 'finish':1,
                            'targetspeed':1, 'target_efficiency':1, 'targettemperature':1, 'turbine_strategy':1, 'rpm':1, 'chord_formula':1 },
                'controls':[ 'banner', 'update_now', 
                             'nickname',
                             #'beta_group', 'beta_info',
                             'location_group', 'centerline', 'root_chord_info', 'hub_nose', 'tip_location', 'tip_chord_info', 'diameter_info', 'radius_nr', 'tip_chord_nr', 'root_chord_nr',

                             'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', # 'tip_mach_info', # 're_info',
                             'velocity_group', 'minspeed', 'targetspeed', 'maxspeed', 
                             'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                             'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature',

                             'strategy_group', 'turbine_strategy', 'chord_formula', 'formula_instructions',
                             'constraints_group', 'rotate_direction', 'rpm', 'tip_speed', 'tip_mach_info', 'target_efficiency',
                             # 'blade_count', 'vary_pitch', 
                             'num_profiles', 'minthick', 'camber',

                             'spline_group', # 'spline_type',
                             'num_points', 'nt_concentrate', 'num_rails', 'patchme',

                             #'save_preset_group', 'save', 'remove',
                             'settings_group', 'share', 'helpok', 'rating_group', 'rate_me'  ] },

        'separator_2'+aft:{'ordr':nextId(), 'type':'Separator' },               # visually separate logicily distinct commands

        #'import_export'+aft: # Cannot have different selection elements in one form - prevent the execute handler from running.
        #    { 'name':'Import / Export airfoil data files',
        #        'type':'Command', 'ordr':nextId(),
        #        'desc':'Intelligently loads any kind of airfoil data file.\n' +
        #                'Exports generated and/or inbuilt library airfoils to text data files in Selig format.\n' +
        #                'Insert airfoils into sketch with attached spline.\n' +
        #                'Loads and saves foils and favorites into your personal library.',
        #        'controls':['banner', 'update_now', 
        #                    'import_export_but',                                   # (radio group)
        #                    'import_group',
        #                    'nose', 'tail', 'chord_info', 'join',              # for import
        #                    'aoa', 'spline_type', 'interpolate',
        #                    #'settings_group',
        #                    'minthick', 'up',
        #                    'constrain', 'endcap', 'enddist', 'share',
        #                    'nofix', 
        #                    'export_group',
        #                    'select_spline',                                   # for export
        #                    'file_group',
        #                    'file_button', 'file_name',
        #                     ] },

        'import_foil'+aft:
            { 'name':'Import airfoil from data file',
                'type':'Command', 'ordr':nextId(),
                'desc':'Intelligently loads any kind of airfoil data file,\n' +
                        'and inserts it into your sketch with attached spline.',
                        #'Loads and saves foils and favorites into your personal library.',
                'required':{'nose':1, 'tail':1, 'file_name':1},
                'controls':['banner', 'update_now', 
                            'file_button', 'file_name',
                            'nose', 'tail', 'chord_info', 'join',              # for import
                            'aoa', 'spline_type', 'interpolate',
                            #'spline_group', # 'spline_type',
                            #'num_points',
                            'nt_concentrate', # 'num_rails',
                            #'settings_group',
                            'minthick',  'wrap_axis', 'up',
                            'constrain', 'camber', 'endcap', 'enddist',
                            'nofix', 'share',
                            'helpok', 'rating_group', 'rate_me'
                             ] },

        'export_foil'+aft:
            { 'name':'Export an airfoil sketch shape to a data file',
                'type':'Command', 'ordr':nextId(),
                'desc':'Select the spline shape of any airfoil to\n'+
                       'export it to a file in Selig .dat format.',
                'required':{'select_spline':1, 'file_name':1},
                'controls':['banner', 'update_now', 
                            'select_spline',                                   # for export
                            'original', 'normalize', 'threed', 'metadata', 'datcsv',  # 'unwrap', for v3
                            #'test_spline',                                     # for export # unwrap_axis ?
                            'file_button', 'file_name', 'share', 'helpok', 'rating_group', 'rate_me'
                             ] },

        'foil_maintenance'+aft:
            { 'name':'Airfoil data maintenance operations', #  [declutter] [normalize] [scale] [close] [spline/points-only] [compare] => can be sideways arrows in menus
                'type':'Command', 'ordr':nextId(), # Better as a DropDown ?
                'desc':'Offers a variety of processing operations on airfoil data\n\n' +
                        '*  Full Tidy: all the below\n\n' +
                        '*  Test: Detect data problems, like excessive angles, holes, duplicates, crossovers, etc\n\n' +
                        '*  Declutter: remove extraneous polar coordinate\n\n' +
                        '*  Normalize: make the front point the nose at (0,0) and the tail (1,0)\n\n' +
                        '*  Scale: Shrink or grow, symmetrically or asymmetrically\n\n' +
                        '*  Rotate: Change the angle of the airfoil data polar coordinates\n\n' +
                        '*  Close: join up any data holes (common at trailing ends)\n\n' +
                        'These utilities are for "foil geeks" only.  It is better to use foils '+
                        'suggested by '+program_name+' because ' +
                        'most 3rd party shapes are extremely inefficient (e.g. our foils outperform all NACA shapes by 57% to 120%)',
                'controls':[ 'banner', 'update_now', 'maint_info', 'votebox' ] },

        'foil_performance'+aft:
            { 'name':'Airfoil performance analysis', # compare
                'type':'Command', 'ordr':nextId(),
                'desc':'Produce polar diagrams, Lift and Drag coefficients, moments, ' +
                        'and more for your chosen range of operating conditions.\n\n'+
                        'Compare multiple different airfoils from our database, '+
                        'your projects, or your personal library.\n\n'+
                        'Can insert results as user parameters and sketch images to '+
                        'easily share with others',
                'required':{'select_spline':1, 'medium':1, 'turbulence':1, 'finish':1, 're':1, 'votebox':1},
                'controls':['banner', 'update_now', 
                            'soon_group', 'soon_info',
                            'select_spline',                                   # from export
                            'chord_info',
                            'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', 're_info',
                            'velocity_group', 'minspeed', 'targetspeed', 'maxspeed', 
                            'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                            'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature', 'mach_info',
                            # 'mass_and_span_group', 'min_mass', 'target_mass', 'max_mass', 'target_span',
                            'settings_group',
                             'share', 'helpok',
                             'votebanner', 'votebox' ] },

        'foil_optimize'+aft:
            { 'name':'Airfoil CFD genetic optimiser',
                'type':'Command', 'ordr':nextId(),
                'desc':'Specify the ranges of all operating conditions along with your weighting '+
                        'criteria to automatically generate the perfect shape to exactly suit your needs.\n\n' +
                        'Run on your own high-performance machine(s) to see results in a week, or distribute ' +
                        'your workload in the cloud to get faster answers.\n\n'+
                        program_name+' already includes pre-computed foils for many conditions, '+
                        'which typically outperform common shapes by 50% or more.  Use '+
                        'our genetic optimiser to get even better results',
                'controls':[ 'banner', 'update_now',
                             'soon_group', 'soon_info',
                             'votebanner', 'votebox' ] },

        'reynolds_calculator'+aft:
            { 'name':'Reynolds (Re) Number and Power Calculators',
                'type':'Command', 'ordr':nextId(),
                'desc':'Intelligent tool for working with the numbers that make ideal foils possible.\n\n'+
                        'Can load and save named inputs as favorites to help make your subsequent use of foils easier',
                'controls':[ 'banner', 'update_now', 'reynolds_info', 'votebanner', 'votebox' ] },

        'global_settings'+aft:
            { 'name':'Settings',
                'type':'Command', 'ordr':nextId(),
                'desc':'Global settings and preferences for how ' + program_name + ' operates.',
                'required':{'text_info':1},
                'controls':[ 'banner', 'update_now', 'text_info', 'announcements', 'displayName', 'email', 'share_errors', 'check_update', 'hide_rating', 'rating_group', 'rate_me' ] }, # ideas; update-checks, cache-last, expert, ...


        'separator_3'+aft:{'ordr':nextId(), 'type':'Separator' },               # visually separate logicily distinct commands

        'foil_bugs_and_requests'+aft:
            { 'name':'Report Bug / Request Feature',
                'type':'Command', 'ordr':nextId(),
                'desc':program_name + ' is under continual development.  It began in 2012 for high-performance ultralight propellers, and has morphed into a powerful suite of tools ' +
                    "for almost anything you can use a foil shape for.\n\n" +
                    "This Fusion360-edition is our first User-Interface to our existing code base, which contains vastly more features than we have so-far exposed in our UI.\n\n"+
                    "If you've found something that doesn't look right, or reached a UI command that we have not yet connected to our code base, or you have a request for a new feature "+
                    "you would like to see, please get in touch!\n\nRun this command to contact the author: Chris Drake.\n\n",
                'controls':[ 'banner', 'update_now', 'bugbanner' ] },
    #              'desc':"\n\nComing Shortly!  This command exists in our CLI tools package installed with this add-in, but we are still working on plumbing it's goodness into the Fusion360 UI.\n\n" +
    #                    "Use our feature-request option if you'd like us to prioritize something specific, or sit back and wait a few weeks for us to join all the dots, then check for an update!\n\n",


        'end'+aft: {'type':'end','ordr':0}  # add above here
        }



    #################################################################################
    ####                                                                          ###
    ##                     Dialogue box controls Definitions.                      ##
    ####                                                                          ###
    #################################################################################
    if sys.platform.startswith('darwin'): blueline= '<img src="'+  os.path.join(abs_resource_folder,'blueline3.png') +'" height=3>' # Mac irritatingly is 3 pixels higher...
    else: blueline= '<img src="'+  os.path.join(abs_resource_folder,'blueline.png') +'" height=1>'

    # This huge blob of text is repeated 3 times, so we define it just once here instead.
    prophtml='<h3>Free Tips</h3>' \
             '<p>If the tips of your turbine blades do not have a duct, we work outwards from the hub reducing the ' \
             'chord up to the 40% point, thereafter we select a chord for the rest of the blade such that the ' \
             'Reynolds number stays constant to the tip.</p>' \
             '<center><img src=formula_free_tips.png></center>' \
             '' \
             '<h3>Inside Duct</h3>' \
             '<p>We maximise the chord of the blade, without exceeding your chosen root chord.  This causes ' \
             'more-efficient higher Reynolds Numbers, which  are safe to use since the duct prevents tip loss.</p>' \
             '<center><img src=formula_duct.png></center>' \
             '' \
             '<h3>Custom</h3>' \
             '<p>Enter a python formula to specify Y (section chord) and Z (trailing edge location) for each X (from 0=hub to 1=tip).</p>' \
             '<center><img src=formula_custom.png></center>' \
             '' \
             '<h3>NOTE:</h3>' \
             '<table border=0><tr><td>*</td><td>' \
             'The length of the axis line you select determines the starting blade chord at the hub (see chord formula).' \
             '</td></tr><tr><td>*</td><td>Both strategies above create blades with straight trailing edges, to minimise wake turbulence.' \
             '</td></tr><tr><td>*</td><td>Picking an existing strategy will populate its corresponding formula into the input box.' \
             '</td></tr><tr><td>*</td><td>Open the <b>formula instructions</b> link for details.</td></tr></table></p>'



    cmdl=[{'id':'banner'+aft,   # This gets overwritten by update messages, if any.   WARNING - must be first in list (see update code which grabs cmdl[0]['dflt'])
        'name':'',
        'type':'TextBoxCommandInput',
        'dflt':'<img border=0 align=left valign=top height=45 src="'+  os.path.join(abs_resource_folder,'banner_aft45.png') +'">' +
               '<b> &nbsp; &nbsp; &nbsp; &nbsp; <font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
               '<br> &nbsp; &nbsp; &nbsp; &nbsp; <font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
               '<br> &nbsp; &nbsp; &nbsp; <a href="http://foils.pro/Airfoil_Tools/">instructions</a> &nbsp; '+
               '<a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a><br>'+ blueline,

# Mac and Windows behave differently (url() does not work on windows, tables are broken, etc)
    #works    'dflt':'<div align="center"><img src="'+  os.path.dirname(os.path.realpath(__file__)) + "/resources/" + 'banner.png">'+program_name+' v<b>'+str(VERSION)+'<br><img src=resources/banner.png></b><a href="http:fusion360.autodesk.com">html.</a></div>',
    #works    'dflt':'<div align="center"><img src="'+  abs_resource_folder + '/banner.png">'+program_name+' v<b>'+str(VERSION)+'</b><a href="http:fusion360.autodesk.com">html.</a></div>',
    #OK!    'dflt':'<div align="center" style="overflow: visible;"><img src="'+  abs_resource_folder + '/banner.png" width=207 border=0 height=43></div>', # 207 ok 208 ng  43ok 44ng
#        'dflt':'<table border=0 cellpadding=0 cellspacing=0><tr><td rowspan=2><img src="'+  os.path.join(abs_resource_folder,'banner_aft.png') +'"></td><td align=right>' +
#               '<b><font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
#               '<br><font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
#               '<br><a href="http://foils.pro/Airfoil_Tools/">instructions</a> &nbsp; <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a>'+
#               '</td></tr><td height=1px style="height:1px" bgcolor="#0091e2"></td></tr></table>', # 207 ok 208 ng  43ok 44ng
#                '<div align="right" style="overflow: visible; background-image: url('+  os.path.join(abs_resource_folder,'banner.png') +'); background-repeat: no-repeat;"><b>'+
#                '<b><font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
#                '<br><font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
#                '<br><a href="http://foils.pro/Airfoil_Tools/">instructions</a> &nbsp; <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a>'+
#                '</div>', # 207 ok 208 ng  43ok 44ng
        'last':0},

        { 'id':'update_now'+aft,
        'name':'Update Now?',
        'type':'BoolValueInput',
        'invisible':True, # Hides it
        'desc':'An update to ' +program_name + ' is waiting to be installed.',
        'tip':'Check this box to update ' + program_name + ' to version ' + NEXTVERSION},
        #'tip':'Check this box to update ' + program_name + ' to the next version.'},

        { 'id':'text_info'+aft,
        'name':'Text info',
        'type':'RadioButtonGroupCommandInput',
        'choices':['Summary','Full','Off (None)'],
        'dflt':'Summary',
        'tip':'How much extra information about the operation\n'+
               'of ' + program_name + ' do you want to show in the\n'+
               'Text-Commands window at the bottom of your screen?'},

        { 'id':'share_errors'+aft,
        'name':'Share Errors',
        'type':'BoolValueInput',
        #'dflt':True,
        'tip':'Send us crash data?',
        'desc':'If our program crashes or encounters an error\n'+
              'this option sends to our web server:\n'+
              '* the details of what went wrong\n'+
              '* your input data which caused the problem.'},

        { 'id':'check_update'+aft,
        'name':'Check Now',
        'type':'BoolValueInput',
        #'dflt':True,
        'tip':'Check this box to check right now for any new updates.',
        'desc':"Note that "+program_name+" always checks for updates when first loaded each day.  This option merely checks again right now, even if it's already checked today.",
        'last':0},

        { 'id':'hide_rating'+aft,
        'name':'Hide Rater',
        'type':'BoolValueInput',
        #'dflt':True,
        'tip':'Remove our rate-us message',
        'desc':"Check this box after you've rated "+program_name+" 5-stars in the app store, to remove our "+'"Rate Us" message from the bottom of our menu options',
        'last':0},

        { 'id':'announcements'+aft,
        'name':'Announcements',
        'type':'BoolValueInput',
        #'dflt':True,
        'desc':'Receive '+program_name+' product announcements (e.g. new version notices and feature updates) by email?',
        'tip':'Be notified when new features arrive.'},

        { 'id':'email'+aft,
        'name':'Your email',
        'type':'StringValueInput',
        'tip':'Where to send your announcement emails',
        'desc':'We will email new '+program_name+' product announcements (e.g. new version notices and feature updates) to your email address here.',
        #'html':'<img src=user_parms_aft.png>', # HTML: Gets wrapped in clipa and clipb
        'last':0},

        { 'id':'displayName'+aft,
        'name':'Your name',
        'type':'StringValueInput',
        'tip':'For any announcements we might send.',
        #'desc':'',
        #'html':'<img src=user_parms_aft.png>', # HTML: Gets wrapped in clipa and clipb
        'last':0},

        {'id':'helpbanner'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p>'+ program_name+' &copy; Copyright 2020 <br>is written and maintained by</p>\n\n'+
            '<p>                              <center>Chris Drake</center></p>\n\n' +
            '<p>Use the links above to access our online help or contact Chris.</p><hr>'+
            '<p>' + program_name + ' is under continual development. &nbsp; It began in 2012 for high-performance '+
            'ultralight propellers, and has since become a powerful suite of tools ' +
            'for almost anything you can use a foil shape for.</p>\n\n' +
            '<p>This Fusion360-edition is our first User-Interface to our existing code base.  We are working to '+
            'bring you all the features it supports in this: our easy-to-use UI.</p><hr>'+
            '<p><b>Our inbuilt airfoil database:</b><br>We began with a huge collection of every public airfoil we could find online. '+
            'We next discarded all patented/copyright ones, and added every NACA 4-series permutation (and many 5-series).</p>'+
            '<p>We then ran 768 permutations of our solver (64 Re/12 N) on every one, as well as "gradient-descent squashing" on each (top and bottom surfaces individually).</p>'+
            '<p>We targeted Max CL/CD (or 1/CD for struts), weighted by normal distribution, across +/- 3degrees AoA for performance optimisation.</P>'+
            '<p>Finally, we took the top ten performers from each permutation we targeted, and ran genetic particle-swarm optimisation on those across 12 representative Mach numbers.</p>'+
            '<p>The foils we include here are the result of over <mark style="background-color: #FFFF80">1 million CPU-hours of processing</mark>, and <b>should</b> outperform anything else you can find online, usually by a large margin.</p>'+
            '<p>Our future cloud service will allow you to generate even more ideal foils suited to your exact usage and weighting criteria. </p></div>',
    #                   "If you've found something that doesn't look right, or reached a UI command that we have not yet connected to our code base, or you have a request for a new feature "+
    #                   "you would like to see, please get in touch!\n\nRun this command to contact the author: Chris Drake.\n\n</div>' ,",
        'rows':54,
        'last':0},

        {'id':'votebanner'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p>Do you need this feature?</p>\n'+
            '<p>'+ program_name + ' is a Fusion-360 user-interface added to our existing Airfoil utility system.</p>' +
            '<p>All features in our menu options exist, but it takes a lot of effort to wrap a UI around them.</p>' +
            '<p>If you want us to prioritize completing the UI for this feature, please vote below.</p>'+
                '</div>', # 207 ok 208 ng  43ok 44ng
        'rows':12,
        'last':0},

        { 'id':'votebox'+aft,
        'name':'Vote?',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'We prioritize our development based on the number of votes each feature receives',
        'tip':'Check this box to vote for this feature.'},


        {'id':'bugbanner'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p>Please use our  <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum link</a> to report bugs.</p>'+
            '<p>Remember to screen-shot any error boxes that popped up, and add them to your forum post to help us find and fix what went wrong.</p>'+
            '</div>', # 207 ok 208 ng  43ok 44ng
        'rows':7,
        'last':0},

        { 'id':'maint_info'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        #'dflt':infoa+'nose to tail = chord'+infob,
        'dflt':'All our supported maintenance operations are currently available through our <b>Import</b> and <b>Export</b> menu options.<br><br>\n\n'+
               'Let us know if you want a dedicated maintenance menu option (check the vote box below, and leave us a note in the <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a> about anything new you need to be supported).',
        'rows':10,
        'last':0},

        { 'id':'reynolds_info'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        #'dflt':infoa+'nose to tail = chord'+infob,
        'dflt':'Use our <b>Wing</b> or <b>Strut</b> options to access our calculator right now.\n\n'+
               'A future '+program_name+' version will support this stand-alone calculator.\n',
        'rows':5,
        'last':0},


        { 'id':'nickname'+aft,
        'name':'Nickname',
        'type':'StringValueInput',
        'tip':'(Optional) Give your foil a name. (No spaces or punctuation please)',
        'desc':'If you name your foil, the name will become a prefix in user parameters for the characteristics of this foil.'+
    #V2            'plus your foil and settings can be saved as a preset, so you can quickly re-use your parameters again in future. '+
                '\n\nThe nickname must start with a letter, then contain letters, numbers, and _ (no spaces).',
        'html':'<img src=user_parms_aft.png>', # HTML: Gets wrapped in clipa and clipb
        'last':0},

        { 'id':'wingorstrut'+aft,
        'name':'Wing or Strut',
        'type':'ButtonRowCommandInput',
        'tip':'', # leave this out - it overrides the button tips if used: 'Do you want a Wing (lifting shape) or Strut (Symmetrical low-drag shape)',
        'desc':'*  Select Wing to insert a shape optimised for\n'+
                '    best lift with lowest drag ( Max Cl/Cd )\n'+
                '*  Select Strut to get a shape optimised for\n'+
                '    lowest drag ( Min Cd )',
        'buttons':[{'name':'Wing (lifting shape)','res':os.path.join(resource_folder,'wing')},{'name':'Strut (low-drag symmetrical shape)','res':os.path.join(resource_folder,'strut')}],
        'html':'<img src=foilsep.png><br>(screen-shot showing wing vs strut - coming here soon)',
        'last':0},

        { 'id':'up'+aft,
        'name':'Up',
        'type':'ButtonRowCommandInput',
        'tip':'', # leave this out - it overrides the button tips if used: 'Do you want a Wing (lifting shape) or Strut (Symmetrical low-drag shape)',
        'desc':'*  Select upright to insert a lifting shape\n'+
                '    like a wing\n',
    #            '*  Select downwards to insert a downforce\n'+
    #            '    shape like a spoiler',
        'dflt':0, # first button is 0
        'buttons':[{'name':'Produces lift (e.g. wing)','res':os.path.join(resource_folder,'wing')},{'name':'Produces down force (e.g. spoiler)','res':os.path.join(resource_folder,'inverted')}],
        'html':'<center><img src=up_aft.png></center><br>*  Select downwards to insert a downforce<br>\n'+
                ' &nbsp;    shape like a spoiler<br><center><img src=down_aft.png></center>',
        'last':0},

        { 'id':'nose'+aft,
        'name':'Nose point',
        'type':'SelectionInput',
        'filter':['SketchPoints'],
        'tip':'Chose a point to set the front of your foil',
        'desc':'The Nose and Tail determine the size and direction of your foil.',
        'html':'<center><img src=create_wing_aft.png></center>',
        'crsr':'Select nose point', # The text in the tooltip shown next to the cursor.
        'lim':[1,1], # Required
        'last':0},

        { 'id':'tail'+aft,
        'name':'Tail point',
        'type':'SelectionInput',
        'filter':['SketchPoints'],
        'tip':'Chose a point to set the trailing end of your foil',
        'desc':'The Nose and Tail determine the size and direction of your foil.',
        'html':'<center><img src=create_wing_aft.png></center>',
        'crsr':'Select tail point',
        'lim':[1,1], # Required
        'last':0},

        { 'id':'chord_info'+aft,
        'name':'Chord',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'nose to tail=chord'+infob,
        'rows':1,
        'unit':'mm',
        'disabled':True,
        'last':0},

        { 'id':'join'+aft,
        'name':'Join point',
        'type':'SelectionInput',
        #'filter':['ConstructionPoints','ConstructionLines','SketchPoints','SketchLines','LinearEdges','JointOrigins'], 
        'filter':['SketchPoints','SketchLines','LinearEdges','JointOrigins'], 
        'tip':'(Optional, Recommended) Pick a point or \nline to locate where to insert your foil.',  # Future note: use nose..tail to work out line direction
        #'tip':'Pick a point or to locate foil.',  # Future note: use nose..tail to work out line direction
        #'desc':'The ideal Angle-of-Attack (AoA or alpha)\ndiffers depending on foil size and conditions.\nIf you select a join point, '+
        #       'your foil will be\ninserted and rotated by the ideal AoA relative\nto the direction indicated by the Nose to the\nTail line.\n\n'+
        #       "Tip: you can select the nose-to-tail line itself if\nyou like - and if you do this, the foil will be\nrotated about it's "+
        #       "central moment (usually\nabout 25\% in from the nose)\n\n"+
        #       "Tip: If you want to join at the Nose or Tail <b>point</b>, you first need to add a <i>second</i> point on top of it (Fusion360 doesn't let our add-in select the same point in 2 different selection inputs)" +
        'html':'The ideal Angle-of-Attack (AoA or alpha) differs depending on foil size and conditions.<br>\n' + 
               'If you select a join point, your foil will be inserted and rotated by the ideal AoA relative to the direction indicated by the Nose to the Tail line.<br><br>\n\n'+
                "Tip: you can select the nose-to-tail line itself if you like - and if you do this, the foil will be rotated about its central moment (usually about 25% in from the nose)<br><br>\n\n"+
                "Tip: If you want to join at the Nose or Tail <u><b>point</b></u>, you first need to add a <i>second</i> point on top of it (Fusion360 doesn't let our add-in select the same point in 2 different selection inputs)<br>\n" +
                '<center><img src=create_wing_aft.png></center>',
        #'html':'<center><img src=create_wing_aft.png></center>',
        'lim':[0,1], # Sets setSelectionLimits(min=0,max=1) to make this optional.
        'autosel':True, # Tells our UI code to auto-select this non-required input after the tail got chosen
        'crsr':'Select moment-center join point or line',
        'last':0},

        # hints for 'helpok'
        { 'id':'chord'+aft,
        'name':'Nose and Tail',
        'type':'TextBoxCommandInput',
        'tip':'Needed to work out where to insert the foil, which way it points, and the chord, which is needed to work out the Reynolds Number',   # for 'helpok'
        'disabled':True,
        'last':0},

        { 'id':'re'+aft,
        'name':'Reynolds Number',
        'type':'TextBoxCommandInput',
        'tip':'Reynolds number is automatically determined from the size (chord) and speed you select, along with the type of medium (air, water) and its properties (temperature/altitude/pressure etc)',   # for 'helpok'
        'disabled':True,
        'last':0},
 
        { 'id':'radius_nr'+aft,    # Not a UI element, but rather, a storage area to pass data from the UI to the create.
        'name':'Blade radius',
        'type':'TextBoxCommandInput',
        'tip':"The radius is automatically determined by the distance between the rotation axis and the blades' tip",   # for 'helpok'
        'disabled':True,
        'invisible':True, # Hides it
        'last':0},
 
        { 'id':'root_chord_nr'+aft,    # Not a UI element, but rather, a storage area to pass data from the UI to the create.
        'name':'Hub chord',
        'type':'TextBoxCommandInput',
        'tip':"The Hub chord is automatically determined from the length of your chosen rotation axis line",   # for 'helpok'
        'disabled':True,
        'invisible':True, # Hides it
        'last':0},
 
        { 'id':'tip_chord_nr'+aft,    # Not a UI element, but rather, a storage area to pass data from the UI to the create.
        'name':'Tip chord',
        'type':'TextBoxCommandInput',
        'tip':"The Tip chord is automatically determined - it is either 0, or, if you chose a tip-location that is a line - the length of that line.",   # for 'helpok'
        'disabled':True,
        'invisible':True, # Hides it
        'last':0},

        #----------------------------   Import / Export stuff

        #{ 'id':'import_export_but'+aft,
        #'name':'            ',                                                  # line up the box with the rest of the controls
        #'type':'RadioButtonGroupCommandInput',
        #'choices':['Import','Export'],
        #'desc': 'Import loads an airfoil file into your design and scales it to the size you select.  Export saves an existing airfoil shape to a Selig-format .dat file.',
        #'tip':'Import to load, export to save.',
        #'last':0},

        { 'id':'aoa'+aft,
        'name':'Attack Angle',
        'type':'ValueInput',
        'dflt':'0 deg',
        'unit':'deg',   # nb: returns to us in radians
        'desc':'Use 0 to import as-is.\n\nEnter an angle, and select a join-point, and your foil will be rotated about the join point by the angle you specify.',
        'tip':'Rotates the foil after import.',
        'last':0},

        { 'id':'spline_type'+aft,
        'name':'Spline type',
        'type':'ButtonRowCommandInput',
        'tip':'', # leave this out - it overrides the button tips if used: 'Do you want a Wing (lifting shape) or Strut (Symmetrical low-drag shape)',
        'desc':'*  Select fit-point to spline through each\n'+
               '    data point (recommended)\n',
    #          '*  Select cubic for control-point data\n'+ # See the HTML below
    #          '    (works best with very few data points)',
        'dflt':0, # first button is 0
        'buttons':[{'name':'Fit-point spline','res':os.path.join(resource_folder,'fitpoint')},{'name':'Cubic (degree 3) spline','res':os.path.join(resource_folder,'cubic')}],
        'html':'<center><img src=fit_point.png></center><br>*  Select cubic for control-point data<br>\n'+
                ' &nbsp;    (works best with very few data points)<br><center><img src=degree_3.png></center>',
        'last':0},

        { 'id':'interpolate'+aft,   # see also num_points and interpolate
        'name':'set # of points',
        'type':'ValueInput',
        'dflt':'0',
        'unit':'',
        #'invisible':True,
        'desc':'Enter 0 to import all coordinates, or, pick a number of points you want to have and your input data will be interpolated (expanded or reduced using splines) to suit',
        'tip':'(optional) Choose # of points to insert',
        'html':'<center><img src=point_density.png></center>',
        'last':0},

        #{ 'id':'wrap_radius'+aft,
        #'name':'Wrap radius',
        #'type':'ValueInput',
        #'dflt':'0',
        #'unit':'mm',
        ##'invisible':True,
        #'desc':'Enter a non-zero number to bend the foil after import.  Your foil will be wrapped around an imaginary cylinder of the radius you enter (Useful if your foil will be moving in a circle, like propellers and turbines).',
        #'tip':'(optional) Bend the imported foil around a cylinder.',
        #'last':0},

        { 'id':'wrap_axis'+aft,
        'name':'Wrap axis',
        'type':'SelectionInput',
        #'filter':['ConstructionLines','SketchLines','LinearEdges'], 
        'filter':['SketchLines','LinearEdges'], 
        'tip':'(optional) Rotational axis.  Pick a center line to indicate where this foil section will rotate around.',
        'desc':'If you select this optional axis, your foil will be '+
               'wrapped (bent in Z axis) around an imaginary '+
               'cylinder centered at this line (Useful if your '+
               'foil will be moving in a circle, like propellers '+
               'and turbines).\n\n'+
               'Hint: put the axis vertical, and the chord '+
               'horizontally on a different plane:',
        'html':'<center><img src=wrap_axis_aft.png></center>',
        'crsr':'Select rotation axis line',
        'lim':[0,1], # Optional
        'last':0},

        { 'id':'nofix'+aft,
        'name':'No fixups',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'We automatically perform extensive repairs on input airfoil coordinates, if necessary.  Check this box to prevent this feature.',
        'tip':'Check this box to disable data-repair.',
        'last':0},

        { 'id':'select_spline'+aft,
        'name':'Select Foil',
        'type':'SelectionInput',
        'filter':['SketchCurves'], # SketchCurves: Select any shape of sketch entity (need to manually prevent non-splines). See https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-03033DE6-AD8E-46B3-B4E6-DADA8D389E4E
        'allowed':{'adsk::fusion::SketchFittedSpline':1, 'adsk::fusion::SketchFixedSpline':1}, # So we can "reject" lines when we only want to let them pick splines...
        'tip':'Click on the airfoil curve you wish to export.\n   It MUST be a spline.', 
        'tip.foil_performance'+aft:'Click on the airfoil curve you wish to analyze.\n   It MUST be a spline.', 
        'desc':'Select the sketch spline of an airfoil shape.\nthe data points will be normalised upon writing',
        'desc.foil_performance'+aft:'Select the sketch spline of teh airfoil shape you wish to analyze.',
        'html':'<center><img src=create_wing_aft.png></center>',
        #'lim':[0,1], # Sets setSelectionLimits(min=0,max=1) to make this optional.
        'crsr':'Select an airfoil sketch spline to export',
        'crsr.foil_performance'+aft:'Select an airfoil sketch spline to analyze',
        'last':0},

        { 'id':'original'+aft,
        'name':'Original',
        'type':'BoolValueInput',
        'dflt':True,
        'desc':'If your selected foil was inserted by ' + program_name + ', this '+
               'option will output the original points that made up the shape.\n\n'+
               "Note that if you've modified the shape since it was created, those changes "+
               "will not be reflected in the output.",
        'tip':'Use the original master spline coordinates.'},

        { 'id':'normalize'+aft,
        'name':'Normalize',
        'type':'BoolValueInput',
        'dflt':True,
        'desc':'Select this option to:\n'+
               '* scale the exported foil into unit space (X\n\n'+
               '   ranges from 0 at the nose, to 1 at the tail)\n\n'+
               '* rotate and translate the foil, putting the\n'+
               '   nose and tail at Y=0',
        'tip':'Convert points putting nose on (0,0) and tail on (1,0).'},
        # TODO - graphic showing input and output

        { 'id':'threed'+aft,
        'name':'3D',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'This will output the full (X, Y, Z) for\n'+
               'each point around your foil circumference.',
        'tip':'Include the Z values in the output.'},

        { 'id':'unwrap'+aft,
        'name':'Unwrap',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':"If your selected foil is not a planar shape (e.g. it's from a propeller, and has been wrapped around the rotational axis cylinder), this option will first un-roll your foil before export.",
        'tip':'Flatten into a 2D shape.'},

        { 'id':'metadata'+aft,
        'name':'Metadata',
        'type':'BoolValueInput',
        'dflt':True,
        'desc':"If attributes for this foil exist, they're included in JSON format on line 1 of your output file\n\n"+
               "Attributes will include everything about how "+program_name+" created this foil in the first place (if it did).",
        'tip':'Include info about the foil and its creation.'},

        { 'id':'datcsv'+aft,       # Turbulence
        'name':'Format',
        'type':'RadioButtonGroupCommandInput',
        'choices':['.dat  SELIG format', '.csv  spreadhseet file'],
        'dflt':'.dat  SELIG format',
        'tip':"Which output file format to use - both are plain text."},

        { 'id':'test_spline'+aft,
        'name':'Test Foil',
        'type':'SelectionInput',
        'filter':['SketchCurves'], # SketchCurves: Select any shape of sketch entity (need to manually prevent non-splines). See https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-03033DE6-AD8E-46B3-B4E6-DADA8D389E4E
        'tip':'Click on the airfoil curve you wish to export.', 
        'desc':'Select the sketch spline of an airfoil shape.\nthe data points will be normalised upon writing',
        'html':'<center><img src=create_wing_aft.png></center>',
        #'lim':[0,1], # Sets setSelectionLimits(min=0,max=1) to make this optional.
        'crsr':'Select an airfoil sketch spline to export',
        'last':0},
        # file button?  https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-69478fef-f96f-4e7c-b5af-766301072042

        { 'id':'file_button'+aft,
        'name':'Select file',
        'type':'ButtonRowCommandInput',
        'tip':'', # leave this out - it overrides the button tips if used: 'Do you want a Wing (lifting shape) or Strut (Symmetrical low-drag shape)',
        'multi':True, # multi-select option needed here
        'desc':'Click the icon to enter the filename',
        'buttons':[{'name':'Select file','res':os.path.join(resource_folder,'openfile')}],
        'last':0},
        
        { 'id':'file_name'+aft,
        'name':'',
        'type':'TextBoxCommandInput',
        'tip':'The name of the file that you want to import from or export to',
        'dflt':infoa + nofiletxt + infob,   # See also: nofiletxt='(no file selected yet)'
        'rows':2,
        #'unit':'mm',
        'disabled':True,
        'last':0},

        #{ 'id':'import_group'+aft,
        #'invisible':True,
        #'name':'Import',
        #'type':'GroupCommandInput',
        #'tip':''},

        #{ 'id':'export_group'+aft,
        #'invisible':True,
        #'name':'Export',
        #'type':'GroupCommandInput',
        #'tip':''},

        #{ 'id':'file_group'+aft,
        #'invisible':True,
        #'name':'File',
        #'type':'GroupCommandInput',
        #'tip':''},


        { 'id':'soon_group'+aft,
        'name':'COMING SOON',
        'type':'GroupCommandInput',
        'tip':'This new feature is currently under development. It doe not yet work.',
        'desc':'This is a new addition coming soon to ' +program_name + ' and does not function in this version.\n\n' + 
              'Keep an eye out for updates! (when Updates exist, ' +program_name + ' menu will tell you at the top).',
        'last':0},

        {'id':'soon_info'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p>NOTE: This new feature is not yet operational.</p></div>', # 207 ok 208 ng  43ok 44ng
        'desc':'This is a new feature coming soon to ' +program_name + ' and does not currently function in this version.\n\n' + 
              'Keep an eye out for updates! (when Updates exist, ' +program_name + ' menu will tell you at the top).',
        'rows':2,
        'last':0},


        #----------------------------   Propeller stuff

        { 'id':'beta_group'+aft,
        'name':'BETA FEATURE',
        'type':'GroupCommandInput',
        'tip':'This new feature has limited functionality at present.',
        'desc':'This is a new addition to ' +program_name + ' and does not yet contain all the features that are planned for it.\n' + 
              'Limited functionality is however available right now.',
        'last':0},

        {'id':'alpha_info'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p><mark style="background-color: #FFB0B0">NOTE: This upcoming new feature IS NOT FUNCTIONAL at present.</mark></p></div>', # 207 ok 208 ng  43ok 44ng
        'tip':'This new feature has limited functionality at present.',
        'desc':'This is an upcoming new addition to ' +program_name + ' and is not functionl in this release.\n\n'+
               'An update which enables this feature will arrive soon.',
        'rows':2,
        'last':0},

        {'id':'beta_info'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'<div><p><mark style="background-color: #FFFF80">NOTE: This new feature has limited functionality at present.</mark></p></div>', # 207 ok 208 ng  43ok 44ng
        'tip':'This new feature has limited functionality at present.',
        'desc':'This is a new addition to ' +program_name + ' and does not yet contain all the features that are planned for it.\n' + 
              'Limited functionality is however available right now.',
        'rows':2,
        'last':0},

        { 'id':'select_outline'+aft,
        'name':'Blade outline',
        'type':'SelectionInput',
        'filter':['SketchCurves'], # SketchCurves: Select any shape of sketch entity (need to manually prevent non-splines). See https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-03033DE6-AD8E-46B3-B4E6-DADA8D389E4E
        'allowed':{'adsk::fusion::SketchFittedSpline':1, 'adsk::fusion::SketchFixedSpline':1}, # So we can "reject" lines when we only want to let them pick splines...
        'tip':'Click on an outline of the blade shape you want\n   It MUST be a spline.', 
        #'tip.foil_performance'+aft:'Click on the airfoil curve you wish to analyze.\n   It MUST be a spline.', 
        'desc':'Use a sketch spline to draw the outline of one blade you want.  This will be used to insert blade-section profiles of the chord-length your draw, located (X, Y, and Z) where you want.',
        #'desc.foil_performance'+aft:'Select the sketch spline of teh airfoil shape you wish to analyze.',
        #'html':'<center><img src=create_wing_aft.png></center>',
        #'lim':[0,1], # Sets setSelectionLimits(min=0,max=1) to make this optional.
        'crsr':'Select the sketch-spline of your blade outline',
        #'crsr.foil_performance'+aft:'Select an airfoil sketch spline to analyze',
        'last':0},

        { 'id':'strategy_group'+aft,                                            # We put the name in its own group, so the actual control doesn't look silly (Fusion UI can't do long names in radios very well)
        'name':'Creation Strategy',
        'type':'GroupCommandInput',
        'tip':'Define blade-shapes here.',
        'html':prophtml,
        'last':0},

        { 'id':'turbine_strategy'+aft,
        'name':'            ',                                                  # line up the box with the rest of the controls
        'type':'RadioButtonGroupCommandInput',
        'choices':['Free tips','Inside Duct','Custom'],                         # Caution - this text is hard-coded below, take care if you change it!
        #'desc': 'Hello desc',
        'html':prophtml,
        'tip':'Select a blade-shape strategy.',
        'last':0},

        {'id':'chord_formula'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'readonly':False, # This is an input box
        'dflt':'# replace this text with your Python formula for section cords, or select a strategy above', # 207 ok 208 ng  43ok 44ng
        #'desc': 'See html',
        'tip': 'Enter the blade-shape formula here.',
        'html':prophtml, # huge text - view above
        'rows':3,
        'last':0},

        {'id':'formula_instructions'+aft,
        'name':'', # full width
        'type':'TextBoxCommandInput',
        'dflt':'Examples at <a href="http://foils.pro/Airfoil_Tools/blade_strategy.html">formula instructions</a></p>',
        'readonly':True,
        'rows':1,
        'last':0},


        { 'id':'location_group'+aft,                                            #----------------------------   Location
        'name':'Location',
        'type':'GroupCommandInput',
        'last':0},

        { 'id':'conditions_group'+aft,                                          #----------------------------   Conditions
        'name':'Conditions',
        'type':'GroupCommandInput',
         'tip':'Some of these are optional.',
        'last':0},

        { 'id':'constraints_group'+aft,                                         #----------------------------   Constraints
        'name':'Constraints',
        'type':'GroupCommandInput',
         'tip':'Most of these are optional.',
        'desc':'Fill in constraints you have. Omitted and redundant values will be optimised or computed.',
        'last':0},


        { 'id':'power_group'+aft,                                               #----------------------------   Power
        'name':'Input motor power',
        'type':'GroupCommandInput',
        'expanded':False, # make them open this if they want to change stuff
        'desc':'Specify the range of power available from your motor, in Horsepower HP or Kilowatts kW etc.',
        'tip':'How powerful is your motor?'},

        { 'id':'powercruise'+aft,
        'name':'Cruise',
        'type':'ValueInput',
        'dflt':'0.0 W',
        'unit':'W', # Watts. Can also accept hp
        #'invisible':True,
        'desc':'What power does your motor output when operating in its most efficient condition?\n\n'+
               'Add the suffix hp if you want to use horsepower instead of Watts',
        'tip':'Power at peak motor efficiency.'},

        { 'id':'powermax'+aft,
        'name':'Max',
        'type':'ValueInput',
        'dflt':'0.0 W',
        'unit':'W',
        'desc':'What is the peak power output of your motor?\n\n'+
               'Add the suffix hp if you want to use horsepower instead of Watts',
        'tip':'Max power your motor can produce.'},


        { 'id':'torque_group'+aft,                                               #----------------------------   Torque
        'name':'Input motor torque',
        'type':'GroupCommandInput',
        'expanded':False,
        'desc':'Optional: provide the best and max torque data, and reduction ratio for your motor',
        'tip':'Specify the torque range of your motor.'},

        { 'id':'torquecruise'+aft,
        'name':'Cruise',
        'type':'ValueInput',
        'dflt':'0.0 N m',
        'unit':'N m', 
        'desc':'What torque does your motor output when operating in its most efficient condition?\n\n'+
               'Use "N m" for Newton-meters, or "lbforce ft" for foot-pounds etc.',
        'tip':'Torque at peak motor efficiency.'},

        { 'id':'torquecruiserpm'+aft,
        'name':'Cruise RPM',
        'type':'ValueInput',
        'dflt':'0.0 rpm',
        'unit':'rpm',
        'desc':'When your motor is outputting the above-entered torque, at what RPM is it spinning?',
        'tip':'RPM for the above torque.'},

        { 'id':'torquemax'+aft,
        'name':'Max',
        'type':'ValueInput',
        'dflt':'0.0 N m',
        'unit':'N m', 
        'desc':'What is the peak torque output of your motor?\n\n'+
               'Use "N m" for Newton-meters, or "lbforce ft" for foot-pounds etc.',
        'tip':'Max torque your motor can produce.'},

        { 'id':'torquemaxrpm'+aft,
        'name':'Max RPM',
        'type':'ValueInput',
        'dflt':'0.0 rpm',
        'unit':'rpm',
        'desc':'When your motor is outputting the above-entered Max torque, at what RPM is it spinning?',
        'tip':'RPM at max torque.'},

        { 'id':'reduction'+aft,
        'name':'Reduction',
        'type':'ValueInput',
        'type':'TextBoxCommandInput',
        'readonly':False, # This is an input box, and # Example of how to do 1-line arbitrary text input
        'dflt':'1 : 1',
        'rows':1,
        #'desc':'If gears or pulleys are fitted, enter the input:output ratio.\n\n'+ 'e.g. If your motor spins 3.5 times faster than your prop, use:-\n'+ '   3.5 : 1         -or-\n     7 : 2',
        'html':'<p>If gears or pulleys are fitted, enter the input:output ratio.  e.g. If your motor spins 3.5 times faster than your prop, use:'+
               '<table border=1 bgcolor=white ><tr><td style="color:black;"> 3.5 : 1 </td></tr></table> or <table border=1 bgcolor=white style="color:black;"><tr><td color=black> 7 : 2 </td></tr></table></p>',
        'tip':'Input:Output gearing ratio.'},

        { 'id':'thrust_group'+aft,                                              #----------------------------   Thrust
        'name':'Target Thrust',
        'type':'GroupCommandInput',
        'expanded':True,
        'html':'<p>Enter the target force, and we will calculate the necessary input power for you.</p>'+
               '<p>Tip: multiply by gravity (9.8) to convert Kg to N. e.g. for 100kg, enter: <table border=1 bgcolor=white ><tr><td style="color:black;"> 100kg * 9.8 N/kg </td></tr></table></p>',
        'tip':'How much force do you need?'},                          

        { 'id':'target_thrust'+aft,
        'name':'Thrust',
        'type':'ValueInput',
        'dflt':'0.0 N',
        'unit':'N',     # Caution - these units are * 100 internally...
        'html':'<p>Enter the target force, and we will calculate the necessary input power for you.</p>'+
               '<p>Tip: multiply by gravity (9.8) to convert Kg to N. e.g. for 100kg, enter: <table border=1 bgcolor=white ><tr><td style="color:black;"> 100kg * 9.8 N/kg </td></tr></table></p>',
        'tip':'How much force do you need?'},                          

        { 'id':'target_efficiency'+aft,
        'name':'Efficiency %',
        'type':'ValueInput',
        'dflt':'59',
        'unit':'', # 'percent' doesn't seem to work
        'desc':'Enter a percentage between 1 and 59% (the betz limit - anything above 59% is impossible, since the flow will just go around your turbine if you try)\n\n' +
              'Turbines slow down the flow of the medium as they extract energy - the amount of slowing depends on their efficiency, and this affects the ideal angle of the blades.',
        'tip':'How much of the available energy will you harvest?'},                          

        { 'id':'prop_velocity_group'+aft,                                       #----------------------------   Vehicle Speed
        'name':'Vehicle Speed',
        'type':'GroupCommandInput',
        'desc':'This is the speed of your vehicle through the Air/Water.  This is NOT the speed of the air/water through your propeller (which will be MUCH faster.  Use 0 for fans and other things that do not move.',
        'tip':'How fast do you want to go?'},


        { 'id':'spline_group'+aft,
        'name':'Spline Control Options',
        'type':'GroupCommandInput',
        'tip':'Change settings relating to how the out splines are built.',
        'expanded':False,
        'last':0},
        # Goes with 'spline_type', 'interpolate'/num_points, 'num_rails'

        { 'id':'num_points'+aft,   # see also num_points and interpolate
        'name':'set # of points',
        'type':'ValueInput',
        'dflt':'60',
        'unit':'',
        'desc':'Pick the number of control points you want to have in each of your spline sections',
        'html':'<center><img src=point_density.png></center>',
        'tip':'(optional) Choose how many control points to have',
        'last':0},

        { 'id':'nt_concentrate'+aft,
        'name':'end density',
        'type':'ValueInput',
        'dflt':'10',
        'unit':'',
        #'desc':'Pick the number of control points you want to have in each of your spline sections',
        'tip':'(optional) How much to concentrate extra points at the nose and tail (10=lots, 1=none)',
        'html':'<center><img src=point_density.png></center>',
        'invisible.import_foil'+aft:True,
        'last':0},

        { 'id':'cloud_foil'+aft,
        'name':'use cloud foils',
        'type':'BoolValueInput',
        'dflt':True,
        #'desc':'Sends your data to our server.',
        'tip':'Get latest airfoil profile and performance data using our online web service.'},

        { 'id':'num_rails'+aft,
        'name':'set # of rails',
        'type':'ValueInput',
        'dflt':'4',
        'unit':'',
        'desc':"4 rails would do the tail, nose, top middle, and bottom middle.  0 would do none.  Even numbers are best (odd will miss out the nose)\n\n"+
               "Note that 1 extra rail is automatically added if your trailing edge has a minimum thickness (one top edge rail and an added bottom edge rail)",
        'tip':'How mant rails do you wan to connect the sections?',
        'html':'<center><img src=num_rails.png></center>',
        'last':0},

        #----------------------------   Turbine stuff

        { 'id':'rotate_direction'+aft,
        'name':'Rotation',
        'type':'ButtonRowCommandInput',
        'tip':'', # leave this out - it overrides the button tips if used: 'Do you want a Wing (lifting shape) or Strut (Symmetrical low-drag shape)',
        'desc':'Which direction do you want it to rotate',
        'dflt':1, # first button is 0
        'buttons':[ {'name':'Left / Counter-clockwise','res':os.path.join(resource_folder,'propccw')}, {'name':'Right / Clockwise','res':os.path.join(resource_folder,'propcw')} ],
        'html':'<p><img src=propccw_aft.png align=left style="float:left;"><br> &nbsp; Left rotates counter-clockwise<br> &nbsp; when viewed from astern.<br></p><hr>'+
               '<p><img src=propcw_aft.png  align=left style="float:left;"><br> &nbsp; Right is clockwise rotation<br> &nbsp; when viewed from astern.</p><p></p>\n'+
               '<br><p>(astern means looking at the blades while we are in the flow that has gone through them)</p>\n',
        'last':0},

        # strategy_group (heading), 'turbine_strategy' (radio), 'chord_formula', 'formula_instructions;, - see prop definitions

        { 'id':'centerline'+aft,
        'name':'Rotation axis',
        'type':'SelectionInput',
        #'filter':['ConstructionLines','SketchLines','LinearEdges'], 
        'filter':['SketchLines','LinearEdges'], 
        'tip':"Rotational axis line.  Where the blades\nrotate around, and the hub's chord.",
        'desc':'Blade foil sections will be cylindrically wrapped around this center line relative to their distance away from it.\n\nThe length of your axis line will also be used for the starting chord of the blades.',
        'html':'<center><img src=prop_turbine_locators.png></center>',
        #'html':'<center><img src=turbine_centerline_aft.png></center>',
        'crsr':'Select rotation axis line',
        'lim':[1,1], # Required
        'last':0},

        { 'id':'hub_nose'+aft,
        'name':'Hub nose',
        'type':'SelectionInput',
        'filter':['SketchPoints'],
        'tip':'Chose a point to set the front of your turbine',
        'tip.create_propeller'+aft:'Chose a point to set the front of your propeller',
        'desc':'The Hub nose points into the flow direction (it is the front of the turbine, which the flow comes towards).',
        'desc.create_propeller'+aft:'The Hub nose points into the flow direction (it is the front of the propeller, which the flow comes towards).',
        #'html':'<center><img src=turbine_nose_aft.png></center>',
        'html':'<center><img src=prop_turbine_locators.png></center>',
        'html.create_propeller':'<center><img src=prop_turbine_locators.png></center>',
        'crsr':'Select hub nose point', # The text in the tooltip shown next to the cursor.
        'lim':[1,1], # Required
        'last':0},

        { 'id':'root_chord_info'+aft,
        'name':'Root chord',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(select rotation axis)'+infob,
        'rows':1,
        'unit':'mm',
        'disabled':True,
        'last':0},

        { 'id':'hub_edge'+aft,
        'name':'Hub edge',
        'type':'SelectionInput',
        #'filter':['ConstructionPoints','ConstructionLines','SketchPoints','SketchLines','LinearEdges','JointOrigins'], 
        'filter':['SketchPoints'],
        'tip':'Pick a point away from the center line to define hub radius.', 
        'desc':'The blade root airfoil will be wrapped around and connected at the cylinder formed by the center line and this point.',
        'html':'<center><img src=turbine_edge_aft.png></center>',
        'crsr':'Select hub edge point', # Future; accept circles too
        'lim':[1,1], # Required
        'last':0},

        { 'id':'tip_location'+aft,
        'name':'Tip location',
        'type':'SelectionInput',
        #'filter':['ConstructionLines','SketchLines','LinearEdges',], 
        #'filter':['ConstructionPoints','ConstructionLines','SketchPoints','SketchLines','LinearEdges','JointOrigins'], 
        'filter':['SketchPoints','SketchLines','LinearEdges','JointOrigins'], 
        #'tip':'Pick a point or line to locate turbine blade tips.',
        'tip':'Pick a point or line to indicate both the\nlocation and chord of the turbine blade tips.',
        'tip.create_propeller'+aft:'Pick a line to indicate both the\nlocation and chord of the propeller blade tips.',
        'desc':'This must be a point or line perpendicular to your rotation axis, and on a different Z plane.\n\n'+ # Note: If you specify a "Blade Length" or select "Optimal", this selection merely determines blade direction, not length nor chord.',
                "* The length of this line will be the chord\n   of the blade tips (or 0 if a point).\n"+
                "* The distance this line is away from the\n   axis determine the length of the blades\n\n"+
                "Hint: make a 3D sketch, put the rotation axis in the XY plane (e.g. horizontal), and this tip location in a different Z plan (e.g. vertical in the XY plane).",
        'html':'<center><img src=prop_turbine_locators.png></center>',
        #'html':'<center><img src=create_wing_aft.png></center>',
        'lim':[1,1], # Sets setSelectionLimits(min=0,max=1) to make this optional.
        'crsr':'Select blade-tip location point or line',
        'last':0},

        { 'id':'tip_chord_info'+aft,
        'name':'Tip chord',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(select tip location)'+infob,
        'rows':1,
        'unit':'mm',
        'disabled':True,
        'last':0},

        { 'id':'diameter_info'+aft,
        'name':'Blade diameter',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(computed)'+infob,
        'rows':1,
        'unit':'mm',
        'disabled':True,
        'last':0},

        { 'id':'tip_speed'+aft,
        'name':'Tip speed',
        'type':'ValueInput',
        'dflt':'0.0 mps',
        'unit':'mps',
        #'invisible':True,
        'desc':'Omit this to use optimal defaults, or, use this or RPM to set a desired rotational speed.\n\nAdd one of these suffixes if you want to use different units:'+
            '\n\n* mps   meters per second'+
            '\n\n* fps   foot per second'+
            '\n\n* mph   miles per hour'+
            '\n\n* knots',
        'tip':'Set the speed of the turbine tips'},

        { 'id':'rpm'+aft,
        'name':'RPM',
        'type':'ValueInput',
        'dflt':'0.0 rpm',
        'unit':'rpm',
        #'invisible':True,
        'desc':'Omit this to use optimal defaults, or, use this or Tip Speed to set a desired speed for the turbine blade tips.',
        'tip':'Set rotational speed of the turbine'},

        { 'id':'tip_mach_info'+aft,
        'name':'Tip Mach',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(computed)'+infob,
        'tip':"The Mach number of the blade tips is automatically determined from your speed, RPM, blades' diameter, altitude, and medium density",
        'disabled':True,
        'rows':1,
        'last':0},
    
        { 'id':'blade_len'+aft,
        'name':'Blade Length',
        'type':'ValueInput',
        'dflt':'0.0 m',
        'unit':'m',
        #'invisible':True,
        'desc':'Omit this to optimal defaults, or, Specify how long you want each blade to be (from centerline to tip)\n\n'+
            'Add any distance suffix if you want to use different units (e.g. mm, cm, m, micron, in, ft, yd, mi, nauticalMile, mil)',
        'tip':'How long is each blade?'},

        { 'id':'blade_count'+aft,
        'name':'Blade Count',
        'type':'ValueInput',
        'dflt':'0',
        'unit':'',
        #'invisible':True,
        'desc':'Omit this to optimal defaults, or, Specify how many blades you want',
        'tip':'How many blades?'},

        { 'id':'vary_pitch'+aft,
        'name':'Pitch variation',
        'type':'ValueInput',
        'dflt':'0 deg',
        'unit':'deg',
        'desc':'Use 0 for fixed-pitch.\n\nIf you have a variable pitch mechanism, enter the total number of degrees variation it supports (e.g. if it does -10 to +20 degrees, enter 30 deg)',
        'tip':'Variable pitch limits.'},

        { 'id':'minthick'+aft,
        'name':'Min thickness',
        'type':'ValueInput',
        'dflt':'0.0 mm',
        'unit':'mm',
        #'invisible':True,
        'desc':'Makes the trailing edge no thinner than this - useful when your manufacturing process cannot do sharp ends or thin walls.',
        'html':'<center><img src=min_thickness_aft.png></center>',
        'tip':'Set minimum thickness for the trailing edge.'},

        { 'id':'num_profiles'+aft,
        'name':'# of sections',
        'type':'ValueInput',
        'dflt':'12',
        'unit':'',
        #'invisible':True,
        'desc':'Profiles are arranged with more at the hub area (where twist and wrap are most pronounced) and fewer out to the tips (which change less).',
        'tip':'How many profile sections do you want?'},

        { 'id':'patchme'+aft,
        'name':'Patch it',
        'type':'BoolValueInput',
        'dflt':True,
        'tip':'Add surface-patches over the inserted sketch.',
        'desc':'This will patch the tip profile, patch the hub profile, surface-loft between them both through all the blade section profiles, and stich that all together to create a solid blade.',
        'last':0},

        #----------------------------   Presets

        { 'id':'preset'+aft,   # #Quick-insert (medium;Air(alt+temp)/Water(sea/fresh/depth/temp)/other m2/s, chord, thick, speed (m,km,knots,mph,foot,etc) ,conditions(rough+turbulent, smooth+laminar...)
        'name':'Preset',        # sea temp average = 16C Air temp =15C... set 16 for start?
        'type':'DropDownCommandInput',
        'choices':[#'Select a preset...',
                    'Your previous settings',
                    #'', # divider between customer presets, and built-ins   long range recon drone
    #V2                'Your foil: testfoil', # populate rest here -         your previous settings, Your last wing/your last strut/your last prop...
                    '', # divider between customer presets, and built-ins
                    'Small toy glider','Medium R/C plane', 'Fast jet toy', 'Ultralight aircraft', 'GA propeller plane', 'Business jet', 'Airliner', 'Fighter jet like', 'Slow hobby drone', 'Long range recon drone', 'Fast jet drone',
                    '', # divider between customer presets, and built-ins
                    'Fast car', 'Highway truck',
                    '', # divider between customer presets, and built-ins
                    'Hydrofoil surf board', 'Hydrofoil powerboat', 'Hydrofoil transporter',
                    '', # divider between customer presets, and built-ins
                    #'Wind Turbine', 'Model Turbine', 'Water Turbine', 
                    #'', # divider between customer presets, and built-ins
                    '(none) - fill in below'],
                    # Logically, the below should be merged with the above... but despite the duplication, doing it this way makes maintenance of these values easier...
        'values':{ 'Small toy glider'      :{'medium'+aft:0, 'turbulence'+aft:1, 'finish'+aft:0,  'minspeed'+aft:3,    'targetspeed'+aft:5,    'maxspeed'+aft:10,    'minaltitude'+aft:0, 'targetaltitude'+aft:20, 'maxaltitude'+aft:200,  'mintemperature'+aft:10, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:2, 'SKIP-target_mass'+aft:123 },
                    'Medium R/C plane'      :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:4,    'targetspeed'+aft:6,    'maxspeed'+aft:12,    'minaltitude'+aft:0, 'targetaltitude'+aft:20, 'maxaltitude'+aft:150,  'mintemperature'+aft:10, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:1, 'SKIP-target_mass'+aft:123 },
                    'Fast jet toy'          :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:5,    'targetspeed'+aft:27,   'maxspeed'+aft:40,    'minaltitude'+aft:0, 'targetaltitude'+aft:20, 'maxaltitude'+aft:200,  'mintemperature'+aft:10, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:0.5, 'SKIP-target_mass'+aft:123 },
                    'Ultralight aircraft'   :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:15,   'targetspeed'+aft:25,   'maxspeed'+aft:32,    'minaltitude'+aft:0, 'targetaltitude'+aft:300, 'maxaltitude'+aft:3000,  'mintemperature'+aft:5,  'targettemperature'+aft:12, 'maxtemperature'+aft:30,  'target_span'+aft:10, 'SKIP-target_mass'+aft:123, 'rpm'+aft: 2650 * radperstorpm }, 
                    'GA propeller plane'    :{'medium'+aft:0, 'turbulence'+aft:1, 'finish'+aft:2,  'minspeed'+aft:31,   'targetspeed'+aft:130,  'maxspeed'+aft:140,   'minaltitude'+aft:0, 'targetaltitude'+aft:8534, 'maxaltitude'+aft:9144,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:13, 'SKIP-target_mass'+aft:123 }, # Piper M600
                    'Business jet'          :{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:3,  'minspeed'+aft:54,   'targetspeed'+aft:272,  'maxspeed'+aft:276,   'minaltitude'+aft:0, 'targetaltitude'+aft:10668, 'maxaltitude'+aft:15544,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:17, 'SKIP-target_mass'+aft:123 }, # Cessna Citation XLS
                    'Airliner'              :{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:3,  'minspeed'+aft:66,   'targetspeed'+aft:252, 'maxspeed'+aft:258,    'minaltitude'+aft:0, 'targetaltitude'+aft:10668, 'maxaltitude'+aft:15544,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:123, 'SKIP-target_mass'+aft:123 }, # 737
                    'Fighter jet like'      :{'medium'+aft:0, 'turbulence'+aft:1, 'finish'+aft:2,  'minspeed'+aft:50,   'targetspeed'+aft:272, 'maxspeed'+aft:276,    'minaltitude'+aft:0, 'targetaltitude'+aft:12000, 'maxaltitude'+aft:18200,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:123, 'SKIP-target_mass'+aft:123 }, 
                    'Slow hobby drone'      :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:0,    'targetspeed'+aft:5,    'maxspeed'+aft:20,    'minaltitude'+aft:0, 'targetaltitude'+aft:50,    'maxaltitude'+aft:1000,  'mintemperature'+aft:10, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:123, 'SKIP-target_mass'+aft:123 },
                    'Long range recon drone':{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:3,  'minspeed'+aft:31,   'targetspeed'+aft:130,  'maxspeed'+aft:136,   'minaltitude'+aft:0, 'targetaltitude'+aft:10000, 'maxaltitude'+aft:16000,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:123, 'SKIP-target_mass'+aft:123 },
                    'Fast jet drone'        :{'medium'+aft:0, 'turbulence'+aft:1, 'finish'+aft:1,  'minspeed'+aft:50,   'targetspeed'+aft:272,  'maxspeed'+aft:276,   'minaltitude'+aft:0, 'targetaltitude'+aft:10000, 'maxaltitude'+aft:16000,  'mintemperature'+aft:16, 'targettemperature'+aft:16, 'maxtemperature'+aft:16,  'target_span'+aft:123, 'SKIP-target_mass'+aft:123 },

                    'Highway truck'         :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:25,   'targetspeed'+aft:27,   'maxspeed'+aft:30,    'minaltitude'+aft:0, 'targetaltitude'+aft:750, 'maxaltitude'+aft:2000,  'mintemperature'+aft:5, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:2.5, 'SKIP-target_mass'+aft:123 },
                    'Fast car'              :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:30,   'targetspeed'+aft:45,   'maxspeed'+aft:60,    'minaltitude'+aft:0, 'targetaltitude'+aft:750, 'maxaltitude'+aft:2000,  'mintemperature'+aft:5, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:1.8, 'SKIP-target_mass'+aft:123 },
                                                                                                                            #
                    'Hydrofoil surf board'  :{'medium'+aft:3, 'turbulence'+aft:1, 'finish'+aft:2,  'minspeed'+aft:4,    'targetspeed'+aft:7,    'maxspeed'+aft:15,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:0.8, 'SKIP-target_mass'+aft:123 },
                    'Hydrofoil powerboat'   :{'medium'+aft:1, 'turbulence'+aft:1, 'finish'+aft:1,  'minspeed'+aft:8,    'targetspeed'+aft:32,   'maxspeed'+aft:52,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:2, 'SKIP-target_mass'+aft:123 },
                    'Hydrofoil transporter' :{'medium'+aft:2, 'turbulence'+aft:0, 'finish'+aft:2,  'minspeed'+aft:6,    'targetspeed'+aft:17,   'maxspeed'+aft:20,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:5, 'SKIP-target_mass'+aft:123 },
                                                                                                                            #
                    #'Wind Turbine'          :{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:2,  'minspeed'+aft:6,    'targetspeed'+aft:17,   'maxspeed'+aft:20,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:5, 'SKIP-target_mass'+aft:123 },
                    #'Model Turbine'          :{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:2,  'minspeed'+aft:6,    'targetspeed'+aft:17,   'maxspeed'+aft:20,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:5, 'SKIP-target_mass'+aft:123 },
                    #'Water Turbine'          :{'medium'+aft:0, 'turbulence'+aft:0, 'finish'+aft:2,  'minspeed'+aft:6,    'targetspeed'+aft:17,   'maxspeed'+aft:20,    'minaltitude'+aft:0, 'targetaltitude'+aft:0, 'maxaltitude'+aft:0,  'mintemperature'+aft:5, 'targettemperature'+aft:12, 'maxtemperature'+aft:16,  'target_span'+aft:5, 'SKIP-target_mass'+aft:123 },

                    '(none) - fill in below':{} },
        'desc':'Select "Your previous settings" to use the most recent values you previously entered into this page',
        'tip':'pick settings for common scenarios'},


        #----------------------------   Medium
        #{ 'id':'medium_group'+aft,
        #'name':'Medium',
        #'type':'GroupCommandInput',
        #'tip':'Specify the medium your foil will be operating in'},

        { 'id':'medium'+aft,
        #'name':'            ',
        'name':'Medium',
        'type':'RadioButtonGroupCommandInput',
        'choices':['Air','Fresh Water','Sea Water','Any Water','Other'],
        'tip':'Specify the medium your foil will be operating in',
        'desc':'Your choice of medium determines the density, and the Reynolds and Mach numbers.',
        'last':0},

        { 'id':'density'+aft,          # Density
        'name':'Fluid density',
        'type':'StringValueInput',
        'disabled':True,
        'invisible':True,
        #'desc':'does this show when disabled?',
        'html':'kilograms per cubic metre (kg/m<sup>3</sup>)',
        'tip':'Density of the gas or fluid medium, in SI units'},

        { 'id':'density_info'+aft,
        'name':'',                      # Full width - this is an output display only
        'type':'TextBoxCommandInput',
        'dflt':infoa+'density and viscosity are computed'+infob,
        #'dflt':'<img border=0 align=left valign=top height=45 src="'+  os.path.join(abs_resource_folder,'banner_aft45.png') +'">' +
        'tip':'Density is automatically determined from your medium, temperature, and altitude selections',   # for 'helpok'
        'disabled':True,
        'rows':1,
        'last':0},

        { 'id':'dynamic_viscosity'+aft,
        'name':'Dynamic viscosity',    # Dynamic viscosity
        'type':'StringValueInput',
        'disabled':True,
        'invisible':True,
        #'desc':'does this show when disabled?',
        'html':'pascal-second (Pa.s), or equivalently kilogram per meter per second (kg.m<sup>-1</sup>.s<sup>-1</sup>).',
        'tip':'Dynamic viscosity of the gas or fluid medium, in SI units'},

        { 'id':'kinematic_viscosity'+aft,
        'name':'Kinematic viscosity',   # Kinematic viscosity
        'type':'StringValueInput',
        'disabled':True,
        'invisible':True,
        #'desc':'does this show when disabled?',
        'html':'square meters per second (m<sup>2</sup>/s)',
        'tip':'Kinematic viscosity of the gas or fluid medium, in SI units'},


        { 'id':'turbulence'+aft,       # Turbulence
        #'name':'Flow\nconditions',
        'name':'Flow',
        'type':'RadioButtonGroupCommandInput',
        'choices':['Laminar and clean', 'Average', 'Turbulent and dirty'],  # these combined with finish to make NCRIT
        'tip':"Tip: Turbulent is a better choice if you are not sure (Foils optimised for clean flows perform badly in dirty ones.)"},


        { 'id':'finish'+aft,
        'name':'Smoothness',            # Finish.  turb+finish mapping:  ncrit = 3*finish + turb-1  # Yields 3 to 14. no. use ncrit=3*finish + turb-1 +(turb==3&&finish<4)-(turb==1&&finish>1); # gives: f1t1=n3 f1t2=n4 f2t1=n5 f1t3=n6 f2t2=n7 f3t1=n8 f2t3=n9 f3t2=n10 f4t1=n11 f3t3=n12 f4t2=n13 f4t3=n14
        'type':'DropDownCommandInput',
        'choices':[#'Select a surface finish...', '',
                    'Very rough',
                    'Average',
                    'Clean',
                    'Very smooth'],
        'tip':'Will the material you make your foil from have a rough and uneven surface, or a clean and smooth one?'},

        { 'id':'re_info'+aft,
        'name':'Re Target',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(computed)'+infob,
        'tip':'The Reynolds Number is automatically determined from your medium, temperature, altitude and speed selections.',
        'desc':'This dimensionless value is one of the key ingredients used by our Computational-Fluid-Dynamics (CFD) solver, which is part of our particle-swarm genetic optimiser which finds ideal-performance foil shapes for you.\n\n',
        'rows':1,
        'disabled':True},

        { 'id':'mach_info'+aft,
        'name':'Mach',
        'type':'TextBoxCommandInput',
        'dflt':infoa+'(computed)'+infob,
        'tip':'Mach number is automatically determined from your speed, altitude, and medium density',   # for 'helpok'
        'disabled':True,
        'rows':1,
        'last':0},
      
        #----------------------------   Velocity
        { 'id':'velocity_group'+aft,
        'name':'Velocity Range',
        'type':'GroupCommandInput',
        'desc':'* The combination of fluid turbulence and foil surface finish combine to determine the "NCrit" - a factor which governs when flow-separation takes place, '+
            'and hence ultimate foil efficiency limits.  Picking these correctly makes a big difference to optimum performance.',
    #    'html':'HTML test...',
        'tip':'Specify the range of speeds you want to support, and an ideal target velocity (cruise speed)'},

    # UI makes FloatSliderCommandInput unusable right now (input boxes are too small to see the numbers, tab button crashes fusion)
    #            { 'id':'speedrange'+aft,
    #            'name':'Min .. Max',
    #            'type':'FloatSliderCommandInput',
    #            'unit':'mps',
    #            #'range':['0 mps','380 mps'],
    #            'range':[0,380],
    #          'tip':'How slow do you need this to still work at?'},

        { 'id':'minspeed'+aft,
        'name':'Min',
        'type':'ValueInput',
        'dflt':'0.0 mps',
        'unit':'mps',
        'invisible':True,
        'desc':'Add one of these suffixes if you want to use different units:'+
            '\n\n* mps   meters per second'+
            '\n\n* fps   foot per second'+
            '\n\n* mph   miles per hour'+
            '\n\n* knots',
        'tip':'How slow do you need this to still work at?'},

        { 'id':'targetspeed'+aft,
        'name':'Target',
        'type':'ValueInput',
        'dflt':'0.0 mps',
        'unit':'mps',
        'desc':'Add one of these suffixes if you want to use different units:'+
            '\n\n* mps   meters per second'+
            '\n\n* fps   foot per second'+
            '\n\n* mph   miles per hour'+
            '\n\n* knots',
        'tip':'What speed do you want to have optimal performance at?'},

        { 'id':'maxspeed'+aft,
        'name':'Max',
        'type':'ValueInput',
        'dflt':'0.0 mps',
        'unit':'mps',
        'invisible':True,
        'desc':'Add one of these suffixes if you want to use different units:'+
            '\n\n* mps   meters per second'+
            '\n\n* fps   foot per second'+
            '\n\n* mph   miles per hour'+
            '\n\n* knots',
        'tip':'How fast do you realistically need to go?'},


        #----------------------------   Altitude
        { 'id':'altitude_group'+aft,
        'name':'Altitude Range',
        'type':'GroupCommandInput',
        'tip':'Specify the range of heights or depths you want to support, and an ideal target (cruising altitude)',
        'desc':'* The combination of medium, temperature, altitude and speed combine to determine the "Reynolds Number" - a key factor used for understanding and optimising flow.\n\n'+
            'Picking these correctly is critical for accomplishing optimum performance.',
        'last':0},

    #    { 'id':'altituderange'+aft,
    #    'name':'Min .. Max',
    #    'type':'FloatSliderCommandInput',
    #    'unit':'m',
    #    #'range':['-11000 m','30000 m'],
    #    'range':[-11000,30000], # deepest ocean, Mariana Trench, is 10994, highest flight, sr-71, is 27432 ... hmm... but maybe we should allow for highly compressed or near vacuum stuff?
    #     'tip':'How low and high do you need this to still work at?'},

        { 'id':'minaltitude'+aft,
        'name':'Min',
        'type':'ValueInput',
        'dflt':'0.0 m',
        'unit':'m',
        'invisible':True,
        'desc':'Specify the range of heights or depths you want to support, and an ideal target (cruising altitude)\n\n'+
            'Add any distance suffix if you want to use different units (e.g. mm, cm, m, micron, in, ft, yd, mi, nauticalMile, mil)',
        'tip':'What is the lowest altitude or depth you need to support?'},

        { 'id':'targetaltitude'+aft,
        'name':'Target',
        'type':'ValueInput',
        'dflt':'0.0 m',
        'unit':'m',
        'desc':'Add any distance suffix if you want to use different units (e.g. mm, cm, m, micron, in, ft, yd, mi, nauticalMile, mil)',
        'tip':'What height or depth do you want to have optimal performance at?'},

        { 'id':'maxaltitude'+aft,
        'name':'Max',
        'type':'ValueInput',
        'dflt':'0.0 m',
        'unit':'m',
        'invisible':True,
        'desc':'Add any distance suffix if you want to use different units (e.g. mm, cm, m, micron, in, ft, yd, mi, nauticalMile, mil)',
        'tip':'What is the highest altitude or depth you need to support?'},


        #----------------------------   Temperature
        { 'id':'temperature_group'+aft,
        'name':'Temperature Offset Range',
        'type':'GroupCommandInput',
        'tip':'Specify the range of temperature conditions you need.',
        'desc':'Temperature affects density, which in turn affects the Reynolds number and thus performance envelope.',
        'last':0},


    #    { 'id':'temperaturerange'+aft,
    #    'name':'Min .. Max',
    #    'type':'FloatSliderCommandInput',
    #    'unit':'C',
    #    #'range':['-273 C','1500 C'],
    #    'range':[0,1773.15],
    #     'tip':'How cold and hot do you need your foil to manage?'},

        { 'id':'mintemperature'+aft,
        'name':'Min',
        'type':'ValueInput',
        'dflt':'16 C',
        'unit':'C',
        'invisible':True,
        'desc':'For sea level, use actual temperature. For high-altitude, this is an offset temperature\n\n'+
            'Add one of these suffixes if you want to use different units:'+
            '\n\n* C   Celsius'+
            '\n\n* F   Fahrenheit'+
            '\n\n* K   Kelvin'+
            '\n\n* R   Rankine',
        'tip':'What is the coldest temperature you need to support?'},

        { 'id':'targettemperature'+aft,
        'name':'Target',
        'type':'ValueInput',
        'dflt':'16 C',
        'unit':'C',
        'desc':'Add one of these suffixes if you want to use different units:'+
            '\n\n* C   Celsius'+
            '\n\n* F   Fahrenheit'+
            '\n\n* K   Kelvin'+
            '\n\n* R   Rankine',
        'tip':'What is the operating temperature of your\nfluid that you want to have optimal performance at?'},

        { 'id':'maxtemperature'+aft,
        'name':'Max',
        'type':'ValueInput',
        'dflt':'16 C',
        'unit':'C',
        'invisible':True,
        'desc':'Add one of these suffixes if you want to use different units:'+
            '\n\n* C   Celsius'+
            '\n\n* F   Fahrenheit'+
            '\n\n* K   Kelvin'+
            '\n\n* R   Rankine',
        'tip':'What is the hottest temperature you need to support?'},


        #----------------------------   Mass and Span option
        { 'id':'mass_and_span_group'+aft,
        'name':'Mass and Span option',
        'type':'GroupCommandInput',
        'tip':'Fill in these settings to auto-generate an ideal chord length for your project.'},

    #    { 'id':'mass_and_span_range'+aft,
    #    'name':'Mass Min .. Max',
    #    'type':'FloatSliderCommandInput',
    #    'unit':'kg',
    #    'range':[0,35000000],       # Saturn 5 - heaviest flying thing
    #     'tip':'What is the empty and fully loaded weight range your foil needs to support?'},

        { 'id':'min_mass'+aft,
        'name':'Min Mass',
        'type':'ValueInput',
        'dflt':'0.0 kg',
        'unit':'kg',
        'invisible':True,
        'html':'<img src=foilsep.png><br>(screen-shot showing how this works - coming here soon)',
        'desc':'Add any mass suffix if you want to use different units (e.g. g, kg, slug, lbmass, ouncemass, tonmass)',
        'tip':'What is the lightest weight your foil needs to support?'},

        { 'id':'target_mass'+aft,
        'name':'Target Mass',
        'type':'ValueInput',
        'dflt':'0.0 kg',
        'unit':'kg',
        'desc':'Tip: Use this option instead of the Nose and Tail points to let '+program_name+
            ' pick the ideal chord length to give your solution optimal performance.\n\n'+
            'Add any mass suffix if you want to use different units (e.g. g, kg, slug, lbmass, ouncemass, tonmass)',
        'tip':'At what weight do you want to have optimal performance?'},

        { 'id':'max_mass'+aft,
        'name':'Max Mass',
        'type':'ValueInput',
        'dflt':'0.0 kg',
        'unit':'kg',
        'invisible':True,
        'desc':'Add any mass suffix if you want to use different units (e.g. g, kg, slug, lbmass, ouncemass, tonmass)',
        'tip':'What is the maximum weight your foil needs to lift?'},

        { 'id':'target_span'+aft,
        'name':'Wingspan',
        'type':'ValueInput',
        'dflt':'0.0 m',
        'unit':'m',
        'desc':'Add any distance suffix if you want to use different units (e.g. mm, cm, m, micron, in, ft, yd, mi, nauticalMile, mil)',
        'tip':'What wingspan do you want? (longer usually performs better)'},

        #----------------------------   Settings
        { 'id':'settings_group'+aft,
        'name':'Settings',
        'type':'GroupCommandInput',
        'tip':''},

        { 'id':'constrain'+aft,
        'name':'Constrain',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'Your foil will be anchored to the join or nose point you selected.  If that point is parametric, and if you select this option, '+
            'your inserted foil will move properly if you change that point.\n\n'+
            'Tip: future '+program_name+' releases will include fully parametric foils: If you choose this option now, and if you give '+
            'your foil a nickname, your foils (in future) will re-compute and re-draw when:\n'+
            '* you change the chord (distance between nose and tail points) or\n'+
            '* you change any foil settings (e.g. the speed, in the user-parameters) or\n'+
            '* you install any update to '+program_name+' that includes new foil shapes that our solver has found to outperform the ones you inserted now.',
        'html':'<center><img align=left src=constrain_aft.png></center>',
        'tip':'Check this box to insert a fully-constrained shape.\n\nNOTE: Be patient - this takes 2+ minutes to insert.'},

        { 'id':'camber'+aft,
        'name':'Camber line',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':"This option inserts a camber line, plus extends "+
               "that line to make it easy for you to insert a "+
               "patch that can be used as a mold-parting-surface "+
               "if you're fabricating your shapes from pre-preg carbon fibre etc",
        'html':'<center><img align=center src=camber_line.png><br><br>'
               '<img align=center src=camber_detail.png></center>',
        'tip':'Add a camber line with mold-parting extensions.'},

        { 'id':'endcap'+aft,
        'name':'3d endcap',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'This creates ellipses from all the top-surface points to the corresponding under-surface points, through a mid-point that is 50% out from the edge.',
        'html':'<center><img src=round_side_aft.png></center>',
        'tip':'Insert rounded 3D "rib" end profiles suitable for lofting a tidy side to your foil.'+
            '\n\nNOTE: Be patient - this takes 1 minute to insert (or 5+ minutes if you also constrain).'},

        { 'id':'enddist'+aft,
        'name':'End distance %',
        'type':'ValueInput',
        'dflt':'100',
        'unit':'',
        'invisible':True,
        'html':'<table border=0>'+
            '<tr><td>100% is circular</td><td><img src=round_close.png></td></tr>'+
    #           '<tr><td colspan=2><img src=round_side.png></td></tr>'+
            '<tr><td>10% would be flat</td><td><img src=flat_close.png></td></tr>'+
    #           '<tr><td colspan=2><img src=flat_side.png></td></tr>'+
            '<tr><td>800% would be very pointy</td><td><img src=pointy_close.png></td></tr>'+
    #           '<tr><td colspan=2><img src=pointy_side.png></td></tr>',
            '</table>',
        'tip':'How rounded do you want the end?'},

        { 'id':'share'+aft,
        'name':'Share',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'Do not select this on test or junk foils, or on confidential projects.\n\n'+
            'This option sends us a copy of your settings from this panel.  We use this data to guide our solver so future '+program_name+
            ' updates will include foil shapes more perfectly suited to the requirements of you and our other users.\n\n'+
            'Optional: fill in the min+max settings above to help us plan our solver performance envelope settings.',
        'tip':'Share your foil settings with us.'},

        { 'id':'helpok'+aft,
        'name':'Help',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'Is the OK button disabled?  Something important must be missing.\n\nClick this checkbox to see what that is.',
        'tip':"Click to see what's missing before the OK lights up"},

        #----------------------------   Save
        { 'id':'save_preset_group'+aft,
        'name':'Save into presets',
        'type':'GroupCommandInput',
        'expanded':False,
        'desc':'To erase a saved entry, first select it in the presets, then click Remove', # 
         'tip':'Saves all your above entries into the preset dropdown.'},

        { 'id':'save'+aft,
        'name':'Save',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'Tip: remember to enter a Nickname to save your settings as.\n\nTo update an already-saved preset, make the changes then simply save it again.',
         'tip':'Check this box to save your settings into the Preset dropdown.'},

        { 'id':'remove'+aft,
        'name':'Remove',
        'type':'BoolValueInput',
        'dflt':False,
        'desc':'Tip: this only works for ones you saved earlier, not for our built-in presets.',
         'tip':'Permanently remove the selected preset from the dropdown.'},


        { 'id':'rating_group'+aft,                                          #----------------------------   Rating Group
        'name':'Rate '+program_name,
        'type':'GroupCommandInput',
         'tip':'Some of these are optional.',
        'last':0},



        { 'id':'rate_me'+aft,
        'name':'', # Rate Airfoil Tools (from the group heading) --- no name here == full-width
        'type':'TextBoxCommandInput',
        'dflt':'<center><font size=+1><b>Please rate us</b> <a href="'+store_url+'"><img border=0 src="'+  os.path.join(abs_resource_folder,'stars5.png') +'"></a></center><br>' + # <img align=left src= stars5.png* stars5s.png*></center>
               '</font>Say thanks for this <b><mark style="background-color: #FFFF80">free</mark> tool</b> with your <a href="'+store_url+'">five star rating</a>. &nbsp;' + #3 Show you<b> &nbsp; &nbsp; &nbsp; &nbsp; <font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
               '<font size=-1><br>Not worth <img border=0 src="'+  os.path.join(abs_resource_folder,'stars5s.png') +'"> ? '+
               'Tell us how to change: <a href="'+fivestar_url+'">We want your 5 stars</a>!</font>',  # &nbsp; &nbsp; &nbsp; &nbsp; <font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
        'rows':6,

# Mac and Windows behave differently (url() does not work on windows, tables are broken, etc)
    #works    'dflt':'<div align="center"><img src="'+  os.path.dirname(os.path.realpath(__file__)) + "/resources/" + 'banner.png">'+program_name+' v<b>'+str(VERSION)+'<br><img src=resources/banner.png></b><a href="http:fusion360.autodesk.com">html.</a></div>',
    #works    'dflt':'<div align="center"><img src="'+  abs_resource_folder + '/banner.png">'+program_name+' v<b>'+str(VERSION)+'</b><a href="http:fusion360.autodesk.com">html.</a></div>',
    #OK!    'dflt':'<div align="center" style="overflow: visible;"><img src="'+  abs_resource_folder + '/banner.png" width=207 border=0 height=43></div>', # 207 ok 208 ng  43ok 44ng
#        'dflt':'<table border=0 cellpadding=0 cellspacing=0><tr><td rowspan=2><img src="'+  os.path.join(abs_resource_folder,'banner_aft.png') +'"></td><td align=right>' +
#               '<b><font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
#               '<br><font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
#               '<br><a href="http://foils.pro/Airfoil_Tools/">instructions</a> &nbsp; <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a>'+
#               '</td></tr><td height=1px style="height:1px" bgcolor="#0091e2"></td></tr></table>', # 207 ok 208 ng  43ok 44ng
#                '<div align="right" style="overflow: visible; background-image: url('+  os.path.join(abs_resource_folder,'banner.png') +'); background-repeat: no-repeat;"><b>'+
#                '<b><font size=+1>'+program_name+'</font></b>&nbsp;&nbsp;&nbsp;'+
#                '<br><font size=-1>v{:1.8f}'.format(VERSION) + ' ' + foildb2020.VERSION(VERSION) + '</font>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;'+
#                '<br><a href="http://foils.pro/Airfoil_Tools/">instructions</a> &nbsp; <a href="https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985">forum</a>'+
#                '</div>', # 207 ok 208 ng  43ok 44ng
        'last':0},



    #            { 'id':'test0'+aft,
    #            'name':'', # omit for full-wid
    #            'type':'TextBoxCommandInput',
    #            'dflt':"<html>hello <b>there</b> <input type=text name=test1 value='test1'>  <input type='submit' name='foo' value='foo' onclick='sendInfoToFusion()'>Click TO crash</input> ",
    #            'desc':'html testing desc',
    #            'html':'<img src=foilsep.png><br>this is a toolClip <b>HTLM</b> test string',
    #             'tip':'html tip test.'},

    #V2    { 'id':'test1'+aft,
    #V2    'name':'', # omit for full-wid
    #V2    'type':'ImageCommandInput',
    #V2    'dflt':resource_folder+'//foilsep.png'},


    # future featurecreep...
    #    { 'id':'log'+aft,
    #    'name':'Show log',
    #    'type':'BoolValueInput',
    #    'dflt':True,
    #    #'dflt':debug, # Default is false for normal users, true for me
    #     'tip':'Check this box to view the runtime log of ' + program_name},

        {'type':'end'}  # add above here
        ]



    #'Then specify your medium (air, water, etc), velocity range, and operating ' +
    #'conditions to insert the optimum foil shape for your purpose.\n\n'+
    #'Tip: If you select only the Nose or Tail (not both), you can then specify ' +
    #'a span and load or weight instead, and ' + program_name + ' will work out ' +
    #'the best chord length for you.\n\n' + 
    #'Tip: A good Join Point is usually 25% in from the nose, which is very close ' +


    #################################################################################
    ####                                                                          ###
    ##        Button, DropDown (slide-out), and dialogue-controls Expanders.       ##
    ####                                                                          ###
    #################################################################################

    # Populate uiel_order and uiel with additional keys that we need later
    uiel_order=[None]*(len(uiel))
    for i in uiel:                                                                  # Make a list, and put all our UI ID's into it in the correct order
        uiel_order[uiel[i]['ordr']]=i
        uiel[i]['id']=i                                                             # Store our id name in ourselves
        myres=resource_folder                                                       # use default if it doesn't have its own - See also aft_make_ui: per-toolbar ones override these
        if os.path.isdir(os.path.join(prog_folder, resource_folder, i)):             # Every control etc has its own set of icons, help, and Clip images
            myres=os.path.join(resource_folder, i)
            #eprint(myres)
        if uiel[i].get('res',0):                                                    # res - if it exists already, use it to create many sub res items...
            for subres in uiel[i]['res']:                                           # For every submenu... ('SolidCreatePanel',)
                useres=os.path.join(myres, subres)                                  # e.g. uiel[i]['res.SketchCreatePanel'] = './resources/create_wing_aft/SketchCreatePanel'
                abs_useres=os.path.join(prog_folder, useres)                        # clips cannot be relative :-()
                if os.path.isdir( abs_useres ):
                    uiel[i]['res.'+subres]=useres
                    uiel[i]['abs_res.'+subres]=os.path.join(prog_folder, useres)
                    #eprint("use:"+useres)
        uiel[i]['res']=myres                                                        # replace the temp list above with the default resource folder location
        uiel[i]['abs_res']=os.path.join(prog_folder, myres)
        (uiel[i]['ctrl'], uiel[i]['cmd'], uiel[i]['def'], uiel[i]['cb'])=({},{},{},{})  # Where we hold all our command objects; indexed by parentID (mywhere)
        if not uiel[i].get('name',0):
            uiel[i]['name']=i                                                       # Give separators a name, so debug output works OK
    if aft: # create aliases from ID-name (with prefix, to prevent toolbar naming collisions) back to base name (no prefix)
        for i in uiel_order:
            uiel[aft+i]=uiel[i]


    # Auto-populate anything missing from cmdl, to avoid key errors later
    for ctrl in cmdl:
        ctrl['abs_res']=abs_resource_folder
        for defaultible in ('dflt', 'tip', 'desc' ):                                # Things that must exist
            if ctrl.get(defaultible,'!!') =='!!': # some defaults can be zero
                ctrl[defaultible]=''
        if ctrl['type'] == 'TextBoxCommandInput' and not ctrl.get('rows',0):
            ctrl['rows']=3                                                          # Default to 3-rows for text lines
        if ctrl.get('id',0):
            cmdh[ctrl['id']]=ctrl                                                   # Create a dict pointing into the list


def bytes_default(self, o):
   try:
       iterable = iter(o)
   except TypeError:
       pass
   else:
       return list(iterable)
   return json.JSONEncoder.default(self, o) # Let the base class default method raise the TypeError

def json_dumps(j): # use the same JSON format for everything
    return json.dumps(j,skipkeys=True,sort_keys=True,ensure_ascii=False, default=lambda o: '<not serializable>') # ,default=bytes_default)

def json_loads(s): # use the same JSON format for everything
    return json.loads(s) # json.loads(s[, encoding[, cls[, object_hook[, parse_float[, parse_int[, parse_constant[, object_pairs_hook[, **kw]]]]]]]])

def remember(sharejson,pnlid):
    global VERSION, pnl, userdata, settings_file
    panel=pnlid['mywhere'] + '_' + pnlid['aftid'] # 'sketch_airfoil_command_aft' + '_' + 'menuid':'create_cowling_aft'
    dlg=pnl[panel]
    if dlg and not sharejson.get('now',0): shareDlg(sharejson,dlg) # Copy user-entered data out of the dialoge form controls into the sharejson dict (if not already done this)
    # Save in a variety of places so we can later be more intelligent about which settings to use for defaults and "previous" when users ask.
    userdata['last']=json_dumps(sharejson) # remember their most recent settings, so we can offer then again in future
    userdata['last.'+panel]=json_dumps(sharejson) # last.sketch_airfoil_command_aft_create_wing_aft
    userdata['last.'+pnlid['aftid']]=json_dumps(sharejson) # last.create_wing_aft
    userdata['last.'+pnlid['mywhere']]=json_dumps(sharejson) # last.sketch_airfoil_command_aft
    save_user_settings(None,settings_file) 



#################################################################################
####                                                                          ###
##                              Dialogue builder                               ##
####                                                                          ###
#################################################################################
#def aft_make_dlg(ui,aftid,mywhere,parentwhere,pctrl):                          # Build the popup dialogue box
def aft_make_dlg(cmd,pnlid):                                                    # Gets called from MyCommandCreatedEventHandler #cmd = args.command
    global cmdl, uiel, pnl, mru, userdata, NEXTVERSION

    dlg = cmd.commandInputs
    app = adsk.core.Application.get()
    ui  = app.userInterface
    panel=cmd.parentCommandDefinition.id

    #meta=dlg.addStringValueInput('meta'+aft,'Internal Settings', json_dumps( {'VERSION':VERSION, 'aft':aft, 'mywhere':self.mywhere, 'parentwhere':self.parentwhere, 'altid':self.altid} )) # In case later executions needs to know where this data came from...
    meta=dlg.addStringValueInput('meta'+aft,'Internal Settings', json_dumps( {'VERSION':VERSION, 'aft':aft} )) # In case later executions needs to know where this data came from...
    meta.isVisible=False
    panelname=dlg.addStringValueInput('panel'+aft,'Internal name of this dialog popup', pnlid['mywhere'] + '_' + pnlid['aftid'])    # So event handlers know what this data is (should we later remove pnlid from them)
    panelname.isVisible=False   # Added before I decided if I need this, so, might not be used...

    mydlg=re.search(pnlid['mywhere']+'_(.*)',panel)             # Get the name of the dialogue (e.g. "create_turbine_aft" out of "solid_airfoil_command_aft_create_turbine_aft" )
    if mydlg: mydlg=mydlg.group(1)
    else: mydlg=''

    if updatedata.get('loaded',0) and updatedata.get('current_version'+aft,0):
        NEXTVERSION='{:1.8f}'.format(updatedata['current_version'+aft])
        eprint("most recent version is {}".format(NEXTVERSION))
        if not pnl.get('toldversion',0):
            vlog("most recent version is {}".format(NEXTVERSION))
            pnl['toldversion']=True
    else:
        msg=updatedata.get('error',0)
        eprint('{}: update check problem: {}'.format(program_name,msg))
        if not pnl.get('toldversion',0):
            vlog('{}: update check problem: {}'.format(program_name,msg))
            pnl['toldversion']=True

    # Build a list of the UI elements we want to have in this menu option
    docontrols={} # which controls we want to include
    ctrllist=[]   # controls order
    for ctrlname in uiel[pnlid['aftid']].get('controls',[]):
        if ctrlname:
            docontrols[ctrlname + aft]=True  # use the ones defined in the menu if it has them
            if cmdh.get(ctrlname+aft,0):
                ctrllist.append(cmdh[ctrlname+aft])
    if not len(docontrols):
        for ctrl in cmdl:
            if ctrl.get('id',0):
                docontrols[ctrl['id']]=True  # create a default set of all controls
                ctrllist.append(ctrl)


    mygroup=dlg
    if 1:   # Go through all possible input elements, and put the right ones for this use case into the dialogue
        #for ctrl in cmdl:
        for ctrl in ctrllist:
            if docontrols.get(ctrl.get('id',0),False): # is it one we want to include?
                # eprint("Doing " + ctrl['name'] + " of type "+ ctrl['type'])
                ctrlmethod=getattr(mygroup,'add'+ctrl['type']) # NB; GroupCommandInput differs  # all control methods start with "add"
                ctrlinstance=None

                if ctrl['type'] == 'StringValueInput':
                    dflt=userdata.get(ctrl['id'],None)                          # If the userdata JSON has a value for this ID, it will be the default
                    if dflt is None: dflt=ctrl.get('dflt',None)                 # Use any code-preferred default if not
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],dflt) # (id, name, initialValue) # initialValue is optional
                    #ctrlinstance.isValueError = True # make it red

                elif ctrl['type'] == 'ValueInput':
                    dflt=userdata.get(ctrl['id'],None)                          # If the userdata JSON has a value for this ID, it will be the default
                    if dflt is None: dflt=ctrl.get('dflt',None)                 # Use any code-preferred default if not
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],ctrl['unit'],adsk.core.ValueInput.createByString(dflt)) # (id, name, unitType, initialValue) # initialValue must be float (in base units) or suitable string
                    eprint("set {} default to {}".format(ctrl['id'],dflt))

                elif ctrl['type'] == 'SelectionInput':
                    tooltip=ctrl['crsr']
                    if ctrl.get('crsr.'+mydlg,''):                               # Use the specific text for just this panel, if it exists
                        tooltip=ctrl['crsr.'+mydlg]
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],tooltip) # (id, name, commandPrompt) # commandPrompt is the text in the tooltip shown next to the cursor.
                    if ctrl.get('lim',0):
                        ctrlinstance.setSelectionLimits(ctrl['lim'][0],ctrl['lim'][1]) # (minimum, maximum) if 0, makes this optional
                    if ctrl.get('filter',0):
                        for filt in ctrl['filter']:
                            ctrlinstance.addSelectionFilter(filt)               # ['ConstructionPoints','ConstructionLines','SketchPoints','SketchLines','LinearEdges','JointOrigins'] usually.

                elif ctrl['type'] == 'FloatSliderCommandInput':
                    ctrlinstance=ctrlmethod( ctrl['id'], ctrl['name'], ctrl['unit'], ctrl['range'][0] , ctrl['range'][1] , True) # (id, name, unitType, min, max, hasTwoSliders) # hasTwoSliders is optional
                    #ctrlinstance=ctrlmethod( ctrl['id'], ctrl['name'], ctrl['unit'], adsk.core.ValueInput.createByString( ctrl['range'][0] ), adsk.core.ValueInput.createByString( ctrl['range'][1] ), True) # Broken - too tired to work out why...

                elif ctrl['type'] == 'ButtonRowCommandInput':
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],ctrl.get('multi',False))      # (id, name, isMultiSelectEnabled)
                    didx=0
                    for but in ctrl['buttons']:
                        try:
                            ctrlinstance.listItems.add(but['name'],(ctrl.get('dflt','') !='' and ctrl.get('dflt',-1) == didx),but['res'])    # Caution - crashes dialog if these are missing
                        except:
                            eprint('panel button {} create fail'.format(but['name']))
                        didx=didx+1

                elif ctrl['type'] == 'RadioButtonGroupCommandInput':
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'])            # (id, name) # name is optional
                    radio=ctrlinstance.listItems
                    dflt=userdata.get(ctrl['id'],None)                          # If the userdata JSON has a value for this ID, it will be the default
                    if dflt is None: dflt=ctrl.get('dflt',None)                 # Use any code-preferred default if not
                    for radioitem in ctrl['choices']:
                        radio.add(radioitem, False)
                        if dflt is not None :                                   # Is there a default?
                            if radioitem==dflt:                                 # Are we inserting the default right now?
                                radio[-1].isSelected=True                       # auto-select it

                elif ctrl['type'] == 'DropDownCommandInput':
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],adsk.core.DropDownStyles.TextListDropDownStyle) # (id, name, dropDownStyle); CheckBoxDropDownStyle/CheckBoxDropDownStyle/TextListDropDownStyle
                    radio=ctrlinstance.listItems
                    for radioitem in ctrl['choices']:
                        radio.add(radioitem, False, 'foo')

                elif ctrl['type'] == 'BoolValueInput':
                    dflt=userdata.get(ctrl['id'],None)                          # If the userdata JSON has a value for this ID, it will be the default
                    if not dflt and ctrl.get('dflt',0): dflt=ctrl.get('dflt',0)
                    if dflt:
                        ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],True,'',dflt) # id, name, isCheckBox, resourceFolder, initialValue
                    else:
                        ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],True)   # id, name, isCheckBox, resourceFolder, initialValue

                elif ctrl['type'] == 'GroupCommandInput':                       # Future - mod the name of this to include "Air" or "Water" etc after they've picked one
                    #ui.messageBox( "thinking about " + ctrl['name'] + ":" + ctrl['type'] + " in " + self.aftid + " of " + self.mywhere)
                    ctrlmethod=getattr(dlg,'add'+ctrl['type'])                  # put all groups in the parent
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'])            # (id, name)
                    ctrlinstance.isExpanded = ctrl.get('expanded',True)         # Defaults to true if expanded isn't set to false
                    ctrlinstance.isEnabledCheckBoxDisplayed = False
                    mygroup = ctrlinstance.children
                    if userdata.get('hide_rating'+aft,0) and ctrl['id'] == 'rating_group'+aft:
                        ctrlinstance.isVisible=False

                elif ctrl['type'] == 'TextBoxCommandInput':                     # Can do full-width HTML (omit name to get full width)
                    text=ctrl['dflt']
                    isReadOnly=ctrl.get('readonly',True)                        # Defaults to true if readonly isn't set to false
                    if ctrl['id']=='banner'+aft and updatedata.get('update_msg',0):
                        text=updatedata['update_msg']                           # Tell user of any important updates when they open our panel
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],text,ctrl['rows'],isReadOnly) # (id, name, formattedText, numRows, isReadOnly)

                elif ctrl['type'] == 'ImageCommandInput':                       # Can do full-width HTML (omit name to get full width)
                    ctrlinstance=ctrlmethod(ctrl['id'],ctrl['name'],ctrl['dflt']) # (id, name, imageFile) # Example uses './/Resources//Fusion360Logo.png'


                if ctrlinstance:
                    if ctrl.get('tip.'+mydlg,''):                               # Use the specific text for just this panel, if it exists
                        ctrlinstance.tooltip=ctrl['tip.'+mydlg]                 #"base tooltip string. This is always shown for commands."
                    elif ctrl['tip']:                                           # otherwise just use the default
                        ctrlinstance.tooltip=ctrl['tip']                        
                    if ctrl.get('desc.'+mydlg,''):
                        ctrlinstance.tooltipDescription=ctrl['desc.'+mydlg]     #"If the tooltip description and/or tool clip are also specified then the tooltip will progressively display more information as the user hovers the mouse over the control."
                    elif ctrl['desc']:
                        ctrlinstance.tooltipDescription=ctrl['desc']
                    if ctrl.get('disabled',0):
                        ctrlinstance.isEnabled=False

                    if ctrl.get('invisible.'+mydlg,0):                          # Visibility can be set for specific panels here - e.g. 'invisible.import_foil'+aft=True
                        ctrlinstance.isVisible=False
                    elif ctrl.get('invisible',0):
                        ctrlinstance.isVisible=False

                    try:                                                        # keep a record of what we made as we make it.
                        if not pnl.get(panel,0):
                            (mru['mruno'], pnl[panel])=( mru['mruno']+1, {'mru':mru['mruno']})      # Create pnl[panel], and populate it with anyting for now - maybe mru will be handy later...
                        pnl[panel][ctrl['id']]=ctrlinstance                     # So our input handlers can easily find stuff...
                    except:
                        eprint("add_and_remember failed")

                    if 1:
                        try:
                            clipfn=make_html(pnlid['mywhere']+"."+pnlid['aftid']+"."+ctrl['id'],ctrl,mydlg) # abs_resource_folder + 'airfoil_toolbar_button_aft.create_duct.nickname.html' 
                            if clipfn:
                                #eprint(clipfn)
                                ctrlinstance.toolClipFilename=clipfn
                        except:
                            eprint('ctrlinstance.toolClipFilename failed: {}'.format(traceback.format_exc()))

    return panel
        # We want the errors - hiding them is silly...
        #        except:
        #            if self.ui:
        #                self.ui.messageBox('Panel command created failed: {}'.format(traceback.format_exc()))





# 2.22, 2.32 in https://40p6zu91z1c3x7lz71846qd1-wpengine.netdna-ssl.com/wp-content/uploads/2017/02/Engineering-Fluid-Mechanics-9th-Edition-Crowe-Solution-Manual.pdf


#################################################################################
####                                                                          ###
##                           Validate - O - Rama                               ##
####                                                                          ###
#################################################################################

def MyValidate(pnlid,args,inputs,caller):
    global mru,pnl,userdata,textpalette,debug,updatedata,settings_file,VERSION,infoa,infob
    try:
        sharejson={} # in case of errors
        app = adsk.core.Application.get()
        ui = app.userInterface
        web_nbreq(None) # proccess any non-blocking operations

        dist=0

        panel=pnlid['mywhere'] + '_' + pnlid['aftid'] # 'sketch_airfoil_command'+'_aft' + '_' + 'menuid':'create_cowling'+'_aft'
        thisctrl=''
        if caller=='changed': # chgargs:
            (mru['mruno'], mru['mru'], mru[args.input.id])=( mru['mruno']+1, args.input.id, mru['mruno'])   # Update most-recently-used counter and pointers
            thisctrl=args.input.id  # e.g. nose_aft - Also has acccess to command.parentCommandDefinition.id  (same as pnlid mywhere+aftid)
        else:
            pass

        if not pnl.get(panel,0): # this line errors if an open panel still exists when this runs (e.g. debug re-start without closing the panel first)
            err(sharejson,"Please stop and re-start the "+program_name+"add-in.\n\n(Internal error encountered - menu missing from the UI - code was probably re-started without closing a panel first)") # signal error to caller and # pops up a messageBox
            args.areInputsValid = False
            return False 
        dlg=pnl[panel]

        eprint("MyValidate(" + panel + '  c=' + thisctrl + ") caller="+caller)

        #valid=0
        invalid=dict(uiel[pnlid['aftid']].get('required',{}))   # dict() makes a copy: these get removed as we do them - when len(invalid)==0 we know we are good to go! # 'targetaltitude':1,

        eprint("p={} Want: {}".format(pnlid['aftid'],invalid)) # VALCHK

        #if pnlid['aftid']=='create_strut'+aft or \
        #   pnlid['aftid']=='create_wing'+aft:
        #    invalid={'chord':1, 'nickname':1, 'medium':1, 'turbulence':1, 'finish':1, 're':1} # these get removed as we do them - when len(invalid)==0 we know we are good to go!
        #if pnlid['aftid']=='create_turbine'+aft:
        #    invalid={'nickname':1, 'hub_radius':1, 'turbulence':1, 'finish':1, 'medium':1, 'targetspeed':1, 'targettemperature':1, 'turbine_strategy':1, 'root_chord_info':1, 'hub_edge':1, 'hub_nose':1, 'tip_location':1, 'rpm':1} # these get removed as we do them - when len(invalid)==0 we know we are good to go! # 'targetaltitude':1,
        #if pnlid['aftid']=='global_settings'+aft:
        #    invalid={'text_info':1}
        #if pnlid['aftid']=='import_foil'+aft:
        #    invalid={'nose':1, 'tail':1, 'file_name':1}
        #if pnlid['aftid']=='export_foil'+aft:
        #    invalid={'file_name':1, 'select_spline':1}


        #######################
        # They want to update #
        if thisctrl=='update_now'+aft:
            if dlg[thisctrl].value == True:   # They just checked the update checkbox
                if ui:
                    click=ui.messageBox('You are using {} version {:1.8f}.\nThe newest version is {}.\n\nYou will need to stop and re-start this add-in after you update.\n\nUpdate now?'.format(program_name,VERSION,NEXTVERSION),'Confirm update',adsk.core.MessageBoxButtonTypes.OKCancelButtonType) # messageBox(text, title, buttons, icon); OKCancelButtonType==1
                    if click==0: #ok
                        userdata['applyupdate']=1
                        save_user_settings(ui,settings_file) 
                        updates(False) # Does the update
                        dlg[thisctrl].isVisible=False # remove update box - done now.
                        if updatedata.get('update_msg',0):
                            updatedata['update_msg']=''                         # don't show the old update message after they updated
                            if updatedata.get('last_check_time',0): updatedata['last_check_time']= updatedata['last_check_time']-24*60*60 # force a re-check on next reload
                            save_other_settings(update_file,updatedata)         # don't show the old update message when they restart, and force a re-check
                        check_update(5) # Run a new update-check to see if there is anything else waiting (important if we just updated the updater)
                        ui.messageBox('Update complete.  To re-start '+program_name+'; either:\n\n'+
                                      '(a) Restart Fusion 360, or\n\n'+
                                      '(b) Restart this add-in as follows:\n'+
                                      '     * push Shift-S to open the Scripts-and-Add-Ins menu,\n'+
                                      '     * Click the Add-Ins tab\n'+
                                      '     * Click on '+program_name+
                                      '\n     * Click Stop\n'+
                                      '     * Click Run','Update complete')

                        try:
                            pnl[panel]['banner'+aft].formattedText = cmdl[0]['dflt'] # line 1067
                        except:
                            eprint("failed setting banner text back to default")

                        # Reset this so new dialogues don't say old stuff:
                        # text=updatedata['update_msg']   
                        # update the update json (and set last-check to 0 for next)?
                        # Re-run updater now

                        # Go fix any panels that already got built:-
                        #pnl[panel][ctrl['id']]=ctrlinstance 

                    else:
                        dlg[thisctrl].value=False # un-check the box because they cancelled


        ###################
        # global settings #
        if thisctrl and pnlid['aftid']=='global_settings'+aft and cmdh[thisctrl]['type']=='BoolValueInput':
            if dlg['text_info'+aft].selectedItem.index >= 0: # is -1 if not set
                if invalid.get('text_info',0): invalid.pop('text_info') # no longer invalid
                if thisctrl=='text_info'+aft:
                    text_info=cmdh['text_info'+aft]['choices'][dlg['text_info'+aft].selectedItem.index] # 'Summary','Full','Off (None)'
                    userdata[thisctrl]=text_info                                    # Stash the kind they want into userdata (it will become default later too)
                    save_user_settings(ui,settings_file)                            # Keep a permanent record
                    vlog(program_name + " Text-Info preference changed to: {}".format(text_info))

            msgs={'share_errors'+aft: "Share-Errors preference changed to: {}",
                    'announcements'+aft:"Announcment subscription setting preference changed to: {}",
                    'displayName'+aft:  "Name changed to: {}",
                    'email'+aft:        "Email changed to: {}",
                    'check_update'+aft: "Running update check now: {}",
                    'hide_rating'+aft:  "Hide-Rating preference changed to: {}" }
            if thisctrl=='check_update'+aft:                                    # Do an immediate check
                check_update(1)
                dlg[thisctrl].value=False                                       # Did the check
            else:                                                               # Don't save the state of the checker box
                userdata[thisctrl]=dlg[thisctrl].value                          # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)
            vlog(program_name + " " + msgs[thisctrl].format(dlg[thisctrl].value))

            if thisctrl=='hide_rating'+aft:                                     # give them immediate feedback on the purpose of this option
                if dlg[thisctrl].value:
                    dlg['rating_group'+aft].isVisible=False
                    err(sharejson,"Thanks for using our FREE "+program_name+" ! - we hope you took a moment to rate us five stars?  If not, please show your appreciation for our hard work with a quick 5-star rating before you hide this request.")
                else:
                    dlg['rating_group'+aft].isVisible=True


            '''
            if thisctrl=='share_errors'+aft:
                userdata[thisctrl]=dlg[thisctrl].value                              # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)
                vlog(program_name + " Share-Errors preference changed to: {}".format(dlg[thisctrl].value))
            if thisctrl=='announcements'+aft:
                userdata[thisctrl]=dlg[thisctrl].value                              # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)
                vlog(program_name + " Announcment subscription setting preference changed to: {}".format(dlg[thisctrl].value))
                shareIt(sharejson,dlg=dlg,Text='send_email_prefernce_update')
            if thisctrl=='displayName'+aft:
                userdata[thisctrl]=dlg[thisctrl].value                              # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)                                # app.currentUser.displayName
                vlog(program_name + " Name changed to: {}".format(dlg[thisctrl].value))
                shareIt(sharejson,dlg=dlg,Text='send_email_prefernce_update')
            if thisctrl=='email'+aft:
                userdata[thisctrl]=dlg[thisctrl].value                              # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)                                # app.currentUser.email
                vlog(program_name + " Email changed to: {}".format(dlg[thisctrl].value))
                shareIt(sharejson,dlg=dlg,Text='send_email_prefernce_update')
            if thisctrl=='hide_rating'+aft:
                userdata[thisctrl]=dlg[thisctrl].value                              # Permanently save this setting - e.g. userdata['share_errors_aft']=True
                save_user_settings(ui,settings_file)
                vlog(program_name + " Share-Errors preference changed to: {}".format(dlg[thisctrl].value))
            '''

        # if userdata.get('text_info_aft',0) == 'Full': vlog("extra stuff!")    # example of appropriate way of telling user extra geeky stuff


        ###########################
        # Delete unwanted presets #
        if thisctrl=='save'+aft: # V2
            dlg['save_button_state'+aft]=dlg['save'+aft].value                      # Remember real value, so we can toggle this on errors without forgetting
            if dlg[thisctrl].value == False and len(dlg['nickname'+aft].value)>0:   # They just un-checked the save checkbox with a nickname entered
                if ui:
                    click=ui.messageBox('Delete saved pre-set "{}" - are you sure?'.format(dlg['nickname'+aft].value),'Confirm delete user preset',adsk.core.MessageBoxButtonTypes.OKCancelButtonType) # messageBox(text, title, buttons, icon); OKCancelButtonType==1
                    if click==0: #ok
                        pass  # remove etc here...


        ######################################################
        # Prepare for sharing (reveal / hide range elements) #
        if thisctrl=='share'+aft:
            showhide=[ 'minspeed', 'maxspeed', 'minaltitude', 'maxaltitude', 'mintemperature', 'maxtemperature' ] 
            if dlg[thisctrl].value:
                for ctrlname in showhide:
                    if dlg.get(ctrlname+aft): dlg[ctrlname+aft].isVisible=True
                web_nbreq('S={}&V={}&P={}'.format(int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds()),VERSION,panel)) # non-blocking socket.
                ''' testing:
                eprint("aftid={}".format( pnlid['aftid']))  # 'create_cowling'+'_aft'
                for ctrlname in dlg: # TEST - preparing for preset cache
                    eprint(ctrlname)
                eprint("and")
                for ctrlname in uiel[pnlid['aftid']]['controls']:
                    eprint(ctrlname)
                '''
            else:
                for ctrlname in showhide:
                    if dlg.get(ctrlname+aft): dlg[ctrlname+aft].isVisible=False


        ######################################
        # Show hidden stuff when appropriate #
        if dlg.get('nt_concentrate'+aft,None) is not None:
            for cpts in ['interpolate','num_points']:
                if dlg.get(cpts+aft,None) is not None:
                    if dlg[cpts+aft].value:
                        dlg['nt_concentrate'+aft].isVisible=True
                    else:
                        dlg['nt_concentrate'+aft].isVisible=False


        ####################################
        # Record a user vote for a feature #
        if thisctrl=='votebox'+aft and dlg[thisctrl].value:
            web_nbreq('vote={}&now={}&V={}'.format(panel,int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds()),VERSION)) # non-blocking socket.

        if dlg.get('votebox'+aft,0) and dlg['votebox'+aft].value:
            args.areInputsValid = True
            eprint("Voting - bypass following checks")
            return True # enable OK if they checked the box (does nothing - we recorded their vote when they clicked it)


        ###########################################
        # Reveal endap distance if they select it #
        if thisctrl=='endcap'+aft:
            ctrlname = 'enddist'
            if dlg[thisctrl].value: dlg[ctrlname+aft].isVisible=True
            else: dlg[ctrlname+aft].isVisible=False


        ###################################################
        # Don't let them pick a co-incident nose and tail # 
        if ( thisctrl == 'nose'+aft or thisctrl == 'tail'+aft ) and dlg['nose'+aft].selectionCount and dlg['tail'+aft].selectionCount and abs(((dlg['tail'+aft].selection(0).entity.geometry.z-dlg['nose'+aft].selection(0).entity.geometry.z)**2 + (dlg['tail'+aft].selection(0).entity.geometry.x-dlg['nose'+aft].selection(0).entity.geometry.x)**2 + (dlg['tail'+aft].selection(0).entity.geometry.y-dlg['nose'+aft].selection(0).entity.geometry.y)**2 )**0.5) < 0.000001:
            vlog("co-incident nose+tail point de-selected (chord must have length)")
            dlg[thisctrl].clearSelection()                                      # De-Select stuff we do not want them selecting

        '''

        if pnlid['aftid']=='create_strut'+aft or \
           pnlid['aftid']=='create_wing'+aft or \
           pnlid['aftid']=='import_foil'+aft:
            # Auto-select the tail point if they just selected the nose
            if dlg.get('nose'+aft,0) and dlg['nose'+aft].selectionCount and invalid.get('nose',0): invalid.pop('nose') # no longer invalid
            if dlg.get('tail'+aft,0) and dlg['tail'+aft].selectionCount and invalid.get('tail',0): invalid.pop('tail') # no longer invalid
            if caller=='changed':
                if thisctrl == 'nose'+aft:
                    if dlg['nose'+aft].selectionCount and not dlg['tail'+aft].selectionCount:
                        dlg['nose'+aft].hasFocus=False
                        dlg['nose'+aft].isEnabled=False
                        dlg['tail'+aft].hasFocus=True
                        dlg['tail'+aft].isEnabled=True

            # Auto-select the join point if they just selected the tail (after seletcing the nose)
            if caller=='changed' and thisctrl == 'tail'+aft and dlg['nose'+aft].selectionCount and dlg['tail'+aft].selectionCount and dlg.get('join'+aft,0) and not dlg['join'+aft].selectionCount:
                dlg['tail'+aft].hasFocus=False
                dlg['tail'+aft].isEnabled=False
                dlg['join'+aft].hasFocus=True
                dlg['join'+aft].isEnabled=True

        '''


        ##############################################
        # Enforce the nickname a valid variable name #
        if dlg.get('nickname'+aft,''):
            if re.findall(r'^([A-Za-z][A-Za-z_0-9]*|)$', dlg['nickname'+aft].value): # Accept blank, or, someting that starts with a letter and can then contain letters, numbers, or _
                dlg['nickname'+aft].isValueError = False
                if invalid.get('nickname',0): invalid.pop('nickname') # no longer invalid - now done generically above
                if dlg.get('save'+aft,0):                               # Save-Presets button
                    dlg['save'+aft].isEnabled=True
                    if dlg.get('save_button_state'+aft,1): # Default is to save
                        dlg['save'+aft].value=True # restore users chosen state
            else:
                invalid['nickname']=2                                   # make it invalid again (and auto-validate will skip this)
                #valid=-99 # prevent OK
                dlg['nickname'+aft].isValueError = True 
                if dlg.get('save'+aft,0):
                    dlg['save'+aft].isEnabled=False
                    dlg['save'+aft].value=False # un check it


        ################################################
        # Update chord and diameter info if they exist #
        if thisctrl=='centerline'+aft and dlg.get('root_chord_info'+aft,0): # hub chord
            dlg['root_chord_info'+aft].formattedText=''
            if dlg[thisctrl].selectionCount:
                if dlg[thisctrl].selection(0).entity.startSketchPoint.geometry.z != dlg[thisctrl].selection(0).entity.endSketchPoint.geometry.z: # ERROR: 'BRepEdge' object has no attribute 'startSketchPoint'
                    dlg[thisctrl].clearSelection()                 # De-Select stuff we do not want them selecting
                    ui.messageBox("Sorry - rotation axis must be aligned to X/Y\n\nTip: select a vertical line.")
                elif dlg[thisctrl].selectionCount:
                    cmdh['root_chord_nr'+aft]['value'] = dlg[thisctrl].selection(0).entity.length  # for insertTurbineUI to use
                    dlg['root_chord_info'+aft].formattedText = adsk.core.Application.get().activeProduct.unitsManager.formatInternalValue( dlg[thisctrl].selection(0).entity.length )

        if dlg.get('hub_nose'+aft,0) and dlg['hub_nose'+aft].selectionCount and dlg.get('centerline'+aft,0) and dlg['centerline'+aft].selectionCount:
            if dlg['hub_nose'+aft].selection(0).entity.geometry.distanceTo(dlg['centerline'+aft].selection(0).entity.startSketchPoint.geometry) != 0 and \
                dlg['hub_nose'+aft].selection(0).entity.geometry.distanceTo(dlg['centerline'+aft].selection(0).entity.endSketchPoint.geometry) != 0:
                dlg['hub_nose'+aft].clearSelection()
                ui.messageBox("The Hub nose must be one end of the Rotation axis\n\nTip: click on the start or end of the axis line.")

        if dlg.get('diameter_info'+aft,0):                                      # diameter
            dlg['diameter_info'+aft].formattedText=''
            if dlg['centerline'+aft].selectionCount and dlg['tip_location'+aft].selectionCount: # and dlg['hub_nose'+aft].selectionCount:
                centerline=dlg['centerline'+aft].selection(0).entity
                if dlg['tip_location'+aft].selection(0).entity.objectType == 'adsk::fusion::SketchPoint': 
                    tippoint=dlg['tip_location'+aft].selection(0).entity
                else: 
                    tippoint=dlg['tip_location'+aft].selection(0).entity.startSketchPoint
                diameter=2 * abs( centerline.startSketchPoint.geometry.z - tippoint.geometry.z )
                if diameter == 0:
                    dlg['tip_location'+aft].clearSelection()                    # De-Select stuff we do not want them selecting
                    ui.messageBox("Sorry - tip must be on a different Z plane than the axis.\n\nTip: select a vertical line for the hub axis, and a second vertical line in front or behind (on a different Z of) the hub axis for the tip.\n\nHint: use a 3d sketch, or two sketches on different Z planes.")
                else:
                    cmdh['radius_nr'+aft]['value'] = diameter / 2               # for insertTurbineUI to use
                    dlg['diameter_info'+aft].formattedText = adsk.core.Application.get().activeProduct.unitsManager.formatInternalValue( diameter )


        if thisctrl=='tip_location'+aft and dlg.get('tip_chord_info'+aft,0): # tip chord
            if not dlg[thisctrl].selectionCount: dlg['tip_chord_info'+aft].formattedText=''
            else:
                if dlg[thisctrl].selection(0).entity.objectType == 'adsk::fusion::SketchPoint':
                    cmdh['tip_chord_nr'+aft]['value'] = 0
                    dlg['tip_chord_info'+aft].formattedText='0'
                else:
                    cmdh['tip_chord_nr'+aft]['value'] =  dlg[thisctrl].selection(0).entity.length 
                    dlg['tip_chord_info'+aft].formattedText=adsk.core.Application.get().activeProduct.unitsManager.formatInternalValue( dlg[thisctrl].selection(0).entity.length )


        #######################################################
        # Auto-populate the strategy formula if they pick one #

        if thisctrl=='turbine_strategy'+aft and dlg.get(thisctrl,0) and dlg[thisctrl].selectedItem.index>=0:
            turbine_strategy=cmdh[thisctrl]['choices'][dlg[thisctrl].selectedItem.index] # 'Free tips', 'Inside Duct', 'Custom'
            free_tip_strategy='[ (u<=0.4) * (1-u) * root_chord    +    (u>0.4) * (section[s-1]["speed"] / speed) * section[s-1]["chord"] ]'   # Turbine blade chord formula - outer sections past 40% at constant RE
            inside_duct_strategy=  '[ root_chord / sin( ( zero_aoa - aoa ) / 180 * pi ) ]'
            existing_strategy=dlg['chord_formula'+aft].text
            if not cmdh[thisctrl].get('previous',''): cmdh[thisctrl]['previous']=existing_strategy
            eprint("exist={} prev={}".format(existing_strategy,  cmdh[thisctrl].get('previous','')  ))
            if turbine_strategy != 'Custom' and \
                not re.match(r'^\s*#',existing_strategy) and \
                existing_strategy != free_tip_strategy and \
                existing_strategy != inside_duct_strategy:                  # They typed something into the chord formula
                cmdh[thisctrl]['previous']=existing_strategy                # Keep anything they already typed into here
                eprint("stored prev={}".format( cmdh[thisctrl].get('previous','')  ))
            if(turbine_strategy == 'Free tips'):
                dlg['chord_formula'+aft].text=free_tip_strategy
                eprint("A set={}".format(dlg['chord_formula'+aft].text))
            elif(turbine_strategy == 'Inside Duct'):
                dlg['chord_formula'+aft].text=inside_duct_strategy
                eprint("B set={}".format(dlg['chord_formula'+aft].text))
            elif(turbine_strategy == 'Custom'):
                if re.match(r'^\s*#',existing_strategy) or existing_strategy == free_tip_strategy or existing_strategy == inside_duct_strategy and cmdh[thisctrl].get('previous',''):
                    dlg['chord_formula'+aft].text=cmdh[thisctrl]['previous']    # replace any of their typing that we might have overwritten above
                    eprint("C set={}".format(dlg['chord_formula'+aft].text))
                eprint("C now={}".format(dlg['chord_formula'+aft].text))
            else:
                err("Code problem with strategy selection - please tell us what happened here.")


        #######################################################
        # Validation applicable to wings/struts/turbines only #
        if pnlid['aftid']=='create_strut'+aft or \
           pnlid['aftid']=='create_turbine'+aft or \
           pnlid['aftid']=='create_propeller'+aft or \
           pnlid['aftid']=='create_wing'+aft:


            #################################################
            # Fill out their presets if they just chose one #
            if thisctrl=='preset'+aft:
                preset=cmdh['preset'+aft]['choices'][dlg['preset'+aft].selectedItem.index] # 'Hydrofoil surf board' , etc
                
                # Need to load their settings into the preset dict here...
                if preset == 'Your previous settings':
                    #err(sharejson,'This option not yet coded - sorry!')
                    last=None # Find the closest recent copy of their settings to apply
                    if userdata.get('last',''): last=userdata['last'] # default to overall last choices from anytime, anywhere
                    if userdata.get('last.'+pnlid['mywhere'],''): last=userdata[ 'last.'+pnlid['mywhere'] ] # use this main-menu last instead, if there is one. e.g. last.sketch_airfoil_command_aft
                    if userdata.get('last.'+pnlid['aftid'],''): last=userdata[ 'last.'+pnlid['aftid'] ]     # use this sub-menu last instead, if there is one. e.g. last.create_wing_aft
                    if userdata.get('last.'+panel,''): last=userdata[ 'last.'+panel ]                       # use this excat main+sub menu last instead, if there is one. e.g. last.sketch_airfoil_command_aft_create_wing_aft
                    #eprint(cmdh)
                    if last:
                        p=json_loads(last)
                        #eprint(p)
                        for v in p: # go through every setting in their last set of choices
                            if cmdh.get(v,None) is None:
                                continue

                            setting=p[v]

                            try:
                                if cmdh[v]['type']=='RadioButtonGroupCommandInput' or cmdh[v]['type']=='ButtonRowCommandInput' or cmdh[v]['type']=='DropDownCommandInput': # button / Dropdown
                                    if cmdh[v]['type']=='DropDownCommandInput' and dlg[v].selectedItem is None:
                                        dlg[v].listItems.add(" ",True)  # Need to first create the selected object before we can select a different one
                                    #?? dlg[v].selectedItem.index
                                    if dlg[v].selectedItem.index >= 0:
                                        dlg[v].listItems[ dlg[v].selectedItem.index ].isSelected=False # Un select previous
                                    dlg[v].listItems[setting].isSelected=True # Select preset
                                elif cmdh[v]['type']=='TextBoxCommandInput':
                                    if not cmdh[v].get('readonly',True):
                                        dlg[v].text=setting # set it
                                    else:
                                        eprint("skipping read-only {} {}".format(cmdh[v]['type'],v))    # OK for readonly of course
                                elif cmdh[v]['type']=='StringValueInput' or cmdh[v]['type']=='BoolValueInput' or cmdh[v]['type']=='ValueInput':
                                    dlg[v].value=setting # set it
                                #elif cmdh[v]['type']=='SelectionInput': don't save mouse selections...
                                else:
                                    eprint("no code to restore: {} {}".format(cmdh[v]['type'],v))    # OK for readonly of course
                            except:
                                eprint("problem loading saved-setting {}".format(v))
                                pass    # debugging


                elif cmdh['preset'+aft]['values'].get(preset,0):                              # One of ours - not "Your foil: theirs"
                    p=cmdh['preset'+aft]['values'][preset]  # We use the name of the value as the key
                            
                    # 'Medium R/C plane'  :{'medium'+aft:0, 'turbulence'+aft:2, 'finish'+aft:0,  'minspeed'+aft:4,    'targetspeed'+aft:6,    'maxspeed'+aft:12,    'minaltitude'+aft:0, 'targetaltitude'+aft:20, 'maxaltitude'+aft:150,  'mintemperature'+aft:10, 'targettemperature'+aft:16, 'maxtemperature'+aft:26,  'target_span'+aft:1, 'SKIP-target_mass'+aft:123 },
                    for v in p: # go through every setting in the preset
                        if not cmdh.get(v,0):
                            continue
                        # Fix the units my presets are in, into the units fusion uses
                        unitid=''
                        if cmdh[v].get('unit',0):
                            unitid=cmdh[v]['unit']       # m or C etc
                        setting=p[v]
                        if units.get(unitid,0):
                            if units[unitid].get('scale',0):
                                setting=setting*units[unitid]['scale']
                            elif units[unitid].get('offset',0):
                                setting=setting+units[unitid]['offset']

                        try:
                            if cmdh[v]['type']=='RadioButtonGroupCommandInput' or cmdh[v]['type']=='DropDownCommandInput': # button / Dropdown
                                if cmdh[v]['type']=='DropDownCommandInput' and dlg[v].selectedItem is None:
                                    dlg[v].listItems.add(" ",True)  # Need to first create the selected object before we can select a different one
                                #?? dlg[v].selectedItem.index
                                if dlg[v].selectedItem.index >= 0:
                                    dlg[v].listItems[ dlg[v].selectedItem.index ].isSelected=False # Un select previous
                                dlg[v].listItems[setting].isSelected=True # Select preset
                            else:
                                dlg[v].value=setting # set it
                        except:
                            eprint("problem loading preset {}".format(v))
                            pass    # debugging
                elif preset != '' and preset != '(none) - fill in below':
                    err(sharejson,"'Sorry - preset '{}' not found.".format(preset))


            #######################
            # Compute their chord #
            if pnlid['aftid']=='create_strut'+aft or \
               pnlid['aftid']=='create_wing'+aft:
                if dlg['nose'+aft].selectionCount and dlg['tail'+aft].selectionCount:
                    dist=dlg['nose'+aft].selection(0).entity.geometry.distanceTo(dlg['tail'+aft].selection(0).entity.geometry) # OK
                    dlg['chord_info'+aft].text=adsk.core.Application.get().activeProduct.unitsManager.formatInternalValue(dist)
                    if invalid.get('chord',0): invalid.pop('chord') # no longer invalid


            ##################################
            # Adjust controls for the medium #
            water=False
            if dlg['medium'+aft].selectedItem.index >= 0: # is -1 if not set
                if invalid.get('medium',0): invalid.pop('medium') # no longer invalid
                medium=cmdh['medium'+aft]['choices'][dlg['medium'+aft].selectedItem.index] # 'Air', 'Fresh Water', 'Sea Water', 'Any Water', 'Other'
                if medium[-5:] == 'Water': # All 3 that end with "Water"
                    water=True
                    dlg['altitude_group'+aft].isVisible=False # No altitude for water
                    if dlg.get('tip_mach_info'+aft,0):
                        dlg['tip_mach_info'+aft].isVisible=False # No mach for water
                    if dlg.get('mach_info'+aft,0):
                        dlg['mach_info'+aft].isVisible=False # No mach for water
                else:
                    dlg['altitude_group'+aft].isVisible=True
                    if dlg.get('tip_mach_info'+aft,0):
                        dlg['tip_mach_info'+aft].isVisible=True
                    if dlg.get('mach_info'+aft,0):
                        dlg['mach_info'+aft].isVisible=True
                if medium == 'Other':           # Enable the extra stuff they need to tell us
                    dlg['density'+aft].isEnabled=True
                    dlg['density'+aft].isVisible=True
                    dlg['dynamic_viscosity'+aft].isEnabled=True
                    dlg['dynamic_viscosity'+aft].isVisible=True
                    dlg['kinematic_viscosity'+aft].isEnabled=True
                    dlg['kinematic_viscosity'+aft].isVisible=True
                    #dlg['density_info'+aft].isEnabled=False
                    dlg['density_info'+aft].isVisible=False
                else:                           # Disable what we don't need
                    dlg['density'+aft].isEnabled=False
                    dlg['density'+aft].isVisible=False
                    dlg['dynamic_viscosity'+aft].isEnabled=False
                    dlg['dynamic_viscosity'+aft].isVisible=False
                    dlg['kinematic_viscosity'+aft].isEnabled=False
                    dlg['kinematic_viscosity'+aft].isVisible=False
                    #dlg['density_info'+aft].isEnabled=True
                    dlg['density_info'+aft].isVisible=True

            if thisctrl=='density'+aft:
                if not pnl.get('toldnodenseyet',0):
                    ui.messageBox("Sorry - the non-Air/non-Water feature will be available in the next release.  It does not currently have any effect now.")
                    pnl['toldnodenseyet']=True
            # 'density'
            # 'dynamic_viscosity'
            # 'kinematic_viscosity'


            #################
            # Compute Ncrit #
            ncrit=''
            if dlg['turbulence'+aft] and dlg['turbulence'+aft].selectedItem:    # Flow conditions
                if dlg['finish'+aft] and dlg['finish'+aft].selectedItem:        # Smoothness
                    turb=dlg['turbulence'+aft].selectedItem.index # 'Laminar and clean', 'Average', 'Turbulent and dirty'
                    fin=dlg['finish'+aft].selectedItem.index # 'Very rough', 'Average', 'Clean', 'Very smooth'
                    if turb>=0 and fin>=0:
                        turb=3-turb
                        fin=fin+1
                        # ncrit=3*finish + turb-1 +(turb==3&&finish<4)-(turb==1&&finish>1); 
                        # gives: f1t1=n3 f1t2=n4 f2t1=n5 f1t3=n6 f2t2=n7 f3t1=n8 f2t3=n9 f3t2=n10 f4t1=n11 f3t3=n12 f4t2=n13 f4t3=n14
                        ncrit=3*fin + turb-1 
                        if  turb==3 and fin < 4 :
                            ncrit = ncrit+1
                        if  turb==1 and fin > 1 :
                            ncrit = ncrit-1
                        dlg['N']=ncrit # for insert2dwing to use in a moment
                        ncrit = " N" + str(ncrit)
                        if invalid.get('finish',0): invalid.pop('finish') # no longer invalid
                        if invalid.get('turbulence',0): invalid.pop('turbulence') # no longer invalid

            #####################
            # Compute Re & Mach #
            v=f=s=a=t=d=Re=Mach=None
            if dlg.get('targettemperature'+aft,0):
                if dlg['targettemperature'+aft]:

                    if dlg['targetspeed'+aft] and dlg['targetspeed'+aft].value<=0:
                        if pnlid['aftid']=='create_strut'+aft or \
                           pnlid['aftid']=='create_wing'+aft:
                            invalid['re']=2 # force invalid - must know speed
                    elif dlg['targetspeed'+aft] and dlg['targetspeed'+aft].value>0:
                        if dlg['targetaltitude'+aft]:
                            if dlg['medium'+aft].selectedItem.index >= 0: # Water of some kind (is -1 if not set)
                                medium=cmdh['medium'+aft]['choices'][dlg['medium'+aft].selectedItem.index] # 'Air', 'Fresh Water', 'Sea Water', 'Any Water', 'Other'
                                s=dlg['targetspeed'+aft].value/100
                                a=dlg['targetaltitude'+aft].value/100
                                t=dlg['targettemperature'+aft].value -273.15
                                if water:
                                    Mach=0
                                else:
                                    Mach=machno(s, t+Altitude_to_Temperature(a))
                                if dist:
                                    chord=dist/100
                                    Re=getRe( medium, chord , s , a , t , dlg) # dlg in case 'Other', so callee can get density etc from user input
                                    if Re!=dlg.get('re',0): # Only tell them suff when they change stuff
                                        if water:
                                            vlog("Re {:,.0f} results from {:.5g}m chord at {:.5g}m/s speed in {:.5g}C temp {}".format(Re,chord,s,t, medium))
                                        else:
                                            vlog("Re {:,.0f} Mach {:.3g} results from {:.5g}m chord at {:.5g}m/s speed in {:.5g}C temp at {:.5g}m altitude in {}".format(Re,Mach,chord,s,t,a, medium))
                                        #eprint("Re {:12,.0f} results from {}m chord at {}m/s speed in {}C temp at {}m altitude".format(Re,chord,s,t,a))
                                    dlg['re_info'+aft].text="{:,.0f}{}".format(Re,ncrit)
                                    dlg['re']=Re        # for insert2dwing to use in a moment
                                    dlg['chord']=chord  # for insert2dwing to use in a moment
                                    dlg['mach_info'+aft].text="{:,.3f}".format(Mach)
                                    dlg['mach']=Mach    # for insert2dwing to use in a moment
                                    if Re>0:
                                        if invalid.get('re',0): invalid.pop('re') # no longer invalid
                                    #adsk.core.Application.get().activeProduct.unitsManager.formatInternalValue(dist)

                                    #eprint("Target Re=" + str(getRe( 'AIR', dist/100, dlg['targetspeed'+aft].value/100 , dlg['targetaltitude'+aft].value/100 , dlg['targettemperature'+aft].value-273.15  , dlg)))

                                # Fill in the hidden "hard stuff" every time they change the preset (unless they are about to enter the hard stuff themsevles)
                                #if thisctrl=='preset'+aft and medium != 'Other': 
                                if medium != 'Other': # 'Other' enables the following 3 input boxes, instead of us computing them
                                    if water: 
                                        d=Water_Density(medium,t) # kg/m^3
                                        dlg['density'+aft].value='{}'.format(d)
                                        v=Water_Viscosity(t)      # Pa.s
                                        dlg['dynamic_viscosity'+aft].value='{}'.format(v)
                                        if dlg.get('density_info'+aft,0):
                                            dlg['density_info'+aft].formattedText="D={:,.2f} kg/m<sup>3</sup> V={:.3e} Pa.s".format(d,v).replace("e-0", "e-")
                                    else:
                                        d=Altitude_to_Density(a)
                                        dlg['density'+aft].value='{}'.format(d)
                                        v=Gas_Viscosity(medium,t,dlg)
                                        dlg['dynamic_viscosity'+aft].value='{}'.format(v)
                                        if dlg.get('density_info'+aft,0):
                                            dlg['density_info'+aft].formattedText= "D={:,.3f} kg/m<sup>3</sup> V={:.3e} Pa.s".format(d,v).replace("e-0", "e-")
                                    if d:
                                        dlg['kinematic_viscosity'+aft].value='{}'.format(v/d) # cSt


            #####################
            # Tip speed and rpm #
            if thisctrl=='rpm'+aft and dlg[thisctrl].value>0 and cmdh['radius_nr'+aft].get('value',0):
                dlg['tip_speed'+aft].value = cmdh['radius_nr'+aft]['value'] * 2 * math.pi * dlg[thisctrl].value / radperstorpm / 60 # compute the tip speed for them

            if thisctrl=='tip_speed'+aft and dlg[thisctrl].value>0 and cmdh['radius_nr'+aft].get('value',0):
                dlg['rpm'+aft].value = 60 * radperstorpm * dlg[thisctrl].value / (cmdh['radius_nr'+aft]['value'] * 2 * math.pi)  # compute the rpm for them

            ###################
            # Tip mach number #
            if  dlg.get('tip_speed'+aft,0) and dlg['tip_speed'+aft].value>0 and d is not None and v is not None:
                ta=t
                if a is not None:
                    ta=t+Altitude_to_Temperature(a)
                dlg['tip_mach_info'+aft].formattedText= "{:,.3f}".format(machno( dlg['tip_speed'+aft].value/100 , ta )).replace("e-0", "e-")
            

            #################################
            # Manual invalidations checking #
            if pnlid['aftid']=='create_turbine'+aft:
                if dlg['targetspeed'+aft] and dlg['targetspeed'+aft].value>0:
                    if invalid.get('targetspeed',0): invalid.pop('targetspeed') # no longer invalid
                if dlg['targettemperature'+aft]:
                    if invalid.get('targettemperature',0): invalid.pop('targettemperature') # no longer invalid
                if dlg['turbine_strategy'+aft].selectedItem.index >= 0:         # is -1 if not set
                    if invalid.get('turbine_strategy',0): invalid.pop('turbine_strategy') # no longer invalid
                if dlg['targetaltitude'+aft]:
                    if invalid.get('targetaltitude',0): invalid.pop('targetaltitude') # no longer invalid
                if dlg['rpm'+aft] and dlg['rpm'+aft].value>0:
                    if invalid.get('rpm',0): invalid.pop('rpm')                     # no longer invalid




                if dlg['centerline'+aft].selectionCount \
                   and dlg['hub_nose'+aft].selectionCount:
                    #and dlg['hub_edge'+aft].selectionCount:
                    center_line=dlg['centerline'+aft].selection(0).entity
                    root_chord=dlg['centerline'+aft].selection(0).entity.length / 100       # in meters
                    hub_nose=dlg['hub_nose'+aft].selection(0).entity
                    #hub_edge=dlg['hub_edge'+aft].selection(0).entity
                    center_point=center_line.endSketchPoint
                    if not hub_nose.geometry.distanceTo(center_point.geometry) > 0:
                        center_point=center_line.startSketchPoint
                    #hub_radius=hub_edge.geometry.distanceTo(center_point.geometry) / 100    # in meters
                    #if root_chord>0 and hub_radius>0:
                    #    if invalid.get('hub_radius',0): invalid.pop('hub_radius') # no longer invalid

        else: # Not a foil (no nickname)        # /Validation applicable to wings/struts/turbines only

            if thisctrl == 'file_button'+aft:
                file_name = ''
                fileDlg = ui.createFileDialog()
                fileDlg.isMultiSelectEnabled = False
                fileDlg.title = 'Import File'
                fileDlg.filter = 'Selig/Lednicer/Coordinate files (*.dat;*.cor;*.txt);;Text or CSV files (*.csv;*.txt);;All files (*.*)'

                if pnlid['aftid']=='export_foil'+aft:
                    # Show file save dialog
                    dlgResult = fileDlg.showSave()
                    if dlgResult == adsk.core.DialogResults.DialogOK:
                        filename=fileDlg.filename # Add .dat or .csv on the end if they didn't have this already
                        if ( dlg['datcsv'+aft] and dlg['datcsv'+aft].selectedItem and dlg['datcsv'+aft].selectedItem.index==1 ) and not re.search(r'\.csv$',filename,re.IGNORECASE):
                            filename=filename + '.csv'
                            if os.path.isfile(filename): ui.messageBox("Caution - existing file will be overwritten:-\n{}".format(filename))
                        elif not ( dlg['datcsv'+aft] and dlg['datcsv'+aft].selectedItem and dlg['datcsv'+aft].selectedItem.index==1 ) and not re.search(r'\.dat$',filename,re.IGNORECASE):
                            filename=filename + '.dat'
                            if os.path.isfile(filename): ui.messageBox("Caution - existing file will be overwritten:-\n{}".format(filename))
                        dlg['file_name'+aft].text = filename
                else:
                    # Show file open dialog
                    dlgResult = fileDlg.showOpen()
                    if dlgResult == adsk.core.DialogResults.DialogOK:
                        dlg['file_name'+aft].text = fileDlg.filename

            if dlg.get('file_name'+aft,0) and dlg['file_name'+aft].text != nofiletxt: # cmdh['file_name'+aft]['dflt']: - The API strips the HTML from the .text attribute
                if invalid.get('file_name',0): invalid.pop('file_name') # no longer invalid


        #########################################################
        # Move the selection focus to the next required element #
        if thisctrl and cmdh[thisctrl]['type'] == 'SelectionInput' and (cmdh[thisctrl].get('lim',None) is None or cmdh[thisctrl]['lim'][0]): # It's a selection *other* than an optional one.
            #eprint("auto-sel from {} now thinking...".format(thisctrl))
            if dlg.get(thisctrl,0) and dlg[thisctrl].selectionCount:                # they now selected something
                controls=uiel[pnlid['aftid']]['controls']                           # Get the list, in order, of what's in this panel
                nextsel=None; afterthis=False
                for c in controls:
                    if afterthis and not nextsel and cmdh[c+aft]['type'] == 'SelectionInput' and (cmdh[c+aft].get('lim',None) is None or cmdh[c+aft]['lim'][0] or cmdh[c+aft].get('autosel',0) ): 
                        nextsel=c+aft
                    if cmdh[c+aft]['id']==thisctrl:                                     # So we can ignore looking for any SelectionInputs until after we've seen the one they're on now.
                        afterthis=True
                if nextsel and dlg.get(nextsel,0) and not dlg[nextsel].selectionCount:  # Not already chosen it
                    dlg[thisctrl].hasFocus=False                                    # Move the cursor for them
                    dlg[thisctrl].isEnabled=False
                    dlg[nextsel].hasFocus=True
                    dlg[nextsel].isEnabled=True
                #elif nextsel and dlg.get(nextsel,0):
                #    eprint("Skipped... {} is already chosen".format(nextsel))
                #else:
                #    eprint("Skipped... nothing comes next")
            #else:
            #    eprint("Skipped... nothing selected yet")
        #elif thisctrl and cmdh[thisctrl]['type'] == 'SelectionInput':
        #    eprint("auto-sel from {} not a lim".format(thisctrl))



        ################################################################################################################
        # Go through everything we know they need to supply us in the menu dialog, and see if they've told us that yet #
        doneok=[]
        for i in invalid:
            if invalid[i] == 1 and dlg.get(i+aft,0):                            # A generic control we can validate automatically which is in the UI
                if cmdh[i+aft]['type'] == 'RadioButtonGroupCommandInput':
                    if dlg[i+aft].selectedItem.index >= 0: doneok.append(i) # is -1 if not set
                elif cmdh[i+aft]['type'] == 'SelectionInput':
                    if dlg[i+aft].selectionCount:
                        if cmdh[i+aft].get('allowed',{}):                       # Only allow selecting things we support
                            sel=dlg[i+aft].selection(0)
                            if cmdh[i+aft]['allowed'].get(sel.entity.objectType,0):
                                doneok.append(i)
                            else:
                                if userdata.get('text_info_aft',0) == 'Full':
                                    ok=[]
                                    for j in cmdh[i+aft]['allowed']: ok.append(j)
                                    msg = "You selected an object of type {} which {} does not recognize as a foil shape.\n\nPlease select one of the following kinds of objects instead:\n\n{}".format(sel.entity.objectType,program_name,ok)
                                    vlog(msg)
                                    dlg[i+aft].clearSelection()                 # De-Select stuff we do not want them selecting
                                    ui.messageBox(msg)

                        else:                                                   # Any selection is OK
                            doneok.append(i)
                elif cmdh[i+aft]['type'] == 'ValueInput':
                    #eprint("ctrl {} value={}".format(i,dlg[i+aft].value))
                    #if dlg[i+aft] and dlg[i+aft].value > 0:
                    if dlg[i+aft].value > 0: doneok.append(i)
                elif cmdh[i+aft]['type'] == 'TextBoxCommandInput':              # these controls are for output, not input
                    if not cmdh[i+aft].get('readonly',True):
                        if dlg[i+aft].text != '' and \
                            not re.match(r'^\s*#',dlg[i+aft].text) and \
                            dlg[i+aft].text != nofiletxt : doneok.append(i) # chord formula # is a comment, and '(no file selected yet)'
                            # not re.search(r'\(.+\)',dlg[i+aft].text): doneok.append(i) # chord formula # is a comment, and '(no file selected yet)'
                elif cmdh[i+aft]['type'] == 'StringValueInput':
                    if dlg[i+aft].value != '': doneok.append(i)                 # How to handle things (temp?) where 0=OK ?
                elif cmdh[i+aft]['type'] == 'BoolValueInput':                   # is either on or off - no "undertermined" state exists.
                    pass # TODO
                elif cmdh[i+aft]['type'] == 'ButtonRowCommandInput':            # default state will always win
                    pass # TODO
                elif cmdh[i+aft]['type'] == 'DropDownCommandInput':             # default state will always win
                    pass # TODO
        for ok in doneok:
            invalid.pop(ok)


        #######################################
        # Help them fill in the form properly #
        if dlg.get('helpok'+aft,0) and dlg['helpok'+aft].value:
            dlg['helpok'+aft].value=False # un-check the box again
            missing='You still need to supply:-\n'
            if not invalid:
                missing="All Good! If you don't need any of the optional\nfeatures, go ahead and click OK now\n\n(If OK is still not clickable, that's a bug! :-\nplease send us a screenshot of all your input)"
            for missed in invalid:
                if cmdh.get(missed+aft,'') and cmdh[missed+aft].get('name'):
                    missing = missing + "\n" + cmdh[missed+aft]['name'] + "\n"         # Return the human name for this missing control ID
                else:
                    missing = missing + "\n" + missed + "\n"                           # Or, e.g. "file_name" if it's a derived control.
                if cmdh.get(missed+aft,'') and cmdh[missed+aft].get('tip'):
                    missing = missing + "    " + cmdh[missed+aft]['tip'] + "\n" # And more info too...
            ui.messageBox(missing)
            test=42
            # cmdl id:xxx  name + tip ?
        #eprint("Still to go: {}".format(invalid))


        #if(valid>=6):   # We need all 5 things (nose, tail, finish, medium, flow, velocity, altitude (if not water) & temp ... before we can do the foil)
        if not len(invalid):                                                    # If everything needed has been removed from the invalid dict:
            args.areInputsValid = True
            eprint("MyValidate OK") # VALCHK
            return True
        else:                                                                   # Something still invalid exists
            args.areInputsValid = False
            eprint(invalid)    # VALCHK
            return False

    except:
        emsg=program_name+' ' + caller + ' event failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())
        err(sharejson,emsg) # opens a messageBox



foilinfo=[ # Supporting info for data that might get inserted into their user parameters
            {'alpha':'Ideal angle of attack (AoA) for foil: '}, # '13.9', + Foil_fn
            {'Cm':'Pitching-moment coefficient relative to quarter chord (center of pressure at design AoA)'}, # '0.9837',
            #{'Foil_fn':''}, # cannot have text parameters: 'Original filename of the input airfoil'}, # 'spline_foil_v13006.dat-s17_n17s0.dat',
            #'cmt' => 'NEWv1.20200425 s(0) foildb2020/e377m.dat (sim 0.000193213738 ) l3 Foil spline 17pts', # foil comments
            {'Re':'Closest test Reynolds number found corresponding to your input criteria - see user_Re'},    # 100000000, # real Re
            {'n':'Closest test log(critical angle) - transition criterion determined from your flow and finish input'}, # 11,
            #{'chord':[ 'cm', 'straight distance from nose to tail' ]}, # 123
            {'r':'criteria used to determine "ideal" (6=bell-curve weighted overall best using +/- 3 significant degrees AoA from ideal)'}, # 1,
            {'CL':'Coefficient of lift at design AoA'}, # '1.5369',
            {'CD':'Coefficient of drag at design AoA'}, # '0.00372',
            {'clcd':'CL/CD (foils) or 1/CD (symmetric) - bigger is always better'}, # '0.00372',
            {'CDf':'friction-drag component contributing to CD'}, # '0.0016',
            {'CDpres':'pressure-drag component contributing to CD'}, # '0.0016',
            {'good':'foil selection criteria: max weighted (see "r") average "clcd"'}, # '0.00372',
            {'CDp':'dynamic pressure max'}, # '0.0016',
            {'Cpmin':'Minimum pressure extent (cavitation indicator) at design AoA'}, # '-21.4023',
            {'sym':'Degree of foil symmetry. 0=exactly symmetric, >0.00125 non-symmetric'}, # '0.097049938590344',
            {'cmt':''}, # 'Spline conversion data (sim = similarity-match with original foil data)'}, # 'NEWv1.20200425 s(0) foildb2020/v13006.dat (sim 0.000016577733 ) l3 Foil spline 17pts',
            {'Top_Xtr':'Top position of forced transition from laminar to turbulent at design AoA'}, # '0.00146'
            {'Bot_Xtr':'Bottom position of forced transition from laminar to turbulent at design AoA'}, # '0.0014',
            # {'Top_Itr':'.'}, # '0.9837',
            # {'Bot_Itr':'.'}, # '78.0185',
            {'xver':'solver version number'}, # '6.99',
            {'diter':'solution iteration count'}, # '6.99',
            # {'src':'solver machine ID'}, # '6.99',
            # {'ok':''}, # 1,
            # {'s':''}, # 0,
            #{'ver':'insertion processor version number'}, # 'v.1.20200504',
            {'iver':'solver manager version number'}, # '1.20200505',
            {'id':''}, # 1,
            # {'xy':''} # [ 1}, 0, '0.855242605109347', '0.0117125705324037', '0.602519634254835', '0.0275105856224779', '0.37458813296931', '0.0368627176605416', '0.190731108083199', '0.0377396028682893',
            ]



# Get the current sketch, or if not in a sketch, create a suitable new sketch and project anything needed into it
def MySketch(*proj):
    app = adsk.core.Application.get()

    sketch = app.activeEditObject                                               # Get the current sketch?
    # Get current sketch, or, prepare to create new one if they're not in one right now (note: creating a new sketch breaks the UI input - erases their selection points)
    if not sketch or sketch.objectType != 'adsk::fusion::Sketch':
        rootComp = None
        try:
            for bestComp in proj:
                rootComp=bestComp.parentSketch.parentComponent                  # Try to create new sketches in the correct component (whichever works first)
                if rootComp: 
                    break
            if not rootComp: rootComp = design.rootComponent                    # Default to the root component of the active design.
            sketches = rootComp.sketches                                        # Create a new sketch on the same plane as what the user selected the nose in
        except:
            rootComp = design.rootComponent                                     # Default to the root component of the active design.
            sketches = rootComp.sketches                                        # Create a new sketch on the same plane as what the user selected the nose in
        # CAUTION - next line breaks the users input
        sketch = sketches.add( proj[-1].parentSketch.referencePlane )           # Use same plane as whatever their chosen nose was (XY, XZ, or YZ) # ERROR: not a planar entity

    # project any foreign points into our current sketch (needed so the X,Y makes sense later)
    for bestComp in proj:
        if bestComp.parentSketch.name != sketch.name:                           # project any foreign things (chord lines, points) into here  ERROR: AttributeError: 'ConstructionPoint' object has no attribute 'parentSketch'
            proj=sketch.project(bestComp)
            proj.isConstruction=True

    return sketch

# Remove tab, space, CR, LF, and EOF characters from all lines
def chompnls(dat):
    zap=False
    for i in range(len(dat)):
        dat[i]=dat[i].decode('ISO-8859-1')                                      # turn it back into text (we read it as binary)
        if zap: dat[i]=''
        eof=re.match(r'(.*)\x1a',dat[i])
        if eof:
            dat[i]=eof.group()
            zap=True # drop everything after the EOF
        while dat[i] != "" and \
            ( dat[i][-1]==chr(9) or \
              dat[i][-1]==chr(32) or \
              dat[i][-1]==chr(10) or \
              dat[i][-1]==chr(13) or \
              dat[i][-1]==chr(26) ):
            dat[i]=dat[i][:-1] # modify the callers original data

# Copy user-entered data out of the dialoge form controls into the sharejson dict
def shareDlg(sharejson,dlg,cache=None,nickname=''): 
    for userinfo in cmdl:
        if userinfo.get('id',0) and dlg.get(userinfo['id'],0) and userinfo.get('type',''): # and userinfo.get('name',None) is not None: # and dlg[userinfo['id']].isVisible:
            id=userinfo['id']
            #eprint("share_checking: {}".format(id))
            pname=nickname + aft+'_user_' + id
            if userinfo['type']=='ValueInput':
                valr = dlg[id].value
                if isinstance(valr, (str, int)): valr=float(valr)
                sharejson[ id ]=valr
            elif userinfo['type']=='StringValueInput' or userinfo['type']=='BoolValueInput':
                if dlg[id].value != '': sharejson[ id ]=dlg[id].value
            elif userinfo['type']=='TextBoxCommandInput':
                if dlg[id].text != '': sharejson[ id ]=dlg[id].text
            elif userinfo['type']=='SelectionInput':
                ent=None
                if cache and cache.get(id,None): ent=cache.get(id,None)
                if not ent: ent=dlg[id]
                if ent and ent.objectType != 'adsk::fusion::SketchPoint' and ent.objectType != 'adsk::fusion::SketchLine' and ent.selectionCount: ent=ent.selection(0).entity
                if ent and ent.objectType == 'adsk::fusion::SketchPoint':
                    sharejson[ id ]='({},{},{})'.format( ent.geometry.x, ent.geometry.y, ent.geometry.z) 
                elif ent and ent.objectType == 'adsk::fusion::SketchLine':
                    sharejson[ id ]='({},{},{})-({},{},{})'.format( ent.startSketchPoint.geometry.x, 
                                                                                ent.startSketchPoint.geometry.y,
                                                                                ent.startSketchPoint.geometry.z,
                                                                                ent.endSketchPoint.geometry.x,
                                                                                ent.endSketchPoint.geometry.y,
                                                                                ent.endSketchPoint.geometry.z)
            elif userinfo['type']== 'StringValueInput' or userinfo['type']== 'BoolValueInput': 
                sharejson[ id ]=dlg[id].value
            elif userinfo['type']== 'ButtonRowCommandInput' or userinfo['type']== 'RadioButtonGroupCommandInput' or userinfo['type']== 'DropDownCommandInput':
                sel=dlg[id].selectedItem
                if sel: sel=sel.index
                sharejson[ id ]=sel # dlg[id].selectedItem.index
                if sel is not None and cmdh[id].get('choices',[]):
                    sharejson[ id+".value" ]=cmdh[id]['choices'][sel] # Also include the name of whatever they selected
                elif sel is not None and cmdh[id].get('buttons',[]):
                    sharejson[ id+".value" ]=cmdh[id]['buttons'][sel]['name'] # Also include the button description name of whatever they selected

                # 'preset_aft' => 6,
            #else:
            #    eprint("cant save non-value parm {} {}".format(id,userinfo.get('type','')))
    sharejson['now']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())
    return sharejson
# shareDlg

# Let us know what they did if they checked the box to allow it
def shareIt(sharejson,dlg=None,Text=None): # dlg is None only when critical errors were encountered
    if dlg: shareDlg(sharejson,dlg)
    if Text or dlg is None or ( dlg.get('share'+aft,0) and dlg['share'+aft].value ):    # dlg is None only when errors occurred and share_errors setting is True.
        tmp={}
        for k in sharejson:
            try:
                #Better to do everything we can, than skip everything other than what we think we can: if type(sharejson[k]) in (str, int, bool, dict, list, tuple, float, type(None) ): tmp[k]=sharejson[k]
                test=json_dumps(sharejson[k]) # skips the next line if it cannot encode it
                tmp[k]=sharejson[k]
            except:
                eprint("Cannot send {} through JSON".format(k))
        if Text and Text != 'send_email_prefernce_update':
            return json_dumps(tmp)  # Caller merely want to serialize sharejson
        eprint("Sharing...")
        web_nbreq('s='+urllib.parse.quote(json_dumps(tmp), safe='')) #non-blocking socket. might need to do this first, so we get a chance to poll it?  or hook CommandHandler?
        #crashes - _thread.start_new_thread(web_req, ('u=u1.20200506&v=1.20200503&a=_aft&p=Airfoil_Tools',))
    return sharejson # so the caller can do: return shareIt() if they want



# Load and optionally repair any kind of airfoil coordinate data file
def loadDatFile(filename,usefoil,sharejson,repair=True,dlg=None):
    app = adsk.core.Application.get()
    ui = app.userInterface
    try:
        with open(filename,'rb' ) as dat_file:
            foil = dat_file.read() # .splitlines(True) - Grr. folds some \r\n into \n\n
            dat_file.close()
    except:
        return err(sharejson,"importExport read failed for file: {}\n{}".format( filename, traceback.format_exc() ),dlg) # signal error to caller and # pops up a messageBox

    # So, if they opt-in to sharing this with us, we get a copy of the precise input (so we ca learn how to auto-correct any more oddness that arrives later from the wild)
    sharejson['importfn']=filename
    sharejson['import']=foil.decode('latin-1')


    # Strip tab, space, CR, LF, and EOF characters from all lines
    if len(foil) < 12:
        return err(sharejson,'This file is not an airfoil (too small):\n{}'.format(filename),dlg) # signal error to caller and # pops up a messageBox

    foil=re.split(b'(?:\r\n|\n\r|\r|\n)',foil)     # deal more intelligently with possibly garbled max/dos line endings that we might encounter
    chompnls(foil)

    # Strip all blank last lines - side effect of our \r\n splitting causes an empty last line always...
    while foil[-1] == '': foil=foil[:-1]
    if len(foil) < 3:
        return err(sharejson,'This file is not an airfoil (not enough lines):\n{}'.format(filename),dlg) # signal error to caller and # pops up a messageBox

    # Repair all the strange kinds of stuffups that typically exist in airfoil data files around the internet...
    if repair:

        # some files have x/c and y/c on line 2
        if re.match(r'^\s*x\/c[\s,:;]*y\/c$',foil[1],re.IGNORECASE): foil=[foil[0],*foil[2:]]

        # some files have the format name \n numrows \s numrows \n \n data \n \n data - where the second set of data is backwards, with duplicated middle
        numrows=re.match(r'^\s*([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)\s*$',foil[1],re.IGNORECASE)

        if numrows and \
           numrows.lastindex>1 and \
           int( float(numrows.group(1)) ) > 1 and \
           int( float(numrows.group(2)) ) > 1 and \
           float( numrows.group(1) ) == int(float(numrows.group(1))) and \
           float( numrows.group(2) ) == int(float(numrows.group(2))) and \
           foil[3+int(float(numrows.group(1)))] == '':
            midsep=3+int(float(numrows.group(1)))
            newfoil=foil[0:2] # headings
            ifrom=2
            while(foil[ifrom]==''): ifrom=ifrom+1 # find the first non-blank
            frontbackwards=re.search(r'((?:\d|[.\-\+]\d|[\+\-]\.\d)[\d\.\-\+e]*)[^\d\.\-\+]+((?:\d|[.\-\+]\d|[\-\+]\.\d)[\d\.\-\+e]*)',foil[ifrom],re.IGNORECASE)
            backfrontwards=re.search(r'((?:\d|[.\-\+]\d|[\+\-]\.\d)[\d\.\-\+e]*)[^\d\.\-\+]+((?:\d|[.\-\+]\d|[\-\+]\.\d)[\d\.\-\+e]*)',foil[midsep+1],re.IGNORECASE)
            if frontbackwards and float(frontbackwards.group(1))<0.5: # this file is backwards at the start
                for i in range(midsep,ifrom,-1): newfoil.append(foil[i-1]) # reverse the backwards start - note the i-1 in append - since the range is plus one
                #newfoil.append(*foil[midsep+1:])
                foil=[*newfoil[0:],*foil[midsep+1:]]
                if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} commenced backwards: repaired".format(filename))
            elif backfrontwards and float(backfrontwards.group(1))>0.5: # this file is backwards at the end
                newfoil=foil[0:2+int(float(numrows.group(1)))]
                for i in range(len(foil)-1,midsep,-1): newfoil.append(foil[i])
                #foil=[*foil[0:2+int(float(numrows.group(1)))],*foil[4+int(float(numrows.group(1))):]] # Remove blank middle
                foil=newfoil
                if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} ended backwards: repaired".format(filename))

            foil[1]='' # blank out the line count number (so it gets deleted next)
            while foil[1] == '':
                foil=[foil[0],*foil[2:]] # Delete all blank lines at the top

        # some files have free-form comments before the numbers...
        while True:
            text=re.match(r'^([a-zA-Z])',foil[1],re.IGNORECASE)
            if text:
                foil=[foil[0],*foil[2:]] # Delete all text lines after the top(name)
            else:
                break

        # some files have a '1.0000     ......' placeholder at the start
        if re.search(r'(\.\.)',foil[1],re.IGNORECASE):
            if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} unnecessary starting placeholder '{}' removed".format(filename,foil[1]))
            foil=[foil[0],*foil[2:]] # Delete it text lines after the top(name)

        # some files have free-form comments after 2 blank lines near the end (note that some files have a blank line in the middle of the numbers... at the (0,0))
        trunc=0
        for i in range(1,len(foil)): # (my $i=1;$i<$#foil;$i++) {
            if trunc<1 and ( foil[i] == '' or re.search(r'(http:|\*modif)', foil[i],re.IGNORECASE) is not None) and foil[i-1] == '':
                trunc=i-1
        if trunc>0:
            foil=foil[0:trunc-1]
            if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} had garbage removed after line {}".format(filename,trunc))

        # some files have free-form comments at the end
        while len(foil) and (re.search('[a-df-z\\\/#;\%\>]',foil[-1],re.IGNORECASE) or foil[-1]=='' or not re.search(r'((?:\d|[.\-\+]\d|[\+\-]\.\d)[\d\.\-\+e]*)[^\d\.\-\+]+((?:\d|[.\-\+]\d|[\-\+]\.\d)[\d\.\-\+e]*)',foil[-1],re.IGNORECASE)): # numbers [incl negative and scientific notation] separated by anything
            foil=foil[:-1]

        # rare uncoded fixups:       top -o CPUaa((?:\d|[.\-])\d+[\d\.\-e]*)[^\d\.\-]+((?:\d|[.\-])\d+[\d\.\-e]*)
        # some files end with 1,0 followed by 0,0
        # push @junk, pop @foil while(("$foil[-2] $foil[-1]"=~/([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)/)&&($1==1)&&($2==$4)&&($3==$1)); # dupe last line
        # if re.search'([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)',foil[0] + " " + foil[1] and \
        #     &&($1==1)&&($2==$4)&&($3==$1)); # dupe last line
        # push @junk, pop @foil while(("$foil[-2] $foil[-1]"=~/([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)/)&&($1==0)&&($2==$4)&&($3==$1)); # dupe nulls
        # push @junk, pop @foil while(("$foil[-2] $foil[-1]"=~/([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)[\s,:;]+([\+\-0-9e\.]+)/)&&($1==1)&&($2==0)&&($3==0)&&($4==0)); # spurious nulls at end

        # If the first line looks like coordinates - insert a new first line
        if len(foil) and re.match(r'^\s*[\+\-0-9e\.]+[\s,:;]+[\+\-0-9e\.]+\s*$',foil[0]):
            foil=['Temporary heading added by' + program_name,*foil]
            if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} was missing a header - temporary one added".format(filename))

    if len(foil) < 3:
        return err(sharejson,'This file is not an airfoil (not enough data):\n{}'.format(filename),dlg) # signal error to caller and # pops up a messageBox


    ll=1
    #eprint("line:{}:'{}'".format(ll,foil[0]))
    for line in foil[1:]:                                               # Skip text heading
        ll=ll+1
        #eprint("line:{}:'{}'".format(ll,line))
        #vlog("line:{}:'{}'".format(ll,line))
        xy=re.search(r'((?:\d|[.\-\+]\d|[\+\-]\.\d)[\d\.\-\+e]*)[^\d\.\-\+]+((?:\d|[.\-\+]\d|[\-\+]\.\d)[\d\.\-\+e]*)',line,re.IGNORECASE) # numbers [incl negative and scientific notation] separated by anything
        if not xy:
            eprint("Skipping no-numbers-found line {}:{} '{}'".format(filename,ll,line))
            if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} line {} skipped: was not polar coordinates: '{}'".format(filename,ll,line))
        else:
            try:
                xy=(float(xy.group(1)),float(xy.group(2)))
            except:
                eprint("Could not determine foil coordinates from the following line {}:{}: '{}'".format(filename,ll,line))
                vlog('Caution: Could not determine foil coordinates from the following line {}:{}:\n{}\n'.format(filename,ll,line))
                #ui.messageBox('Could not determine foil coordinates from the following line:\n\n{}'.format(line))
                xy=[1,0]    # [1,0] is tail, [0,0] is nose
            usefoil['xy']=[*usefoil['xy'],xy[0],xy[1]] # .append(xy)                                        # for insertFoil

    if repair: # not ( dlg.get('nofix'+aft,0) and dlg['nofix'+aft].value ):
        # Normalize (especially for control-splines so that the nose and AoA are good)
        # Need to see where the actual nose is for control-point splines, so we can ensure the line goes through their selection.
        (scale,rot,tx,ty)=normalize(usefoil) # scale, rotate, translate (in that order).  Note that for foils with open ends, we normalize the nose to be opposite the center of the open end, BUT, we insert aligned to the top side of that end - so the nose point will not appear to be correct if joining on the tail
        if userdata.get('text_info_aft',0) == 'Full': vlog("Imported file {} Normalized via scale:{} rotate:{} translate({},{})".format(filename,scale,rot*180/math.pi,tx,ty))

    #eprint("Loaded {}: {}".format(filename,foil[0])) #vlog("Loaded {}: {}".format(filename,foil[0]))

    return False    # no error

    # loadDatFile


# Load or save a foil in a text file
def importExport(pnlid,args,inputs):
    global VERSION, pnl, userdata, settings_file, foilinfo
    try:
        app = adsk.core.Application.get()
        ui = app.userInterface
        design = app.activeProduct
        sharejson={}

        panel=pnlid['mywhere'] + '_' + pnlid['aftid'] # 'sketch_airfoil_command_aft' + '_' + 'menuid':'create_cowling_aft'
        dlg=pnl[panel]
        eprint("importExport")
        filename=dlg['file_name'+aft].text



        # 'file_button', 'file_name', 'nose', 'tail', 'chord_info', 'join', 
        # 'aoa', 'spline_type', 'interpolate', 'wrap_axis',     'minthick', 'up',     'constrain', 'endcap', 'enddist', 'share',    'nofix', 
        ##########
        # IMPORT #
        ##########
        if pnlid.get('aftid')=='import_foil'+aft:

            nose=dlg['nose'+aft].selection(0).entity
            tail=dlg['tail'+aft].selection(0).entity
            if dlg.get('wrap_axis'+aft) and dlg['wrap_axis'+aft].selectionCount: wrap_axis=dlg['wrap_axis'+aft].selection(0).entity
            else: wrap_axis=None
            constrain=dlg.get('constrain'+aft,0) and dlg['constrain'+aft].value     # Make this optional, since it is so slow.
            camber=dlg.get('camber'+aft,0) and dlg['camber'+aft].value

            if dlg.get('join'+aft,0) and dlg['join'+aft].selectionCount:
                join=dlg['join'+aft].selection(0).entity
            else:
                join=tail

            chordlen=nose.geometry.distanceTo(tail.geometry)/100 #(in meters)
            dir=tail.geometry.vectorTo(nose.geometry) # dirn.x / .y / .z : x<0 means nose is on the left.  y<0 means nose is at the bottom.  z<0 means nose is at the back
            if(dir.x<0): swapx=-1
            else: swapx=1
            eprint("Inserting chord of len: {}m".format(chordlen))

            if dlg.get('nofix'+aft,0) and dlg['nofix'+aft].value :
                repair=False
            else: repair=True

            spline_type = dlg.get('spline_type'+aft,0) and dlg['spline_type'+aft].selectedItem.index    # 0 = default = fit-point

            usefoil={'imported':filename,'Cm':0,'alpha':0,'s':spline_type,'xy':[]}                      # Get it ready for loadDatFile/insertFoil to use

            # Read in the file, fix fixable things, and normalise the data
            if loadDatFile(filename,usefoil,sharejson,repair,dlg):
                #return err(sharejson,dlg=dlg) # Pops up a message box of the error
                return sharejson # caller already popped up a message

            up = dlg.get('up'+aft,0) and dlg['up'+aft].selectedItem.index                               # 0 = default = upright
            if up: up=-1
            else: up=1
            doendcap=dlg.get('endcap'+aft,0) and dlg['endcap'+aft].value            # Do they want rounded ends
            if doendcap and dlg.get('enddist'+aft,0): endscale=dlg['enddist'+aft].value/200      # How pointy to make the endcaps
            else: endscale=0


            # CAUTION - next line loses the users selection inputs above
            sketch = MySketch(join,nose,tail)                                   # Get the current sketch (creates new if we're not in one; last element provides the plane, first element provides the component - project all foreign things into sketch too)
            sketchPoints = sketch.sketchPoints


            # Name the sketch for them:
            sname=re.search('.*[\\/:](.+)',filename) # strip path
            # big_test()
            if sketch.name[:6]=='Sketch':
                if not sname: sketch.name=filename
                else: sketch.name=sname.group(1)                                # Name their sketch for them, if they've not already done that themselves

            nt_concentrate=10 # defaults
            if dlg.get('nt_concentrate'+aft,None) is not None: nt_concentrate=int( dlg['nt_concentrate'+aft].value )
            if nt_concentrate<=1: nt_concentrate= 1.000001 # 1 breaks it - this works 

            # insert the imported foil here
            insertFoil( sketch=sketch,      # Works only in X,Y plane (handles Z OK, but needs foil in the XY)
                    nose=nose,              # dlg['nose'+aft].selection(0).entity
                    tail=tail,              # dlg['tail'+aft].selection(0).entity
                    join=join,              # which axis to do any rotations around
                    usefoil=usefoil,        # just imported
                    rotate=0,               # Flag to apply auto-rotation.  360 means none [ignores join] # TODO - allow auto-rotate if the metadata finds this
                    flip=up,                # anti-clockwise? : must be 1 or -1
                    extrarotate=dlg['aoa'+aft].value /(math.pi/180),         # Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
                    wrapline=wrap_axis,     # Center of a cylinder that the foil will be bent around
                    wrapstart=join,         # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                    endscale=endscale,      # pointyness of the end caps (0 if none)
                    textInfo=True,          # Tell the user what we did in the text info area (and, for errors, with UI popups)
                    constrain=constrain,
                    camber=camber,
                    minthick=dlg['minthick'+aft].value / 100 / chordlen,   # in relative meters (e.g. 0.4mm = 0.0004) # ensure the foil can be printed (stop T/E being 0 wide)
                    splinetype=spline_type, # 0=fitpoint, 1=controlpoint
                    interpolate=dlg['interpolate'+aft].value ,          # 0=do not change, 60=expand to 60 points
                    equispaceS=nt_concentrate,  # how much to scrunch up the points at the nose and tail
                    dlg=dlg,                # So we can add metadata (attributes) to the foil as well
                    sharejson=sharejson)    # sharejson is output

            # cmdl holds all dlg[]
            shareDlg(sharejson,dlg,nickname='import') # Copy all form data into sharejson
            shareIt(sharejson,dlg=dlg) # Let us know what they did if they checked the box to allow it.

            remember(sharejson,pnlid)                                                 # remember their most recent settings, so we can offer then again in future
            return False # no error





        # 'select_spline',   'test_spline',    'file_button', 'file_name', 'helpok',
        #                    'original', 'normalize', 'threed', 'unwrap', 'datcsv',
        ##########
        # EXPORT #
        ##########
        else: # self.pnlid.get('aftid')=='export_foil'+aft:

            spline=dlg['select_spline'+aft].selection(0)

            if spline.entity.objectType != 'adsk::fusion::SketchFittedSpline' and \
               spline.entity.objectType != 'adsk::fusion::SketchFixedSpline':
                return err(sharejson,"Sorry - you can only export Fitted or Fixed splines - not a {}".format( spline.entity.objectType ,dlg)) # signal error to caller and # pops up a messageBox

            if dlg['datcsv'+aft] and dlg['datcsv'+aft].selectedItem and dlg['datcsv'+aft].selectedItem.index==1: csv=True # 0=dat, 1=csv
            else: csv=False

            if dlg['threed'+aft] and dlg['threed'+aft].value: threed=True
            else: threed=False

            filename=dlg['file_name'+aft].text                                  # Had had .dat or .csv added on the end if they didn't have this already
            sname=re.search('.*[\\/:](.+)',filename) # strip path
            if not sname: sname=filename
            else: sname=sname.group(1)

            original=False
            if dlg['original'+aft] and dlg['original'+aft].value:
                if spline.entity.attributes.itemByName("attr"+aft,"JSON"):
                    try:
                        orixy=json.loads( spline.entity.attributes.itemByName("attr"+aft,"JSON").value )
                        if threed:
                            foil=points_to_1xyz( orixy['usefoil']['xy'],0 )
                        else:
                            foil=points_to_1xy( orixy['usefoil']['xy'] )
                        original=True 
                    except:
                       pass

            if dlg['normalize'+aft] and dlg['normalize'+aft].value: normalizeme=True 
            else: normalizeme=False

            if dlg.get('unwrap',0) and dlg['unwrap'+aft].value: unwrap=True 
            else: unwrap=False

            if dlg['metadata'+aft] and dlg['metadata'+aft].value: metadata=True
            else: metadata=False


            # Ensure they selected something we can handle
#'adsk::fusion::SketchFittedSpline'
#'adsk::fusion::SketchFixedSpline'
            # If it's a nurbs, only let them do ones we know about
            # spline.entity.geometry.knots

            # Make sure it's got some points
            # spline.entity.geometry.controlPointCount

            # Get all the points
            if not original:
                foil=[]
                for i in spline.entity.geometry.controlPoints.iterator():
                    foil.append([ i.x, i.y, i.z ])


            # Featurecreep...

            # Original
            # spline.entity.attributes.itemByName("attr"+aft,"JSON")


            # Normalize
            # Look around the shape for a disconnected part - that's the tail
            # If not tail - Look around the shape for the highest angle between panels - that's the tail

            if normalizeme and threed:
                err(sharejson,"Normalize only works in 2D right now.  Z data discarded",dlg)        # signal error to caller and # pops up a messageBox


            if normalizeme:
                usefoil={'exported':filename,'Cm':0,'alpha':0,'s':0,'xy':[]}                      # Get it ready for loadDatFile/insertFoil to use
                for xy in foil:
                    usefoil['xy']=[*usefoil['xy'],xy[0],xy[1]]                                     # for normalize

                (scale,rot,tx,ty)=normalize(usefoil) # scale, rotate, translate (in that order).  Note that for foils with open ends, we normlaize the nose to be opposite the center of the open end, BUT, we insert aligned to the top side of that end - so the nose point will not appear to be correct if joining on the tail
                (jx,jy)=points_to_xy( usefoil['xy'] )
                foil=[]
                for i in range(len(jx)):
                    if threed:
                        foil.append([jx[i],jy[i],0])
                    else:
                        foil.append([jx[i],jy[i]])

                nmsg="Normalized via scale:{} rotate:{} translate({},{})".format(scale,rot*180/math.pi,tx,ty)
                if userdata.get('text_info_aft',0) == 'Full': vlog("Export file {} {}".format(filename,nmsg))
            else:
                nmsg=""



            # Unwrap

            # put it into usefoil format and call usefoil

            # Add any known metadata into the export - .e.g - the original unit spline data and perfomance metrics etc.
            if metadata and spline.entity.attributes.itemByName("attr"+aft,"JSON"): sjson=nmsg + "\t" + spline.entity.attributes.itemByName("attr"+aft,"JSON").value
            else: sjson=nmsg
            banner='{} - exported by {} v{} - http://foils.pro/AirfoilTools/\t{}'.format(sname,program_name,VERSION,sjson)

            try:
                with open(filename, 'w') as dat_file:
                    if csv:
                        dat_file.write('"{}"\n'.format(banner.replace('"','""')))   # CSV escapes " by putting ""
                        for xyz in foil:
                            if threed:
                                dat_file.write('{},{},{}\n'.format( xyz[0],xyz[1],xyz[2] ))
                            else:   
                                dat_file.write('{},{}\n'.format( xyz[0],xyz[1] ))

                    else: # .dat
                        dat_file.write('{}\n'.format(banner))
                        for xyz in foil:
                            if threed:
                                dat_file.write('{}\t{}\t{}\n'.format( xyz[0],xyz[1],xyz[2] ))
                            else:   
                                dat_file.write('{}\t{}\n'.format( xyz[0],xyz[1] ))

                    dat_file.close()
            except:
                return err(sharejson,"importExport write failed for file: {}\n{}".format( filename, traceback.format_exc() ),dlg) # signal error to caller and # pops up a messageBox

            sharejson['panel']=panel
            sharejson['export']=foil
            shareIt(sharejson,dlg=dlg) # Let us know what they did if they checked the box to allow it.
            remember(sharejson,pnlid)                                                 # remember their most recent settings, so we can offer then again in future
            return False # no error


        # never happens
        remember(sharejson,pnlid)                                                 # remember their most recent settings, so we can offer then again in future
        return sharejson    # Everything about what we just did

    # Error handling
    except:
        return err(sharejson,program_name+' importExport failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox

    # importExport

# testc.entity.geometry.knots , testc.entity.geometry.controlPointCount, for i in testc.entity.geometry.controlPoints.iterator(): print("({},{},{})".format(i.x,i.y,i.z)), testc.entity.attributes
# testc.attributes.add('ADSK-AttribSample', 'FaceArea', str(selectedFace.area))
# testc.entity.attributes.add("attr"+aft,"JSON",'{"VERSION":1.20201112}')
# testc.entity.attributes.itemByName("attr"+aft,"JSON")
# testc.entity.attributes.itemByName("attr"+aft,"JSON").value 



            #test=dlg['select_spline'+aft]
            #testc=dlg['select_spline'+aft].selection(0)
            #testf=dlg['test_spline'+aft].selection(0)
            #testf.entity.objectType

            # testc.entity.geometry.knots , testc.entity.geometry.controlPointCount, for i in testc.entity.geometry.controlPoints.iterator(): print("({},{},{})".format(i.x,i.y,i.z)), testc.entity.attributes



# Thoughts; Ask... RPM or Tip-Speed (default=optimal) rpm  -  prop/turbine - rpm, torque, hp/kw, amps/volts/phase efficiency, etc etc
#                  Blade-Length or Tip-Location/Direction (default=optimal) [if tip specified, still allow optional length (choices: optimal, to tip, specified)
#                  Hub; use the length of the root chord - when rotated, if it exceeds hub line, that is the center hub edge.
#                  Root-Chord; optimal, centerline-long, user-specified
#                  Tip-Chord; optimal
#                  Winglets/Duct option? (no)
#                  strategy? max l/d .vs. spanwide constant l/d ?
#                  note: friction reduces as Re increases ( http://www.marinewiki.org/index.php/Frictional_Resistance )
#################################################################################
####                                                                          ###
##                              Do the Work at last!!                          ##
####                                                                          ###
#################################################################################
def insertTurbineUI(pnlid,args,inputs): # Does Turbines AND propellers.
    global VERSION, pnl, userdata, settings_file, foilinfo, cmdh
    try:
        app = adsk.core.Application.get()
        ui = app.userInterface
        design = app.activeProduct

        panel=pnlid['mywhere'] + '_' + pnlid['aftid'] # solid_airfoil_command_aft_create_turbine_aft e.g. 'sketch_airfoil_command_aft' + '_' + 'menuid':'create_cowling_aft'
        if re.search('turbine',pnlid['aftid']): spintype='turbine';turbine=True
        else: spintype='propeller'; turbine=False
        dlg=pnl[panel]
        pnl['calcstart']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())
        sharejson={ panel: int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds()) }

        # Set up some constants...
        prop_radius= float( cmdh['radius_nr'+aft]['value'] ) / 100                        # From MyValidate...
        root_chord= float( cmdh['root_chord_nr'+aft]['value'] ) / 100
        tip_chord= float( cmdh['tip_chord_nr'+aft]['value'] ) / 100

        nickname=dlg['nickname'+aft].value                                      # Triggers inserting user parms if named
        center_line=dlg['centerline'+aft].selection(0).entity
        hub_nose=dlg['hub_nose'+aft].selection(0).entity
        tip_sel=dlg['tip_location'+aft].selection(0).entity

        if tip_sel.objectType == 'adsk::fusion::SketchPoint': 
            tip_location=tip_sel
        else: 
            tip_location=tip_sel.startSketchPoint

        nose_point=center_line.startSketchPoint
        center_point=center_line.endSketchPoint
        if not hub_nose.geometry.distanceTo(center_point.geometry) > 0:
            center_point=center_line.startSketchPoint
            nose_point=center_line.endSketchPoint

        if center_point.geometry.z > tip_location.geometry.z: zdir=-1             # Which way did they draw on the Z?
        else: zdir = 1

        #root_chord=dlg['centerline'+aft].selection(0).entity.length / 100       # in meters
        #hub_edge=dlg['hub_edge'+aft].selection(0).entity
        #hub_edge=None


        #hub_radius=hub_edge.geometry.distanceTo(center_point.geometry) / 100    # in meters
        #prop_radius=tip_location.geometry.distanceTo(center_point.geometry)/100 # in meters - can fail - test
        #test=hub_edge.geometry.distanceTo(nose_point.geometry) / 100            # in meters
        #if hub_radius > test:
        #    hub_radius=test

        in_speed=dlg['targetspeed'+aft].value / 100                             # in meters / sec
        rpm=dlg['rpm'+aft].value / radperstorpm                                 # convert from radians per second (fusion360 units) to rpm
        minthick=dlg['minthick'+aft].value / 100                                # in meters (e.g. 0.4mm = 0.0004)
        num_profiles=int(float(dlg['num_profiles'+aft].value))                  # how many sections - defaults to 12
        if num_profiles < 1: num_profiles = 12
        num_profiles = num_profiles + 1                                         # We always skip profile 0...

        turbine_strategy=cmdh['turbine_strategy'+aft]['choices'][dlg['turbine_strategy'+aft].selectedItem.index] # 'Free tips', 'Inside Duct'

        eprint("{} strategy={} in-rpm={} real rpm={}".format(spintype, turbine_strategy, dlg['rpm'+aft].value, rpm ))
        #TEMP:
        #prop_radius=root_chord*2.2 # center_line.geometry.distanceTo(center_point.geometry)

        ''' Things we can use...

                'controls':[ 'banner', 'update_now', 
                             'nickname',
                             'beta_group', 'alpha_info',
                             'location_group', 'centerline', 'root_chord_info', 'hub_nose', 'tip_location', 'tip_chord_info', 'diameter_info',

                             'conditions_group', 'preset', 'medium', 'density', 'density_info', 'dynamic_viscosity', 'kinematic_viscosity', 'turbulence', 'finish', # 'tip_mach_info', # 're_info',
                             'velocity_group', 'minspeed', 'targetspeed', 'maxspeed', 
                             'altitude_group', 'minaltitude', 'targetaltitude', 'maxaltitude', 
                             'temperature_group', 'mintemperature', 'targettemperature', 'maxtemperature',

                             'strategy_group', 'turbine_strategy', 'chord_formula', 'formula_instructions',
                             'constraints_group', 'rotate_direction', 'rpm', 'tip_speed', 'tip_mach_info', # 'blade_len',
                             'blade_count', 'vary_pitch', 'num_profiles', 'minthick', 'camber',

                             #'save_preset_group', 'save', 'remove',
                             'settings_group', 'share', 'helpok', 'rating_group', 'rate_me'  ] },

        '''



        # CAUTION - next line loses the users selection inputs above
        sketch = MySketch(center_line,hub_nose,tip_location)                    # Get the current sketch (creates new if we're not in one; last element provides the plane, first element provides the component - project all foreign things into sketch too)
        sketchPoints = sketch.sketchPoints

        # Name the sketch for them:
        if sketch.name[:6]=='Sketch':
            if nickname:
                sketch.name=nickname+'_' + spintype                              # Name their sketch for them, if they've not already done that themselves
            else:
                sketch.name=spintype                                             # Name their sketch for them, if they've not already done that themselves

        #hub_Re=dlg['re']        # Inserted by the validate code - not relevant - speed is incoming airflow, not profile speed
        N=dlg['N']              # Inserted by the validate code

        rps=rpm/60
        blade_len=prop_radius           # -hub_radius 


        camber=False; num_rails=4; num_points=60; nt_concentrate=10 # defaults

        camber=dlg.get('camber'+aft,0) and dlg['camber'+aft].value
        if dlg.get('num_points'+aft,0) and dlg['num_points'+aft].value > 0: num_points=int( dlg['num_points'+aft].value )
        if dlg.get('num_rails'+aft,None) is not None: num_rails=int( dlg['num_rails'+aft].value )
        if num_rails > num_points: num_rails = num_points
        if dlg.get('nt_concentrate'+aft,None) is not None: nt_concentrate=int( dlg['nt_concentrate'+aft].value )
        if nt_concentrate<=1: nt_concentrate= 1.000001 # 1 breaks it - this works 

        patchme=dlg.get('patchme'+aft,0) and dlg['patchme'+aft].value
        cloud_foil=dlg.get('cloud_foil'+aft,0) and dlg['cloud_foil'+aft].value

        # test
        #le_points = adsk.core.ObjectCollection.create() # Leading-edge spline
        #te_points = adsk.core.ObjectCollection.create() # Trailing edge line

        #num_rails=4                                                             # Add this many splines between sections (min2, so we get the LE and TE at least!)
        #num_rails=8
        rail_points = []
        if num_rails:
            for rail in range(0,num_rails+1):                                       # +1 in case minthick
                rail_points.append(adsk.core.ObjectCollection.create())             # Leading-edge, trailing edge, and other splines

        # For Re computations
        medium=cmdh['medium'+aft]['choices'][dlg['medium'+aft].selectedItem.index] # 'Air', 'Fresh Water', 'Sea Water', 'Any Water', 'Other'
        a=dlg['targetaltitude'+aft].value/100                                   # Altitude
        t=dlg['targettemperature'+aft].value -273.15                            # Temperature
        #strategy='duct' # which kind of turbine to make




        mathok={k: v for k, v in math.__dict__.items() if not k.startswith("__")} # print(eval("3*pi",mathok))

        ##############################################################################################
        # pre-compute starting data for all the proposed sections (for the formula and later to use) #
        section=[]                                                              # Pre-Cache all the section info so "next" and "previous" are available to the formula
        for s in range(0,num_profiles):                                         # 0(hub) upto 12 (tip) (that's 13 remember)
            section_radius=prop_radius * (s/(num_profiles-1))**2                # tip (prop_radius) down to zero
            section_speed=rps * math.pi * 2 * (section_radius)                  # profile rotational speed in M/S
            # https://www.intechopen.com/books/advances-in-wind-power/low-speed-wind-turbine-design
            # Pturb watts = .5 * efficiency(betz) * density kg/m^3 * area m^2 * speed^3 m/s  (where Pturb is the mechanical power of the turbine in Watts )
            flow_extra=0
            section_translational_speed = 0
            max_power_watts=0
            mach=0
            if section_speed>0:
                if turbine:
                    #energy=16/27                    # Placeholder: best possible - betz limit - how much power we plan to extract from the flow
                    speed_loss_factor = (dlg['target_efficiency'+aft].value/100) ** 0.5 # energy = .5 times density times area time velocity squared; so the resulting speed after removing 59% of the energy must be e=e*sqrt(energy). w.g 2m/s=>1.53622915m/s
                    max_power_watts= 0.5 * dlg['target_efficiency'+aft].value/100 * float(dlg['density'+aft].value) * math.pi * prop_radius**2 * in_speed**3 # watts
                    zero_aoa=math.atan((speed_loss_factor * in_speed) / section_speed ) / math.pi * 180 # Work out the needed pitch at this section for the RPM and speed (in_speed is less because of betz limit)
                    section_translational_speed = ( (section_speed)**2 + (speed_loss_factor * in_speed)**2 ) ** 0.5 # Work out the actual airspeed for each section
                    flow_extra = (speed_loss_factor * in_speed) - in_speed      # negative for turbines
                    if medium[-5:] != 'Water':
                        mach=machno(section_translational_speed, t+Altitude_to_Temperature(a))
                    else:
                        mach=0
                    if(flow_extra> -0.01):
                        vlog("ERROR: flow is too much: s={} section_radius={} section_speed={} speed_loss_factor={} section_translational_speed={} flow_extra={} d={} z={}".format(s,section_radius,section_speed,speed_loss_factor,section_translational_speed,flow_extra,dlg['target_efficiency'+aft].value, zero_aoa ))
                        flow_extra=0

                else:
                    zero_aoa=ZeroAoA( in_speed, section_speed, dlg['target_thrust'+aft].value/100 , float(dlg['density'+aft].value) , math.pi * prop_radius**2 )# returns degrees # What is the zero angle-of-attack, knowing the speeds and thrust info?
                    # eprint("ZeroAoA: in_speed={} section_speed={} thrustN={} density={} discarea={} zeroAoA={}".format(in_speed,section_speed,thrustN,density,discarea, math.atan(( in_speed + NtoThrust(thrustN,density,discarea) ) / section_speed ) / math.pi * 180 ))
                    flow_extra=NtoThrust( dlg['target_thrust'+aft].value/100 , float(dlg['density'+aft].value) , math.pi * prop_radius**2 )
                    section_translational_speed = ( (section_speed)**2 + ( in_speed + flow_extra )**2 ) ** 0.5 # Work out the actual airspeed for each section
                    max_power_watts= 0.5 * float(dlg['density'+aft].value) * math.pi * prop_radius**2 * flow_extra**3 # watts
                    if medium[-5:] != 'Water':
                        mach=machno(section_translational_speed, t+Altitude_to_Temperature(a))
                    else:
                        mach=0
            else:
                zero_aoa=90


            section.append({'section_radius':section_radius,                    # Distance of this section from the hub (e.g. tip = prop radius)
                            'u':(s/(num_profiles-1))**2 ,                       # Proportion of this section away from the hub. 0= hub, 1=tip
                            's':s,                                              # integer station number - best not to use, since sections are not equal spaced
                            'speed':section_translational_speed,                # actual airspeed of this section
                            'zero_aoa': zero_aoa,                               # AoA at this station whihc would be directly into the flow
                            'rotational_speed':section_speed,                   # How fast is this turning here - best not to use, since the flow makes this faster (props) or slower (turbines)
                            'flow_extra':flow_extra,                            # How much do our blades change the speed of the medium (in m/s^2)
                            'max_power_watts':max_power_watts,                  # How much energy are we getting out (or putting in)
                            'mach':mach,
                            'chord':0,
                            'foil':{},
                            'aoa':0 })                                          # The AoA that the best suited foil profile will assume (above the zero_aoa of course)


        mmsg="tip_speed={}m/s".format(section_translational_speed)
        if medium[-5:] != 'Water': # All 3 that end with "Water"
            mmsg=mmsg+" mach={:,.3f} T={:,.0f}".format(machno( section_translational_speed , t+Altitude_to_Temperature(a) ), t+Altitude_to_Temperature(a) ).replace("e-0", "e-")

        vlog("New {} {}. nickname={} root_chord={}m prop_radius={}m in_speed={}m/s out_speed={}m/s max_power={}W {} rpm={:,.0f} minthick={} N={} num_profiles={} blade_len={}m medium={} alt={}m temperature_offset={}C abs_temp={}C".format(turbine_strategy, spintype, nickname, root_chord, prop_radius, in_speed, (in_speed + flow_extra), max_power_watts, mmsg, rpm, minthick, N, num_profiles, blade_len, medium, a, t, t+Altitude_to_Temperature(a))) # last calc from above is the tip

        #print(eval("3*pi+x",mathok,{'s':station, 'diameter':radius*2, 'root_chord':root_chord, 'tip_chord':tip_chord, 'section':section, 'flow_speed':in_speed, 'radius':section[station]['radius'], 'speed':section[station]['speed'], 'zero_aoa':section[station]['zero_aoa'], 'rotational_speed':section[station]['rotational_speed'], 'inflow_speed':section[station]['inflow_speed'], 'aoa':section[station]['aoa'] }))
        formula=dlg['chord_formula'+aft].text
        #formula='[ root_chord / sin( ( zero_aoa - aoa ) / 180 * pi ) ]'                     # ducted blade formula - use full area
        #formula='[ (u>0.4) * (section[s-1]["speed"] / speed) * section[s-1]["chord"]    +   (u<=0.4) * (1-u) * root_chord ]'   # Turbine blade chord formula - outer sections past 40% at constant RE
        # formula = [ sin(pi*( u -  0.117647059 ) * 1.117647059 )* tip_chord *(u>0.226) + (u<=0.226)* root_chord ]  # kinda like my 10x4.5L R/C prop
        #f= [ (u<=0.4) * (1-u) * root_chord    +    (u>0.4) * (section[s-1]["speed"] / speed) * section[s-1]["chord"] , -chord/2 , (u>=0.9) * ((0.9-u)*10)**2 * -tip_chord , u**6*tip_chord ]

        move_tail=0     # how much we shift the tail away from the rotation line
        scimitar=0      # how much we advance the foil 
        hedral=0        # changes the plane of this section
        xtilt=0         # tilts about the X axis - use with hedral to accomodate sideways inflow
        liftamt=1       # what kind of shape?  liftamt=1 for wings, liftamt=0 for struts
        manualaoa=0     # angle-of-attack override (e.g. to feather the tips)
        #sharejson['sections']=[]

        # Idea - keep doing halves - stop when chosen foil is the same between 2 halves?
        #for s in range(num_profiles-1,-1,-1):                                   # 11 (tip) down to 0 (hub)
        for redo in range(0,num_profiles+1):                                      # We re-compute all this many times, to back-propogate any answers for the real formula
            for s in range(num_profiles-1,0,-1):                                    # 11 (tip) down to 1 (amost hub)  (python[11] = 12th section)
                localok={'s':s,                                                     # the integer
                        'section':section,                                          # the [list]
                        'diameter':prop_radius*2,
                        'root_chord':root_chord,
                        'tip_chord':tip_chord,
                        'flow_speed':in_speed,
                        # density, viscosity, medium, altitude, etc etc
                        'section_radius':section[s]['section_radius'],
                        'u':section[s]['u'],    # unit chord
                        'num_profiles':num_profiles-1,
                        'speed':section[s]['speed'],
                        'zero_aoa':section[s]['zero_aoa'],
                        'rotational_speed':section[s]['rotational_speed'],
                        'chord':section[s]['chord'],
                        'foil':section[s]['foil'],
                        'aoa':section[s]['aoa'],
                        'mach':section[s]['mach'],
                        'dlg':dlg } # give them full access to everything in the UI too
                try:
                    formres = eval( formula, mathok, localok )                      # Let them use sin() and pi and tan(): acos acosh asin asinh atan atan2 atanh ceil copysign cos cosh degrees erf erfc exp expm1 fabs factorial floor fmod frexp fsum gamma gcd hypot isclose isfinite isinf isnan ldexp lgamma log log1p log10 log2 modf pow radians remainder sin sinh sqrt tan tanh trunc pi e tau inf nan
                    # Here next, so the try catches all errors (note: the user sees the comment - yay!)
                    section_chord=formres[0] # Note: your formula must return a python list in the form: [chord, tailx, rake, hedral] with at least the chord in it (the rest are optional)
                                                            # Required: the length of the chord of this blade section
                    if len(formres)>1: move_tail=formres[1] # Optional, default=0: How far to move the tail of the section forward (+) or backward (-) of the root-to-tip line. 0 places all blade tails in a straight line.
                    if len(formres)>2: scimitar =formres[2] # Optional, default=0: How far to advance this section ahead (+) or behind (-) the root-to-tip line. If you want to rake back your blade tips, this is the number to use
                    if len(formres)>3: hedral   =formres[3] # Optional, default=0: How far to shift this section into (+) or away from (-) the incoming flow. Makes non-flat blades, which trace a shape like a bowl (+) or mushroom (-) as they rotate
                    if len(formres)>4: xtilt    =formres[4] # Optional, default=0: # of degrees to tilt the section along the X axis. 90 degrees, combined with Hedral, at the tip would make winglets
                    if len(formres)>5: liftamt  =formres[5] # Optional, default=1: What kind of foil shape to use: 1 is a wing (makes lift) 0 is a strut (zero lift - handy for reducing vortices, interesting to combine with Hedral and Xt_tilt)
                    if len(formres)>6: manualaoa=formres[6] # Optional, default=0: Number of degrees to manually vary this section angle of attack (useful for feathering tips to reduce vortices)

                except:
                    formres=[]
                    return err(sharejson,'Your formula:-\n\n{}\n\nencountered the error:-\n\n{}\n\nUsing variables:-\n\n{}'.format(formula,traceback.format_exc(),localok)) # pops up a messageBox

                #eprint('formula:-\n\n{}\n\nUses variables:-\n\n{}'.format(formula,localok))

                # Compute Re
                section_radius=section[s]['section_radius']
                section_translational_speed = section[s]['speed']
                zero_aoa=section[s]['zero_aoa']
                Re=getRe( medium, section_chord , section_translational_speed , a , t , dlg)
                mach=machno(section_translational_speed, t+Altitude_to_Temperature(a))

                section_foil=bestFoil( Re=Re, N=N, strut=False, TextInfo=False, liftamt=liftamt, Mach=mach, Cloud=cloud_foil, WantS=0)
                section[s]['aoa']= section_foil.get('alpha',0) 
                section[s]['foil']=section_foil # gets overwritten at the end with more data

                # back propogation - so the formula can use the output of its own formula...
                section[s]['chord']=section_chord
                section[s]['move_tail']=move_tail
                section[s]['scimitar']=scimitar
                section[s]['hedral']=hedral
                section[s]['xtilt']=xtilt
                section[s]['liftamt']=liftamt
                section[s]['manualaoa']=manualaoa


                if redo==num_profiles: # This is the last loop - do it for real now...

                    mmsg="speed={:.5g}m/s".format(section_translational_speed)
                    if medium[-5:] != 'Water': # All 3 that end with "Water"
                        mmsg=mmsg+" mach={:,.3f} abs.T={:,.0f}".format(mach, t+Altitude_to_Temperature(a) ).replace("e-0", "e-")

                    mynote=green + 'Section {} ({}%) radius {:.5g}m chord {:.5g}m rpm {:,.0f} has {} re={:,.0f} zeroAoA={:.3g}degrees. Formula:{}{}'.format(s, int(100*localok['u']), section_radius, section_chord, rpm, mmsg, Re, zero_aoa, formres,norm)
                    #eprint(localok)
                    #eprint(formres)

                    if medium[-5:] == 'Water': # All 3 that end with "Water"
                        mynote=mynote + " in {:.5g}C temp {}".format(t, medium)
                    else:
                        mynote=mynote + " in {:.5g}C temp_offs at {:.5g}m altitude in {}".format(t, a, medium)

                    if dlg.get('rotate_direction'+aft,0) and dlg['rotate_direction'+aft].selectedItem.index: rotate_direction = -1 * zdir
                    else: rotate_direction = 1 * zdir                                 # 0 = default = left

                    extrarotate=-zero_aoa-manualaoa
                    if turbine: flip=-1; myrot=-1                               # Turbines differ from props, because they're pushed, instead of doing the pushing - these adjustments apply that physical difference
                    else: flip=1; myrot=0; extrarotate=-extrarotate

                    
                    if not section_chord: # insert a point instead of a foil (e.g. the tip)
                        tail=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + rotate_direction*scimitar*100 + rotate_direction*move_tail*100 ,                                       center_point.geometry.y + hedral*100 , zdir*section_radius*100) ) # Insert a point to indicate the tail, and where the rotation-join will be (not using hub_edge since it might be from wrong sketch)
                        sectionjson={}
                        sectionjson['points']=adsk.core.ObjectCollection.create() 
                        sectionjson['points'].add(tail)
                        #if s>0: sectionjson['pitch']=section[s-1]['foil']['pitch'] else: sectionjson['pitch']=0
                        sectionjson['pitch']=0

                    else:
                        #if s==12: # debug data
                        #    eprint("localok s=12:")
                        #    eprint(localok)

                        section_foil=bestFoil( Re=Re, N=N, strut=False, TextInfo=True, note=mynote, sharejson=sharejson, liftamt=liftamt, Mach=mach, Cloud=cloud_foil, WantS=0) # Find foil that is the closest Re match to the one we want

                        #try:
                        nose=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + rotate_direction*scimitar*100 + rotate_direction*move_tail*100 + rotate_direction * section_chord*100, center_point.geometry.y + hedral*100 , zdir*section_radius*100) ) # And the nose, to its left
                        tail=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + rotate_direction*scimitar*100 + rotate_direction*move_tail*100 ,                                       center_point.geometry.y + hedral*100 , zdir*section_radius*100) ) # Insert a point to indicate the tail, and where the rotation-join will be (not using hub_edge since it might be from wrong sketch)
                        join=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + rotate_direction*scimitar*100                                  ,                                       center_point.geometry.y + hedral*100 , zdir*section_radius*100) ) # Insert a point to indicate the tail, and where the rotation-join will be (not using hub_edge since it might be from wrong sketch)
                        #except:
                        #    err(sharejson,program_name+' insertTurbineUI failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox

                        sectthick=minthick/section_chord

                        # below, trying to use X axis radius has two problems... insertFoil can't wrap this way, and insertFoil can't do rotation joins properly this way either.
                        #tail=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + section_radius*100 , 0.0, center_point.geometry.y) ) # Insert a point to indicate the tail, and where the rotation-join will be (not using hub_edge since it might be from wrong sketch)
                        #nose=sketchPoints.add( adsk.core.Point3D.create( center_point.geometry.x + section_radius*100 , 0.0, center_point.geometry.y + rotate_direction * section_chord*100 ) ) # And the nose, to its left
                        #sectthick=minthick/section_chord


                        # insert the turbine foil sections here
                        sectionjson={}
                        insertFoil( sketch=sketch, 
                                    nose=nose,              # dlg['nose'+aft].selection(0).entity
                                    tail=tail,              # dlg['tail'+aft].selection(0).entity
                                    join=join,              # join usually = tail; Always rotate about the tail, so the trailing edge is a straight line
                                    usefoil=section_foil,   # callee uses -section_foil['alpha'],
                                    rotate=myrot,           # Flag to apply auto-rotation (0=yes, 360=no, -1=yes backwards) # -1 for turbines works
                                    flip=flip,              # other way for turbines (get pushed) .vs. propellers (do the pushing) #-1 for turbines works
                                    extrarotate=extrarotate,# Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
                                    wrapline=center_line,   # Center of a cylinder that the foil will be bent around
                                    wrapstart=hub_nose,     # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                                    endscale=0,             # Not end-capping this
                                    textInfo=True,          # Tell the user what we did in the text info area (and, for errors, with UI popups)
                                    constrain=False,
                                    camber=camber,          # Also compute and insert a camber line, with extended ends for mold parting surfaces
                                    minthick=sectthick,     # ensure the foil can be printed (stop T/E being 0 wide)
                                    splinetype=0,           # 0=fitpoint, 1=controlpoint
                                    #interpolate=100,       # 0=do not change, 1=expand
                                    equispaceN=num_points,  # How many points to have
                                    equispaceS=nt_concentrate, # how much to scrunch up the points at the nose and tail
                                    dlg=dlg,                # So we can add metadata (attributes) to the foil as well
                                    patch=False,            # Add a patch to it, returning the patch object inside sharejson['patch']
                                    xtilt=xtilt,            # rotate this many degrees about the X axis
                                    sharejson=sectionjson)  # sharejson is output

                    section[s]['foil']=sectionjson      # overwrite the existing foil info with more detailed stuff
                    #sharejson['sections'].append(sectionjson) # so we can in future do parametric things with this

                    # Compute blade lift/drag from cl/cd here...
                    if s>0:
                        area = ( section[s]['section_radius'] - section[s-1]['section_radius'] ) * ( section[s]['chord'] + section[s-1]['chord'] ) / 2 
                        eprint("area={}".format(area))
                        # section[4]['chord'] = 0.15
                        # section[4]['section_radius'] = 0.09375

                    #Collect points for the rail splines
                    if num_rails and sectionjson.get('points'):
                        spoints=(sectionjson['points'].count-1) # ERROr - key error 'points'
                        for rail in range(0,num_rails):
                            railidx=int(rail*spoints/num_rails)
                            #railidx=spoints?
                            eprint("spoints={} ridx={}".format(spoints,railidx))
                            rail_points[rail].add(sectionjson['points'].item(railidx))
                        if minthick>0:
                            rail_points[num_rails].add(sectionjson['points'].item(spoints))

                    eprint("s={} flip={} rot={} extrarot={} ZeroAoa={} foilAoA={} pitch={} wrapped={}".format(s, flip, myrot, extrarotate, section[s]['zero_aoa'],section[s]['aoa'],sectionjson['pitch'],sectionjson['wrapdeg']))
                    #te_points.add(sharejson['points'].item(0)) # Trailing edge point
                    #midpt=int((sharejson['points'].count-1)/2)
                    #le_points.add(sharejson['points'].item(midpt)) # Leading edge point

                    #if station<0.99:
                    #    station=station+0.1 # Bunch the hub so we can loft the tight corner
                    #    if station>0.99:
                    #        station=1 
                    #else:
                    #    station=station+1


                    # Need to prep to patch here for not-sure-what reason...
                    if patchme: #  and (s<=5 or s==(num_profiles-1)) and zz==123:
                        try:
                            patches = sketch.parentComponent.features.patchFeatures # patches = rootComp.features.patchFeatures
                            section[s]['foil']['patchInput'] = patches.createInput(section[s]['foil']['foilobj'], adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                            # section[s]['foil']['mytail']
                            section[s]['foil']['pathInput'] = adsk.fusion.Path.create(section[s]['foil']['foilobj'], 1) # For the subsequent loft 1=connectedChainedCurves
                        except:
                            eprint("patch prep failed.{}".format(traceback.format_exc()))
                            pass

                    ''' loft prep


        # Create profile 3
        ctorPlaneInput2 = ctorPlanes.createInput()
        ctorPlaneInput2.setByOffset(ctorPlane1, offset)
        ctorPlane2 = ctorPlanes.add(ctorPlaneInput2)
        sketch2 = sketchesObj.add(ctorPlane2)
        sketchCirclesObj2 = sketch2.sketchCurves.sketchCircles
        sketchCirclesObj2.addByCenterRadius(centerPoint, 10.0)
        profile2 = sketch2.profiles.item(0)

        # Create loft feature input
        loftFeats = rootComp.features.loftFeatures
        loftInput = loftFeats.createInput(adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
        loftSectionsObj = loftInput.loftSections
        loftSectionsObj.add(profile0)
        loftSectionsObj.add(profile1)
        loftSectionsObj.add(profile2)
        loftInput.isSolid = False

                    #
                    #
                    #loftFeats = design.rootComponent.features.loftFeatures
                    #loftInput = loftFeats.createInput(adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                    #loftSectionsObj = loftInput.loftSections
                    ##loftSectionsObj.add(section[11]['foil']['pathInput'])
                    #loftInput.isSolid = False
                    loftFeats.add(loftInput)
                    '''



        sharejson['sections']=section # .append(sectionjson) # so we can in future do parametric things with this

        if num_rails:
            sharejson['rails']=[]
            sharejson['railpath']=[]
            for rail in range(0,num_rails):
                rail_line=sketch.sketchCurves.sketchFittedSplines.add(rail_points[rail]) 
                sharejson['rails'].append(rail_line.entityToken) # so we can in future do parametric things with this
                try:
                    sharejson['railpath'].append( adsk.fusion.Path.create( rail_line, 0)) # For the subsequent loft 0=noChainedCurves # can ERR
                except:
                    eprint('rail ' + str(rail) + ' failed:-\n{}'.format(traceback.format_exc()))
            if minthick>0: # Also put a rail on the minthick topside
                rail_line=sketch.sketchCurves.sketchFittedSplines.add(rail_points[num_rails]) 
                sharejson['rails'].append(rail_line.entityToken) # so we can in future do parametric things with this
                try:
                    sharejson['railpath'].append( adsk.fusion.Path.create( rail_line, 0)) # For the subsequent loft
                except:
                    eprint('Rail ' + str(rail) + ' failed:-\n{}'.format(traceback.format_exc()))
                # section[s]['foil']['pathInput'] = adsk.fusion.Path.create(section[s]['foil']['foilobj'], 0) # For the subsequent loft

            rail_line.attributes.add("attr"+aft,"JSON", shareIt(sharejson,dlg=dlg,Text=True) ) # Include everytihng about this foil into the object attributes too


        # patch logic: create the loft (keep trying sections away from the hub until one works), then patch those ends - if that fails, delete the loft and find a patch that works, then try to loft from that
        if 1 and patchme: # almost works - except stitch
            pdone=0     # Counts up to 2 - the number of patches we ultimately want to create
            ps=1        # patch surface counter - starts at 1 (hub)
            hubnum=1    # gets set to the first profile near the hub that patched OK
            surfaces = adsk.core.ObjectCollection.create()          # For a later stitch
            patchFeats = design.rootComponent.features.patchFeatures            # For creating, and deleting when failed.
            loftFeats = design.rootComponent.features.loftFeatures  # For creating, and deleting when failed.
            myloft = None
            myerr=''

            retryit = 1 # for the while loop - we're going to un-do stuff that fails until we get a fully working result


            # Loft the sections first. loft is more sensitive than patch, so should fail less often on tight bendy hubs
            while retryit:
                eprint("Patching.")
                eprint("Patching. ps={} pdone={} hubnum={} myloft={}".format(ps, pdone, hubnum, myloft))
                retryit=0
                if myloft is None:

                    loftInput = loftFeats.createInput(adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                    loftSectionsObj = loftInput.loftSections
                    loftRailsObj = loftInput.centerLineOrRails

                    for ps in range(num_profiles-1,hubnum-1,-1):    # num_profiles-1 because it's already had 1 added to it.  hubnum-1 because python skips the last one in a range
                        if section[ps]['foil'].get('pathInput'):    # could be missing if creating it was an exception earlier
                            loftSectionsObj.add(section[ps]['foil']['pathInput'])

                    for rp in sharejson['railpath']:
                        loftRailsObj.addRail(rp)

                    loftInput.isSolid = False
                    lcount=loftFeats.count
                    try:
                        myloft=loftFeats.add(loftInput) # create the loft - # this might throw an exception
                        sharejson['loft_entity']=myloft.entityToken
                    except:
                        myerr=traceback.format_exc()
                        eprint('loft failed:-\n{}'.format(myerr))
                        if lcount < loftFeats.count: # something broken got created
                            dudloft=loftFeats.item(loftFeats.count-1) # should check healthState <> 0 or dudloft.errorOrWarningMessage isn't empty
                            dudloft.deleteMe()
                            hubnum=hubnum+1
                            if hubnum >= num_profiles-1:
                                retryit = 0 # give up
                            else:
                                retryit = 1 # try next section

                # Patch the ends next
                ps=hubnum
                while myloft and pdone<2 and ps<=(num_profiles-1): # no endless loops please
                    eprint("patcihng {} (w={}) pdone={}...".format(ps, sharejson['sections'][ps]['foil']['wrapdeg'], pdone))
                    pcount=patchFeats.count
                    try:
                        section[ps]['foil']['patch']=patchFeats.add(section[ps]['foil']['patchInput']) # this might throw an exception
                        #section[ps]['foil']['patch']=patchFeats.add(section[ps]['foil']['pathInput']) # this might throw an exception
                        section[ps]['foil']['patch_entity']=section[ps]['foil']['patch'].entityToken
                        pdone=pdone+1
                        surfaces.add(section[ps]['foil']['patch'].bodies.item(0)) # for a subsequent stitch
                        ps=ps+1
                        if ps<num_profiles-1: ps=num_profiles-1 # skip to the tip as soon as a hub one works
                    except:
                        myerr=traceback.format_exc()
                        eprint('Trying next profile because patching profile {} failed:\n{}'.format(ps, myerr)) 
                        if myloft:
                            eprint('deleting unpatchable but otherwise working loft')
                            try:
                                myloft.deleteMe()   # sometimes fails again
                            except:
                                pass
                            myloft=None
                            retryit=1
                        hubnum=hubnum+1
                        ps=ps+1 # try again, with the next one (when the hub fails, rey again, working out to the tip until one works...

                        if pcount < patchFeats.count: # something broken got created
                            dudpatch=patchFeats.item(patchFeats.count-1)
                            dudpatch.deleteMe()
                            if hubnum >= num_profiles-1:
                                retryit = 0 # give up

                sharejson['hubnum']=hubnum # in case something later needs to know

            # Stitch it all togther into a solid now
            if myloft:
                surfaces.add(myloft.bodies.item(0))
        
                # Define tolerance with 1 cm.
                tolerance = adsk.core.ValueInput.createByReal(1.0)
        
                # Create a stitch input to be able to define the input needed for an stitch.
                stitches = design.rootComponent.features.stitchFeatures
                stitchInput = stitches.createInput(surfaces, tolerance, adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
        
                # Create a stitch feature.
                try:
                    stitch = stitches.add(stitchInput)
                    sharejson['stitch_entity']=stitch.entityToken
                except:
                    vlog('Auto-stitch failed, sorry:\n{}'.format(traceback.format_exc()))

            else:
                vlog('Unable to auto-loft or patch ends, sorry:-\n{}'.format(myerr))




        # (old way) Patch the ends first
        if 0 and patchme:
            pdone=0     # Counts up to 2 - the number of patches we ultimately want to create
            ps=4        # patch surface counter - starts at 1 (hub)
            # temp - 3  - coz 1 and 2 do not loft propper
            hubnum=ps    # gets set to the first profile near the hub that patched OK
            surfaces = adsk.core.ObjectCollection.create()          # For a later stitch
            patchFeats = design.rootComponent.features.patchFeatures            # For creating, and deleting when failed.
            loftFeats = design.rootComponent.features.loftFeatures  # For creating, and deleting when failed.
            myloft = None
            myerr=''

            while pdone<2 and ps<=(num_profiles-1): # no endless loops please
                vlog("patcihng {} (w={})...".format(ps, sharejson['sections'][ps]['foil']['wrapdeg']))
                try:
                    #patches = sketch.parentComponent.features.patchFeatures # 
                    #patchInput = patches.createInput(section[ps]['foil']['foilobj'], adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                    #section[ps]['foil']['patchInput']=patchInput
                    #section[ps]['foil']['patch']=patches.add(patchInput)
                    section[ps]['foil']['patch']=patches.add(section[ps]['foil']['patchInput'])
                    section[ps]['foil']['patch_entity']=section[ps]['foil']['patch'].entityToken
                    if hubnum==0:
                        hubnum=ps # Remember the closest we could get to the hub that worked.
                    domex=1
                    pdone=pdone+1
                    if domex:
                        surfaces.add(section[ps]['foil']['patch'].bodies.item(0)) # for a subsequent stitch
                    ps=ps+1
                    if ps<num_profiles-1:
                        ps=num_profiles-1
                except:
                    # Need to feature.deleteMe() now... # ff=design.rootComponent.features
                    vlog('Trying next profile because patching profile {} failed:\n{}'.format(ps, traceback.format_exc()))
                    ps=ps+1 # try again, with the next one (when the hub fails, work out to the tip until one works...
                    pass # avoid the below...
                    # RuntimeError: 5 : SurfacePatch12 / 1 Reference Failures   The operation failed.     Try adjusting the values or changing the input geometry. //   The operation failed.
                    # Try adjusting the values or changing the input geometry. / Compute Failed // Modeling Error: Can't create patch toolbody, please check the inputs.

            # Loft the sections next - bzzt - this first? loft is more sensitive than patch - either may fail though...
            if hubnum>0: # worked
                loftInput = loftFeats.createInput(adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                loftSectionsObj = loftInput.loftSections
                loftRailsObj = loftInput.centerLineOrRails

                for ps in range(num_profiles-1,hubnum-1,-1):    # num_profiles-1 because it's already had 1 added to it.  hubnum-1 because python skips the last one in a range
                    if section[ps]['foil'].get('pathInput'):
                        loftSectionsObj.add(section[ps]['foil']['pathInput'])

                for rp in sharejson['railpath']:
                    loftRailsObj.addRail(rp)

                loftInput.isSolid = False
                myloft=loftFeats.add(loftInput) # create the loft

                # Stitch it all togther into a solid now
                surfaces.add(myloft.bodies.item(0))
        
                # Define tolerance with 1 cm.
                tolerance = adsk.core.ValueInput.createByReal(1.0)
        
                # Create a stitch input to be able to define the input needed for an stitch.
                stitches = design.rootComponent.features.stitchFeatures
                stitchInput = stitches.createInput(surfaces, tolerance, adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
        
                # Create a stitch feature.
                stitch = stitches.add(stitchInput)

            else:
                vlog('Unable to patch, sorry:\n{}'.format(traceback.format_exc()))



        # Demos: 2000-up, 250 hub, 50 edge
        # Finally - move it all to where the user wanted it
        #sk1 = root.sketches.add(root.xYConstructionPlane)
        #coll = adsk.core.ObjectCollection.create()
        #for crv in sk1.sketchCurves:
        #    coll.add(crv)
        #for pnt in sk1.sketchPoints:
        #    coll.add(pnt)
        #transform = adsk.core.Matrix3D.create()
        #transform.translation = adsk.core.Vector3D.create(0,0,2)
        #sk1.move(coll, transform)

        shareIt(sharejson,dlg=dlg) # Let us know what they did if they checked the box to allow it

        #le_line=sketch.sketchCurves.sketchFittedSplines.add(le_points) 
        #te_line=sketch.sketchCurves.sketchFittedSplines.add(te_points) # Normally a straight line at present
        remember(sharejson,pnlid)                                                 # remember their most recent settings, so we can offer then again in future

    # Error handling
    except:
        return err(sharejson,program_name+' insertTurbineUI failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox

    # insertTurbineUI





#################################################################################
####                                                                          ###
##                              Do the Work at last!!                          ##
####                                                                          ###
#################################################################################
def insert2DWing(pnlid,args,inputs):
    global VERSION, pnl, userdata, settings_file, foilinfo
    ui = None
    try:
        sharejson={}
        app = adsk.core.Application.get()
        ui = app.userInterface
        design = app.activeProduct

        panel=pnlid['mywhere'] + '_' + pnlid['aftid'] # 'sketch_airfoil_command_aft' + '_' + 'menuid':'create_cowling_aft'
        dlg=pnl[panel]

        if dlg.get('votebox'+aft,0) and dlg['votebox'+aft].value:               # Not a command
            return True
        if dlg.get('nickname'+aft,None) is None:                                # Not a command
            return True


        pnl['calcstart']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())


        # Set up some constants...
        cache={} # sketches.add() breaks user selections input, so cache selections now
        nickname=dlg['nickname'+aft].value                                      # Triggers inserting user parms if named
        medium=cmdh['medium'+aft]['choices'][dlg['medium'+aft].selectedItem.index] # 'Air', 'Fresh Water', 'Sea Water', 'Any Water', 'Other'
        nose=dlg['nose'+aft].selection(0).entity
        tail=dlg['tail'+aft].selection(0).entity
        join=None
        if dlg.get('join'+aft,0) and dlg['join'+aft].selectionCount:
            join=dlg['join'+aft].selection(0).entity
        else:
            eprint("No join chosen") # sometimes misses this?

        if dlg.get('wrap_axis'+aft) and dlg['wrap_axis'+aft].selectionCount: wrap_axis=dlg['wrap_axis'+aft].selection(0).entity
        else: wrap_axis=None

        if dlg.get('up'+aft,0) and dlg['up'+aft].selectedItem.index:            # Up is item 0, so, item 1 (true) is inverted
            (flip,dflip)=(-1,-1)                                                # dflip for user parms, flip for us (might be reversed if we rotated >90 degrees...)
        else:
            (flip,dflip)=(1,1) 

        cloud_foil=dlg.get('cloud_foil'+aft,0) and dlg['cloud_foil'+aft].value

        sketch = app.activeEditObject                                           # Get the current sketch?
        # Get current sketch, or, prepare to create new one if they're not in one right now (note: creating a new sketch breaks the UI input - erases their selection points)
        if not sketch or sketch.objectType != 'adsk::fusion::Sketch':
            rootComp = None
            try:
                if join: rootComp=join.parentSketch.parentComponent             # Try to create new sketches in the correct component
                if not rootComp: rootComp=nose.parentSketch.parentComponent
                if not rootComp: rootComp = design.rootComponent                # Default to the root component of the active design.
                sketches = rootComp.sketches                                    # Create a new sketch on the same plane as what the user selected the nose in
            except:
                rootComp = design.rootComponent                                 # Default to the root component of the active design.
                sketches = rootComp.sketches                                    # Create a new sketch on the same plane as what the user selected the nose in
            # CAUTION - next line breaks the users input
            try:
                sketch = sketches.add( nose.parentSketch.referencePlane )           # Use same plane as whatever their chosen nose was (XY, XZ, or YZ)
            except: # above line can fail with "not a planr entry"
                sketch = sketches.add( app.activeProduct.rootComponent.xYConstructionPlane )

        # project any foreign points into our current sketch (needed so the X,Y makes sense later)
        if nose.parentSketch.name != sketch.name:                               # project any foreign point into here # ERROR: 'ConstructionPoint' object has no attribute 'parentSketch'
            proj=sketch.project(nose)
            if proj.count: nose=proj.item(0)
        if tail.parentSketch.name != sketch.name:
            proj=sketch.project(tail)
            if proj.count: tail=proj.item(0)

        rotate=360                                                              # Flag so we know whether or not to rotate later (<360 means yes)
        if join:                                                                # Only rotate if they selected a join point
            rotate=0                                                            # Flag. 0 menas auto-rotate by AoA amount
            if join.parentSketch.name != sketch.name:
                proj=sketch.project(join)
                if proj.count:
                    join=proj.item(0)
                    join.isConstruction=True
        else:
            join=nose                                                           # Default to join to nose-point dlg['nose'+aft].selection(0).entity
            eprint("Not auto-rotating") # sometimes misses this?

        cache['nose'+aft]=nose
        cache['tail'+aft]=tail
        cache['join'+aft]=join

        constrain=dlg.get('constrain'+aft,0) and dlg['constrain'+aft].value     # Make this optional, since it is so slow.
        camber=dlg.get('camber'+aft,0) and dlg['camber'+aft].value
        strut=panel.find('strut') >= 0                                          # struts have the word "strut" in the panel name
        doendcap=dlg.get('endcap'+aft,0) and dlg['endcap'+aft].value            # Do they want rounded ends
        if doendcap and dlg.get('enddist'+aft,0): endscale=dlg['enddist'+aft].value/200      # How pointy to make the endcaps
        else: endscale=0

        Re=dlg['re']        # Inserted by the validate code
        N=dlg['N']          # Inserted by the validate code
        chord=dlg['chord']  # Inserted by the validate code
        mach=dlg['mach']    # Inserted by the validate code
        minthick=dlg['minthick'+aft].value / 100                                # in meters (e.g. 0.4mm = 0.0004)
        sectthick=minthick/chord                                                # Scale from desired-size down to unit size (which later gets scaled back up)

        # Find foil that is the closest Re match to the one we want
        usefoil=bestFoil( Re=Re, N=N, strut=strut, TextInfo=True, Mach=mach, Cloud=cloud_foil)    # Says . Best foil is a tested Re 1,400,000 n12 and named optbs_spline_sym_naca0007.dat-s17_n17s0_nsopt-Re620000-N7.dat. drag coefficient: 0.0029749 (=friction drag 0.00206 + pressure drag 0.00091)
        if userdata.get('text_info_aft',0) == 'Full':
            vlog("{} density: {} kg/M^3, dynamic_viscosity: {} Pa.s, kinematic_viscosity: {} cSt, Mach:{}".format(medium,dlg['density'+aft].value, dlg['dynamic_viscosity'+aft].value, dlg['kinematic_viscosity'+aft].value,mach))    # example of appropriate way of telling user extra geeky stuff

        # Insert computed and foil data into user parameters (before they get lost when we create a new sketch...)
        for parms in foilinfo:
            for pkey in parms:
                desc=parms[pkey]
                if desc: # blank means skip
                    pname=nickname + aft + '_' + pkey
                    #val = adsk.core.ValueInput.:createByString("5 mm")
                    valr = usefoil.get(pkey,0)
                    if isinstance(valr, (str, int)): valr=float(valr)
                    if isinstance(valr, (float)):                               # Skip "None" or other dud stuff
                        val = adsk.core.ValueInput.createByReal(valr)
                        unit=''
                        if pkey=='alpha':   # add text data into the comments
                            desc=desc + usefoil.get('Foil_fn','')
                            unit='deg'
                            val = adsk.core.ValueInput.createByString('{} deg'.format(valr))
                        if pkey=='xver':   # add text data into the comments
                            desc=desc + ', {} version:{:1.8f} db:{:1.8f}'.format(program_name,VERSION,float(foildb2020.VERSION(0)))
                        #sharejson[pkey]=valr # Now done in shareDlg/
                        if nickname:
                            if design.userParameters.itemByName(pname):
                                eprint("Not adding user parameter {} because it already exists in the parameters".format(pname))
                            else:
                                design.userParameters.add(pname, val, unit, desc)
                    else:
                        eprint("Not adding user parameter {} because it already exists in the parameters".format(pname))

        if nickname:
            pname=nickname +aft+ '_chord' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByString('{} m'.format(chord)), 'm', 'straight distance from nose to tail')
            pname=nickname +aft+ '_user_Re' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(Re), '', 'Your computed Reynolds Number')
            pname=nickname +aft+ '_user_Mach' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(mach), '', 'Your computed Mach Number')
            pname=nickname +aft+ '_user_ncrit' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(N), '', 'Your computed log(critical angle)')
            pname=nickname +aft+ '_user_flip' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(dflip), '', 'Lift(1) or Downforce(-1)')
            pname=nickname +aft+ '_user_ver' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(VERSION), '', program_name+' version number')
            pname=nickname +aft+ '_user_dbver' # Add this computed value separately
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal(foildb2020.VERSION(0)), '', program_name+' foil database version number')

        shareDlg(sharejson,dlg,cache=cache,nickname=nickname) # Copy all form data into sharejson
        # Save user-entered numbers inside the design data for future use
        for userinfo in cmdl:
            if userinfo.get('id',0) and dlg.get(userinfo['id'],0) and userinfo.get('type','') and userinfo.get('name','') and dlg[userinfo['id']].isVisible:
                if userinfo['type']=='ValueInput':
                    valr = dlg[userinfo['id']].value
                    if isinstance(valr, (str, int)): valr=float(valr)
                    val = adsk.core.ValueInput.createByReal(valr)
                    if nickname:
                        if design.userParameters.itemByName(pname):
                            eprint("Not adding user parameter {} because it already exists in the parameters".format(pname))
                        else:
                            design.userParameters.add(pname, val, userinfo.get('unit',''), userinfo.get('tip',''))


        # Name the sketch for them:
        if nickname and sketch.name[:6]=='Sketch':
            sketch.name=nickname+'_foil'                                        # Name their sketch for them, if they've not already done that themselves

        num_points=60; nt_concentrate=10 # defaults
        if dlg.get('num_points'+aft,0) and dlg['num_points'+aft].value > 0: num_points=int( dlg['num_points'+aft].value )
        if dlg.get('nt_concentrate'+aft,None) is not None: nt_concentrate=int( dlg['nt_concentrate'+aft].value )
        if nt_concentrate<=1: nt_concentrate= 1.000001 # 1 breaks it - this works 

        # Perform the 2D wing/strut insert
        #vlog("Using {} splinetype {}".format(num_points,usefoil.get('s',0))); 60 type 1
        insertFoil( sketch=sketch, 
                    nose=nose,                  # dlg['nose'+aft].selection(0).entity
                    tail=tail,                  # dlg['tail'+aft].selection(0).entity
                    join=join,
                    usefoil=usefoil,
                    rotate=rotate,              # = means AoA, -1 mean -AoA, 360 means no.
                    flip=flip,    
                    # extrarotate=0,            # Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
                    wrapline=wrap_axis,         # Center of a cylinder that the foil will be bent around
                    wrapstart=join,             # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                    endscale=endscale,          # pointyness of the end caps
                    textInfo=True,              # Tell the user what we did in the text info area (and, for errors, with UI popups)
                    constrain=constrain,        
                    camber=camber,              # Also compute and insert a camber line, with extended ends for mold parting surfaces
                    minthick=sectthick,         # ensure the foil can be printed (stop T/E being 0 wide)
                    splinetype=usefoil.get('s',0), # 0=fitpoint, 1=controlpoint
                    equispaceN=num_points,      # How many points to have
                    equispaceS=nt_concentrate,  # how much to scrunch up the points at the nose and tail
                    dlg=dlg,                    # So we can add metadata (attributes) to the foil as well
                    sharejson=sharejson) 

        if sharejson.get('points',0):           # Not serializable
            del sharejson['points']

        # Return everything about this foil to the caller
        sharejson['chord']=chord
        sharejson['user_Re']=Re
        sharejson['user_Mach']=mach
        sharejson['user_ncrit']=N
        sharejson['user_flip']=dflip
        sharejson['nickname']=nickname
        sharejson['ver']=VERSION
        sharejson['dbver']=foildb2020.VERSION(0)
        sharejson['count']=1+ userdata.get('count',0)
        userdata['count']=sharejson['count']
        sharejson['now']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())
        sharejson['panel']=panel
        for i in ('calcstart', 'opened', 'loaded', 'openc'):
            sharejson[i]=pnl[i]


        shareIt(sharejson,dlg=dlg) # Let us know what they did if they checked the box to allow it

        userdata['count']=sharejson['count']

        # Insert metadata that our re-computer might need into user parms
        if nickname:
            pname=nickname + aft + '_json'
            if not design.userParameters.itemByName(pname): design.userParameters.add(pname, adsk.core.ValueInput.createByReal( sharejson['count'] ), '', json_dumps(sharejson) )

        #userdata['last']=json_dumps(sharejson) # remember their most recent settings, so we can offer then again in future
        #save_user_settings(ui,settings_file) 
        remember(sharejson,pnlid)                                                 # remember their most recent settings, so we can offer then again in future

        # Replace this if we decide in future not to hook all events
        if 0:
            if web_nbreq(None): maxt=2
            else: maxt=0 # proccess any non-blocking operations, wating upto max 2s
            while maxt>0 and web_nbreq(None): # proccess any non-blocking operations
                eprint('waiting {}'.format(maxt))
                time.sleep(0.25)
                maxt=maxt-0.25

    # Error handling
    except:
        return err(sharejson,program_name+' insert2DWing failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox # ERROr: 'sharejson' referenced before assignment

    #insert2DWing






#################################################################################
####                                                                          ###
##                           The real work gets done here                      ##
####                                                                          ###
#################################################################################
def insertFoil( sketch=None,        # Works only in X,Y plane: ignores Z (except if wrapping)
                nose=None,          # dlg['nose'+aft].selection(0).entity
                tail=None,          # dlg['tail'+aft].selection(0).entity
                join=None,          # point or line (will add Cm if line) to join upon
                usefoil=None,       # from bestFoil()
                rotate=None,        # =360 means none. =0 means AoA. any negative means -AoA.  >360 means that specific angle.  Gets added to extrarotate.
                flip=None,          # Lift or downforce?
                extrarotate=0,      # Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
                wrapline=None,      # Center of a cylinder that the foil will be bent around
                wrapstart=None,     # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                endscale=None,      # pointyness of the end caps
                textInfo=None,      # Tell the user what we did in the text info area (and, for errors, with UI popups)
                constrain=None,     #
                camber=None,        # Also compute and insert a camber line, with extended ends for mold parting surfaces
                minthick=0,         # Gap to have at the trailing edge
                splinetype=0,       # 0=fitpoint, 1=controlpoint (-s 0 or -s 1)
                interpolate=60,     # 0=do not change, 1=expand
                equispaceN=0,       # Number of points to re-spline this shape into
                equispaceS=0,       # Nose concentration factor for equispace
                dlg=None,           # So we can add metadata (attributes) to the inserted foil as well
                patch=None,         # Create a surface patch over this foil
                xtilt=0,            # rotate this many degrees about the X axis
                sharejson=None):    # Return info
    global VERSION,foilinfo

    try:

        (nosex,nosey,nosez)=(nose.geometry.x, nose.geometry.y, nose.geometry.z)
        (tailx,taily,tailz)=(tail.geometry.x, tail.geometry.y, tail.geometry.z)
        (joinx,joiny,joinz)=(nose.geometry.x, nose.geometry.y, nose.geometry.z)     # will be updated later if a join was chosen
        scale=nose.geometry.distanceTo(tail.geometry) # OK

        # Create an object collection for the points.
        points = adsk.core.ObjectCollection.create() 
        xyzpoints=[] # copy of points, in pure python only
        sketchPoints = sketch.sketchPoints
        dim = sketch.sketchDimensions

        joindx=0    # So we can constrain the CM to the nose later
        joindy=0
        joinzz=tailz
        radius=0

        #vlog("insertFoil: sketch:{} nose:{} tail:{} join:{} scale:{} usefoil:{} rotate:{} flip:{} extrarotate:{} wrapline:{} wrapstart:{} endscale:{} textInfo:{} constrain:{} camber:{} minthick:{} splinetype:{} interpolate:{} equispaceN:{} equispaceS:{} dlg:{} patch:{} xtilt:{} sharejson:{}".format( sketch, nose, tail, join, scale, usefoil, rotate, flip, extrarotate, wrapline, wrapstart, endscale, textInfo, constrain, camber, minthick, splinetype, interpolate, equispaceN, equispaceS, dlg, patch, xtilt, sharejson ))
        #vlog("insertFoil:  scale:{} usefoil:{} rotate:{} flip:{} extrarotate:{} wrapstart:{} endscale:{} textInfo:{} constrain:{} camber:{} minthick:{} splinetype:{} interpolate:{} equispaceN:{} equispaceS:{} patch:{} xtilt:{} sharejson:{}".format( scale, usefoil, rotate, flip, extrarotate, wrapstart, endscale, textInfo, constrain, camber, minthick, splinetype, interpolate, equispaceN, equispaceS, patch, xtilt, sharejson ))


        CM = usefoil.get('Cm',0)
        if isinstance(CM, (str, int)): CM=float(CM)
        CM=0.25-CM # CM is relative to quarter chord

        # Decide where to join and if to rotate by AoA or not (based on users join point selection)
        if rotate<360 or extrarotate: # rotate=0 Indicates they selected a join point or line
            #eprint("Join chosen:{}".format(rotate))
            if rotate<0:
                rotate = usefoil.get('alpha',0)
                if isinstance(rotate, (str, int)): rotate=float(rotate)
                rotate = -rotate # if caller was negative, it means they want a rotate the other way
            else:
                rotate = usefoil.get('alpha',0)
                if isinstance(rotate, (str, int)): rotate=float(rotate)

            #if join.objectType=='adsk::fusion::SketchLine':
            if join.objectType!='adsk::fusion::SketchPoint': 
                (joinstart, joinend) = (join.startSketchPoint, join.endSketchPoint)
                if joinstart.geometry.distanceTo(nose.geometry) > joinend.geometry.distanceTo(nose.geometry): # use their nose selection as the correct orientation to the real front, regardless of what the "start" of the input line might be
                    (joinstart, joinend) = (joinend, joinstart)
                if abs(joinstart.geometry.x-nosex) > 0.000001 or abs(joinstart.geometry.y-nosey) > 0.000001 or abs(joinend.geometry.x-tailx) > 0.000001 or abs(joinend.geometry.y-taily) > 0.000001:
                    if textInfo:
                        vlog('** CAUTION: Your join line is in a different location to your chord line - insertion point may be unreliable **')
                joindx=CM*(joinend.geometry.x - joinstart.geometry.x ) 
                joindy=CM*(joinend.geometry.y - joinstart.geometry.y ) 
                joinzz=joinend.geometry.z
                join=sketchPoints.add( adsk.core.Point3D.create( joinstart.geometry.x + joindx, joinstart.geometry.y + joindy, joinzz) ) # Insert a point to indicate where the rotation-join is, being the CM...

            (joinx,joiny,joinz)=(join.geometry.x, join.geometry.y, join.geometry.z)     # will be nose unless otherwise specified

        if ((tailz-nosez)**2 + (tailx-nosex)**2 + (taily-nosey)**2 )**0.5 :
            nose_tail_angle = math.asin(  (taily-nosey) /   ((tailz-nosez)**2 + (tailx-nosex)**2 + (taily-nosey)**2 )**0.5 )*180/math.pi # seems OK # ERROr - /0
        else:
            nose_tail_angle = 0




        # 3D atan2() :-
        H = ((tailz-nosez)**2 + (tailx-nosex)**2 + (taily-nosey)**2 )**0.5
        if (tailx-nosex) < 0:
            nose_tail_angle = math.asin((taily-nosey)/H) - math.pi/2
        else:
            nose_tail_angle = math.asin(-(taily-nosey)/H) + math.pi/2

        nose_tail_angle = -nose_tail_angle * 180/math.pi + 90


        #eprint("\x1b[32;1m Nose angle now={}  OK2D={} \x1b[0m".format(nose_tail_angle,math.atan2(taily - nosey, tailx - nosex) * 180 / math.pi))
        
        if nose_tail_angle > 90 or nose_tail_angle < -90:
            flip=flip*-1 # we rotate to orient it the way they want, which means we need to flip if we go past vertical (while still respecting their origianl flip request)

        xy=usefoil['xy'].copy() # get the points
        #xy=[*usefoil['xy']]                                     # for normalize

        #eprint("insertFoil: rotate {} + {} = {}. nta={} joinzz={}. nose({},{},{}) tail({},{},{})".format(rotate, extrarotate, rotate + extrarotate, nose_tail_angle, joinzz, nose.geometry.x, nose.geometry.y, nose.geometry.z, tail.geometry.x, tail.geometry.y, tail.geometry.z  ))
        #eprint(xy)
        #vlog("insertFoil: rotate {} + {} = {}. nta={} joinzz={}. nose({},{},{}) tail({},{},{})  interpolate{} splinetype{} xy{} ".format(rotate, extrarotate, rotate + extrarotate, nose_tail_angle, joinzz, nose.geometry.x, nose.geometry.y, nose.geometry.z, tail.geometry.x, tail.geometry.y, tail.geometry.z, interpolate, splinetype, xy  ))


        if interpolate:
            if(splinetype): # can't sensibly change control points, so, convert them to fit-points first.
                (splinex,spliney)=jspline(-9999,splinetype,splinetype,1,points_to_xy(xy)) # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)
                splinetype=0 # switch to control-points now
                xy=xy_to_points(splinex,spliney) # Because we changed the splinetype
                # usefoil['s']=0
            else:
                (splinex,spliney)=jspline(-interpolate*2,splinetype,splinetype,1,points_to_xy(xy)) # expand them (at least by double - more if needed - see jspline)
            (splinex,spliney)= equispace(interpolate,10, splinetype,xy_to_points(splinex,spliney) ) # Convert our input spline into 60 fitpoints, with concentration-at-nose factor of 10
        else:
            (splinex,spliney)=points_to_xy(xy)

        if wrapline and splinetype: # can't wrap control points - the Z goes all wonky.  Convert to fit-point.
            (splinex,spliney)=jspline(-9999,splinetype,splinetype,1,points_to_xy(xy)) # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)
            splinetype=0 # switch to control-points now
            xy=xy_to_points(splinex,spliney) # Because we changed the splinetype
            minp=int((len(xy)+1)/2)
            if minp < 60: minp=60
            (splinex,spliney)= equispace(minp,10, splinetype,xy_to_points(splinex,spliney) ) # Convert our input spline into 60 fitpoints, with concentration-at-nose factor of 10
        elif equispaceN:
            eprint("Equispaced before x={} y={}".format(*points_to_xy(xy)))
            (splinex,spliney)=jspline(-9999,splinetype,splinetype,1,points_to_xy(xy)) # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)
            (splinex,spliney)= equispace(equispaceN, equispaceS, splinetype,xy_to_points(splinex,spliney) ) # Convert our input spline into 60 fitpoints, with concentration-at-nose factor of 10
            eprint("Equispaced after x={} y={}".format(splinex,spliney))

        if minthick>0:
            (splinex,spliney)= enforceThickness(minthick, xy_to_points(splinex,spliney) ) # move points away from each other if they are closer than minthick
            #Caller does the scale (splinex,spliney)= enforceThickness(minthick/scale*100, xy_to_points(splinex,spliney) ) # move points away from each other if they are closer than minthick

        splinez = [0.0] * len(splinex) # Might get used if wrapping around a cyl

        # join (kinda more rounded, but skips a few points at the tail)
        # (splinex,spliney)=jspline(2,0,0,0,points_to_xy(xy)) # expand them not too much (first 2 is how much)
        #if debug: vlog("mid insertFoil: X={} Y={}".format(splinex,spliney))

        #eprint("pline 1 l={}".format(len(splinex)))

        if constrain:
            shifto=1/8+1/16+1/32+1/64 # hopefully a distance that is extremely unlikely to exactly coincide with any existing geometry in the users sketch (to prevent auto-constraints appearing)
            shifto=0 # Disable this now
        else:
            shifto=0

        middle=int((len(splinex)-1)/2)

        # Find approximate camber Y point to drop the CM onto
        cmx=CM
        cmy=0
        for i in range(middle,len(splinex)): # find closest bottom point
            if splinex[i]>cmx:
                cmy=spliney[i-1]
                break
        for i in range(0,middle): # find top
            if splinex[i]<cmx:
                cmy=flip * (cmy + 0.5*abs(cmy-spliney[i-1]))
                break
        # move the CM to match the foil
        if nose_tail_angle: # put it the way they asked for next
            (x,y)=(cmx,cmy)
            cmx= x * math.cos(-nose_tail_angle*math.pi/180) + y*math.sin(-nose_tail_angle*math.pi/180) 
            cmy= -x * math.sin(-nose_tail_angle*math.pi/180) + y*math.cos(-nose_tail_angle*math.pi/180)
        if ( rotate>0 and rotate<360 ) or extrarotate:  # tilt to to the optimal AoA first
            (x,y)=(cmx+(nosex-joinx)/scale,cmy+(nosey-joiny)/scale) # test rotate - negative is anticlockwise.
            cmx=  x * math.cos(flip*(rotate+extrarotate)*math.pi/180) + y * math.sin(flip*(rotate+extrarotate)*math.pi/180) -(nosex-joinx)/scale
            cmy= -x * math.sin(flip*(rotate+extrarotate)*math.pi/180) + y * math.cos(flip*(rotate+extrarotate)*math.pi/180) -(nosey-joiny)/scale
        #if not wrapline: # why?
        cmpt=sketchPoints.add(adsk.core.Point3D.create( cmx*scale+nosex+shifto, cmy*scale+nosey+shifto, joinzz))
        #eprint('rotate={} nose_tail_angle={} cmx={} cmy={} extra={} total={}'.format(rotate,nose_tail_angle,cmx,cmy,extrarotate,(rotate+extrarotate)))

        controlPoints=[]

        # Prepare constants so we can wrap this if they want
        if wrapline:

            if wrapline.startSketchPoint.geometry.z != wrapline.endSketchPoint.geometry.z:
                return err(sharejson,"Sorry - wrapping is only supported on flat planes.  Adjust your sketch so your foil nose+tail and join are on a sketch XY plane, and your axis is on a DIFFERENT XY sketch plane, and try again (e.g. a 3D sketch, or two sketches on different planes)",dlg)

            if tail.geometry.x != nose.geometry.x:
                radius=wrapline.startSketchPoint.geometry.z - joinzz
            #elif wrapline.startSketchPoint.geometry.x == joinzz:                # TODO - make same XY sketch wrapping work using Z...
            #    radius=join.geometry.x - wrapline.startSketchPoint.geometry.x 
            else:
                radius=wrapline.startSketchPoint.geometry.x - joinzz
                joinzz = joinzz + wrapline.startSketchPoint.geometry.z
            #if debug: vlog("radius={} wrap: ({}, {}, {}) joinzz={}  nose:({}, {}, {})  tail:({}, {}, {})  join:({}, {}, {}) ".format( radius, wrapline.startSketchPoint.geometry.x , wrapline.startSketchPoint.geometry.y, wrapline.startSketchPoint.geometry.z, joinzz, nosex,nosey,nosez, tailx,taily,tailz, joinx,joiny,joinz ))
            if not radius:
                #  wrap: (0.0, 0.0, 0.0) joinzz=0.0  nose:(4.1682999999999995, 5.0, 0.0)  tail:(0.0, 5.0, 0.0)  join:(2.3203844070513977, 5.0, 0.0) 
                return err(sharejson,"Sorry - can't wrap a distance of 0.\nYour foil nose+tail and join must be on a sketch XY plane, and your axis must also be on a DIFFERENT XY sketch plane with a different Z (e.g. a 3D sketch, or two sketches on different Z planes)",dlg)

            circumference=math.pi * 2 * radius
            wrappx=wrapline.startSketchPoint.geometry.x
        else: # Not wrapping
            wrappx=0

        oblique = nose.geometry.z - tail.geometry.z # For people who insert foils at strange angles
        if oblique != 0 and tail.geometry.x != nose.geometry.x:
            if tail.geometry.x - nose.geometry.x:
                rotamt=math.atan(oblique/(tail.geometry.x - nose.geometry.x)) # how much to rotate to connect to their wanted line   # ERROr - /0
            else:
                rotamt=0

        elif oblique > 0:
            rotamt=math.pi/2

        elif oblique != 0:
            rotamt=-math.pi/2

        else: 
            oblique=0 # they picked a Z direction for the foil.
            rotamt=0  # won't be used when oblique==0 anyhow

        #eprint("oblique={}. zz={}. Rotating {} degrees translating {} cm nose({},{},{}) tail({},{},{})".format(oblique, joinzz, rotamt*180/math.pi,oblique, nose.geometry.x, nose.geometry.y, nose.geometry.z, tail.geometry.x, tail.geometry.y, tail.geometry.z  ))
        minx=maxx=wrapdeg=None # So we know how far we wrapped it, so we can work out if Fusion360 will be able to patch it

        for i in range(0,len(splinex)): # The last one will connect back to the first (it's the same as the first anyhow) - fusion notices this and drops it.
            # maybe we should create them all someplace constrained, then move them?
            spliney[i]=spliney[i]*flip

            if nose_tail_angle: # put it the way they asked for next
                (x,y)=(splinex[i],spliney[i])
                splinex[i]= x * math.cos(-nose_tail_angle*math.pi/180) + y*math.sin(-nose_tail_angle*math.pi/180) 
                spliney[i]= -x * math.sin(-nose_tail_angle*math.pi/180) + y*math.cos(-nose_tail_angle*math.pi/180)

            if ( rotate>0 and rotate<360 ) or extrarotate:  # tilt to to the optimal AoA first
                (x,y)=(splinex[i]+(nosex-joinx)/scale,spliney[i]+(nosey-joiny)/scale) # test rotate - negative is anticlockwise.
                splinex[i]=  x * math.cos(flip*(rotate+extrarotate)*math.pi/180) + y * math.sin(flip*(rotate+extrarotate)*math.pi/180) -(nosex-joinx)/scale
                spliney[i]= -x * math.sin(flip*(rotate+extrarotate)*math.pi/180) + y * math.cos(flip*(rotate+extrarotate)*math.pi/180) -(nosey-joiny)/scale


            if oblique !=0 :
                #need to... (a) add obz to all points
                #then (b) rotate this back so the tail lands where it belongs
                (x,z)=(splinex[i],splinez[i]/scale)
                splinex[i]= x * math.cos(rotamt) + z*math.sin(rotamt) 
                splinez[i]=( -x * math.sin(rotamt) + z*math.cos(rotamt)) * scale
                obz=oblique # * (1-splinex[i])    # tilt it to match the join line
            else:
                obz=0


            # Wrap foil around cyl:
            # a) translate point to common starting line
            # b) rotate by X distance around circumference (new X and Z - Y is untouched)
            # wrapline=None,  # Center of a cylinder that the foil will be bent around - this is a tangent to the chord, at some Z distance away
            # wrapstart=None, # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not 
            #                   have to be part of the foil - only the X and distance-from-line is used.)

            if wrapline:
                #axis= # adsk.core.InfiniteLine3D.create(wrapline.startSketchPoint.geometry, wrapline.endSketchPoint.geometry.vectorTo( wrapline.startSketchPoint.geometry )) #InfiniteLine3D doesn't seem to have any distance stuff. Grr.

                x=splinex[i]*scale+nosex+shifto
                dist=(x-wrappx)/circumference * math.pi * 2 # How many radians do we need to rotate to match this distance around the circumference?
                splinex[i]=(( radius * math.sin(dist) ) - nosex-shifto) / scale # The last 2 just pre-un-do the scale/translate that's about to come next
                splinez[i]=  - radius * math.cos(dist) + radius 

                #eprint("wrap i={} x={} radius={} dist={} sin={} cos={}".format(i,x,radius,dist,math.sin(dist),math.cos(dist)))  # cos 1..1=flat, 1..0 = 45 degrees, 1..-1 = 180, beyond -1 and back towards 1 = past 180



            xyz=[ splinex[i]*scale+nosex+shifto+wrappx, spliney[i]*scale+nosey+shifto, splinez[i]+joinzz+obz ]
            #eprint("i={}. #={} [{}, {}, {}]".format(i,len(xyzpoints),*xyz))
            xyzpoints.append(xyz) # keep a python copy too
            point=adsk.core.Point3D.create( *xyz ) # caution - seems to auto-add constraints when my points hit other points. # +1 to prevent auto-constraining going on
            if minx is None or xyz[0]<minx: minx=xyz[0]  # Remember actual extents, so we know how far we wrapped
            if maxx is None or xyz[0]>maxx: maxx=xyz[0]

            # do-planars...
            #if not first_nonend and len(xyzpoints)>1 and ( xyzpoints[0][0]!= xyz[0] or xyzpoints[0][1]!= xyz[1] or xyzpoints[0][2]!= xyz[2] ):
            #    first_nonend=len(xyzpoints)
            #    eprint("end{}".format(first_nonend))
            #    eprint(xyzpoints[0])
            #    eprint(xyz)

            #if not len(points):
            #    point.translateBy( adsk.core.Vector3D.create( -1,-1,0 )) # prevent co-incidence...
            #point=adsk.core.Point3D.create( ) # no xyz to avoid constraint adding
            #point=adsk.core.Point3D.create( 0,0,0 ) 
            #point.set( *xyz ) # fails - becomes co-incident again
            #point.move( adsk.core.Vector3D.create( *xyz ))  # no such method
            #point.translateBy( adsk.core.Vector3D.create( *xyz ))  # fails - becomes co-incident again
            #point.geometry.x=xyz[0]
            controlPoints.append(point)
            points.add(point) 
            if splinetype:
                sketchPoints.add(point) # Show the spline control points on the sketch as well
            #sketchPoints.add(point) # Show the spline control points on the sketch as well
            #eprint("Added point {} at ({}, {}, {})  [spline was ({}, {})] shifted:{}".format(i,splinex[i]*scale+nosex, spliney[i]*scale+nosey, splinez[i], splinex[i],spliney[i],shifto))
            adsk.doEvents()

        if minx is not None and maxx is not None and radius: wrapdeg= (maxx - minx) / (math.pi * radius * 2) * 360 # How many degrees did we wrap this?
        if splinex[0]==splinex[1] and spliney[0]==spliney[1] and splinez[0]==splinez[1]:
            eprint("Caution: duplicated starting point")
        if splinex[-2]==splinex[-1] and spliney[-2]==spliney[-1] and splinez[-2]==splinez[-1]:
            eprint("Caution: duplicated ending point")

        #Generates the spline curve
        if not splinetype:  # fit point
            if xyzpoints[0][0]==xyzpoints[-1][0] and xyzpoints[0][1]==xyzpoints[-1][1] and xyzpoints[0][2]==xyzpoints[-1][2]:  # Starting and ending point is coincident
                midpt=adsk.core.Point3D.create( ( xyzpoints[-2][0]+xyzpoints[-1][0] ) /2 , ( xyzpoints[-2][1]+xyzpoints[-1][1] ) /2 , ( xyzpoints[-2][2]+xyzpoints[-1][2] ) /2 )   # Make a temporary ending point half-way back up the spline
                #oripoint=points.item(points.count-1)                # Save the original
                # back =points.item(points.count-1).vectorTo(midpt)   
                forth=midpt.vectorTo(points.item(points.count-1))   # prepare to move the fake end to the right place
                points.removeByIndex(points.count-1)    # Remove the former last point
                points.add(midpt)                       # replace the real end with a fake end

                #points.item(points.count-1).move(back) # move the last point back, so it is no longer co-incident
                #back =sketchPoints.item(sketchPoints.count-1).geometry.vectorTo(midpt)
                #forth=midpt.vectorTo(sketchPoints.item(sketchPoints.count-1).geometry)
                #sketchPoints.item(sketchPoints.count-1).move(back) # move the last point back, so it is no longer co-incident
                foilobj=sketch.sketchCurves.sketchFittedSplines.add(points) # Caution: auto-joins coincident points (the start and the end) dropping the duplicates.  $#foilobj==$#points-1
                foilobj.endSketchPoint.move(forth) # move the fake end point back to where we really wanted it (thus keeping any pointy ends)
                # YAY:             foilobj.endSketchPoint.move(foilobj.endSketchPoint.geometry.vectorTo(start)) # move the last point back to the start point without merging them (keeping a pointy end)
            else:
                foilobj=sketch.sketchCurves.sketchFittedSplines.add(points) # Caution: auto-joins coincident points (the start and the end) dropping the duplicates.  $#foilobj==$#points-1
                #foilobj=sketch.sketchCurves.sketchFittedSplines.add(sketchPoints) # Caution: auto-joins coincident points (the start and the end) dropping the duplicates.  $#foilobj==$#points-1

            #points[0].translateBy( adsk.core.Vector3D.create( xyzpoints[0][0]-xyzpoints[first_nonend][0], xyzpoints[0][1]-xyzpoints[first_nonend][1], xyzpoints[0][2]-xyzpoints[first_nonend][2] )) # prevent co-incidence...
            #eprint([ xyzpoints[0][0]-xyzpoints[first_nonend][0], xyzpoints[0][1]-xyzpoints[first_nonend][1], xyzpoints[0][2]-xyzpoints[first_nonend][2] ])
            #points[0].translateBy( adsk.core.Vector3D.create( xyzpoints[0][0]-xyzpoints[1][0], xyzpoints[0][1]-xyzpoints[1][1], xyzpoints[0][2]-xyzpoints[1][2] )) # prevent co-incidence...
            #points[0].translateBy( adsk.core.Vector3D.create( 0,0,0 ))
            #eprint([ xyzpoints[0][0]-xyzpoints[1][0], xyzpoints[0][1]-xyzpoints[1][1], xyzpoints[0][2]-xyzpoints[1][2] ])
            #points[0].translateBy( adsk.core.Vector3D.create( -(xyzpoints[0][0]-xyzpoints[first_nonend][0]), -(xyzpoints[0][1]-xyzpoints[first_nonend][1]), -(xyzpoints[0][2]-xyzpoints[first_nonend][2]) )) # prevent co-incidence...
            #foilobj.translateBy( adsk.core.Vector3D.create( -(xyzpoints[0][0]-xyzpoints[first_nonend][0]), -(xyzpoints[0][1]-xyzpoints[first_nonend][1]), -(xyzpoints[0][2]-xyzpoints[first_nonend][2]) )) # prevent co-incidence...
            #points[0].translateBy( adsk.core.Vector3D.create( -xyzpoints[1][0], -xyzpoints[1][1], -xyzpoints[1][2] ))
            #foilobj.startSketchPoint.move( adsk.core.Vector3D.create(  -xyzpoints[1][0], -xyzpoints[1][1], -xyzpoints[1][2] ))
        else:               #  control-point degree 3
            knots=[ 0,0,0,0 ]
            for i in range(1,len(controlPoints)-3):
                knots.append(i)
            degree=3
            knots = [ *knots[0:], knots[-1]+1, knots[-1]+1, knots[-1]+1, knots[-1]+1 ] 
            foilobj=sketch.sketchCurves.sketchFixedSplines.addByNurbsCurve( adsk.core.NurbsCurve3D.createNonRational(controlPoints, degree, knots, False ))
            #sketchCurves.sketchFittedSplines.addByNurbsCurve( adsk.core.NurbsCurve3D.createNonRational(controlPoints, degree, knots, False )) # Semi-works (breaks if user edits)

        sharejson['usefoil']=usefoil # so we have the original data too
        sharejson['wrapdeg']=wrapdeg
        sharejson['pitch']=flip*(rotate+extrarotate)
        sharejson['entityToken']=foilobj.entityToken # so we can in future do parametric things with this
        foilobj.attributes.add("attr"+aft,"JSON", shareIt(sharejson,dlg=dlg,Text=True) ) # Include everytihng about this foil into the object attributes too

        # Constrain it now, working from the nose to the tail...
        if constrain:
            #sketch.isComputeDeferred = True                                     # Don't waste time computing stuff until the end ... doesn't work :-(
            lasti=-1
            firstcname=None
            lastcname=None # for future re-calcs
            # Add driven chord, so we can tell later if it gets changed
            rh=dim.addDistanceDimension(nose, tail, adsk.fusion.DimensionOrientations.HorizontalDimensionOrientation, pointme(tail.geometry,scale,'x',1), False) # False means driven # ERROR: InternalValidationError : rDim
            if rh: firstcname=rh.parameter.name # how to set sizes if not driven: rh.parameter.expression='{} cm'.format(abs(nosex-tailx)) # ERROr: AttributeError: 'NoneType' object has no attribute 'name'
            rv=dim.addDistanceDimension(nose, tail, adsk.fusion.DimensionOrientations.VerticalDimensionOrientation, pointme(tail.geometry,scale,'y',1), False) # False means driven
            # Note also the CM in dimensions
            dim.addDistanceDimension(nose, cmpt, adsk.fusion.DimensionOrientations.HorizontalDimensionOrientation, pointme(cmpt.geometry,scale,'x',1), True)
            dim.addDistanceDimension(nose, cmpt, adsk.fusion.DimensionOrientations.VerticalDimensionOrientation, pointme(cmpt.geometry,scale,'y',1), True)

            if (joindx or joindy) and not join.sketchDimensions.count:  # Constrain the CM, unless they already did themselves.
                rh=dim.addDistanceDimension(nose, join, adsk.fusion.DimensionOrientations.HorizontalDimensionOrientation, pointme(join.geometry,scale,'x',0)) # ERROR: Sketch geometry is over constrained
                rh.parameter.expression='{} cm'.format(abs(joindx))
                rv=dim.addDistanceDimension(nose, join, adsk.fusion.DimensionOrientations.VerticalDimensionOrientation, pointme(join.geometry,scale,'y',0))
                rv.parameter.expression='{} cm'.format(abs(joindy))

            for j in range(middle,len(splinex)): # Move along the undersurface from the nose to the tail
                for i in[j,middle+(middle-j)]:
                    if i==lasti: continue # Don't do nose twice
                    lasti=i 
                    if i<foilobj.fitPoints.count:
                        eprint("Constraining point {} at ({}, {})  [spline was ({}, {}), dimension was ({},{})]".format(i,splinex[i]*scale+nosex, spliney[i]*scale+nosey, splinex[i],spliney[i],  splinex[i]*scale,spliney[i]*scale))
                        sp=foilobj.fitPoints.item(i)
                        if shifto:
                            if splinex[i]<0 and spliney[i]<0:
                                sp.move( adsk.core.Vector3D.create(  -shifto, -shifto, joinzz ))    # Can't have negative constraints, so put it where it goes now (hopefully that won't merge this point with any it now lands on!)
                            elif splinex[i]<0:
                                sp.move( adsk.core.Vector3D.create(  -shifto, 0.0, joinzz )) # Leave Y alone
                            elif spliney[i]<0: 
                                sp.move( adsk.core.Vector3D.create(  0.0, -shifto, joinzz ))   # Leave X alone
                        while (sp.sketchDimensions.count):
                            eprint('Caution: erased user constraint {}'.format(i))
                            try:
                                sp.sketchDimensions.item(0).deleteMe() # should never happen, but if our random-shift bumped exactly into something of the users, it might...
                            except:
                                eprint('Caution: user constraint erasure failed {}'.format(i))
                        try:
                            rh=dim.addDistanceDimension(nose, sp, adsk.fusion.DimensionOrientations.HorizontalDimensionOrientation, pointme(sp.geometry,scale,'x',i==j))
                            rh.parameter.expression='{} cm'.format(abs(splinex[i]*scale))
                            if not firstcname: firstcname=rh.parameter.name # remember the first one we added
                        except:
                            eprint('Caution: add H constraint failed {}'.format(i))
                        try:
                            rv=dim.addDistanceDimension(nose, sp, adsk.fusion.DimensionOrientations.VerticalDimensionOrientation, pointme(sp.geometry,scale,'y',i==j))
                            rv.parameter.expression='{} cm'.format(abs(spliney[i]*scale))
                            lastcname=rv.parameter.name
                        except:
                            eprint('Caution: add V constraint failed {}'.format(i))
                        adsk.doEvents()
            sharejson['firstcname']=firstcname  # Constraint names - needed to perform re-calcs in future.
            sharejson['lastcname']=lastcname
            #sketch.isComputeDeferred = False   # doesn't help.
            #adsk.doEvents()


        if camber:
            camberp = []
            for j in range(middle,len(xyzpoints)): # Move along the undersurface from the nose to the tail - note, range includes tails
                underpt=j
                overpt=middle+(middle-j)
                midx=(xyzpoints[overpt][0]+xyzpoints[underpt][0])/2
                midy=(xyzpoints[overpt][1]+xyzpoints[underpt][1])/2
                midz=(xyzpoints[overpt][2]+xyzpoints[underpt][2])/2
                camberp.append([midx,midy,midz])


            # Need to extend xyzpoints so the camber sticks out the ends, then use middle+1 below
            p1=camberp[0];j=1;p2=camberp[j] # tail
            while p1[0]==p2[0] and p1[1]==p2[1] and p1[2]==p2[2]: j=j+1;p2=camberp[j] # in case of any dupes
            endsz=( ( p1[0]-p2[0] )**2 + ( p1[1]-p2[1] )**2 + ( p1[2]-p2[2])**2 )**0.5  # How long is the distance between the end 2 points
            p3=[0,0,0]
            for j in range(0,3): p3[j]=  p1[j] - (p2[j] - p1[j])*(scale/2/endsz)          # mirror the point the other side of the line, then extend it out to half the chord length
            camberp = [p3,*camberp]
            
            p1=camberp[-1];j=len(camberp)-2;p2=camberp[j] # nose
            while p1[0]==p2[0] and p1[1]==p2[1] and p1[2]==p2[2]: j=j-1;p2=camberp[j] # in case of any dupes
            endsz=( ( p1[0]-p2[0] )**2 + ( p1[1]-p2[1] )**2 + ( p1[2]-p2[2])**2 )**0.5  # How long is the distance between the end 2 points
            p3=[0,0,0]
            for j in range(0,3): p3[j]=  p1[j] - (p2[j] - p1[j])*(scale/2/endsz)          # mirror the point the other side of the line, then extend it out to half the chord length
            camberp.append(p3)
            
            # Need to prevent coindident points merging, since that hampers patch from working
            fixtail=camberp[1]  # tail ([0]  is extension)
            fixnose=camberp[-2] # nose ([-1] is extended end)
            midnosept=[ ( camberp[-2][0]+camberp[-1][0] ) /2 , ( camberp[-2][1]+camberp[-1][1] ) /2 , ( camberp[-2][2]+camberp[-1][2] ) /2 ]   # Make a temporary ending point half-way back up the spline
            midtailpt=[ ( camberp[0][0] + camberp[1][0] ) /2 , ( camberp[0][1] + camberp[1][1] ) /2 , ( camberp[0][2] + camberp[1][2] ) /2 ]   # Make a temporary ending point half-way back up the spline
            camberp[1]=midtailpt    # replace nose and tail coincident points with fake ones temporarily
            camberp[-2]=midnosept

            '''
            if xyzpoints[0][0]==xyzpoints[-1][0] and xyzpoints[0][1]==xyzpoints[-1][1] and xyzpoints[0][2]==xyzpoints[-1][2]:  # Starting and ending point is coincident
                midpt=adsk.core.Point3D.create( ( xyzpoints[-2][0]+xyzpoints[-1][0] ) /2 , ( xyzpoints[-2][1]+xyzpoints[-1][1] ) /2 , ( xyzpoints[-2][2]+xyzpoints[-1][2] ) /2 )   # Make a temporary ending point half-way back up the spline
                forth=midpt.vectorTo(points.item(points.count-1))   # prepare to move the fake end to the right place
                points.removeByIndex(points.count-1)    # Remove the former last point
                points.add(midpt)                       # replace the real end with a fake end
                foilobj=sketch.sketchCurves.sketchFittedSplines.add(points) # Caution: auto-joins coincident points (the start and the end) dropping the duplicates.  $#foilobj==$#points-1
                foilobj.endSketchPoint.move(forth) # move the fake end point back to where we really wanted it (thus keeping any pointy ends) # YAY: 
            '''

            camberpoints = adsk.core.ObjectCollection.create()
            for j in range(len(camberp)):
                camberpoint=adsk.core.Point3D.create( *camberp[j] ) 
                camberpoints.add(camberpoint) 

            camberobj=sketch.sketchCurves.sketchFittedSplines.add(camberpoints)
            sharejson['camber_entity']=camberobj.entityToken

            # Move the fake ends to the right place now
            #camberobj.fitPoints.item(camberobj.fitPoints.count-2)==camberobj.endSketchPoint
            camberobj.fitPoints.item(1).move(                           camberpoints.item(1).vectorTo(                    adsk.core.Point3D.create( *fixtail ) ))   # move the tail back to where we wanted it originally (hopefully avoids auto-coincidence)
            camberobj.fitPoints.item(camberobj.fitPoints.count-2).move( camberpoints.item(camberpoints.count-2).vectorTo( adsk.core.Point3D.create( *fixnose ) ))   # move the nose back to where we wanted it originally (hopefully avoids auto-coincidence)


        if endscale:    # Add the end caps
            shifto=0
            for j in range(middle+1,len(splinex)-1): # Move along the undersurface from the nose to the tail - note, range, never hits end...
                tweak=1-abs(j-( middle+1 + len(splinex) )/2) / len(splinex) # 0.5 to 1 to 0.5 - makes the middle 50% longer than the ends
                
                underpt=j
                underpt=adsk.core.Point3D.create( splinex[underpt]*scale+nosex+shifto, spliney[underpt]*scale+nosey+shifto, joinzz)
                overpt=middle+(middle-j)
                overpt=adsk.core.Point3D.create( splinex[overpt]*scale+nosex+shifto, spliney[overpt]*scale+nosey+shifto, joinzz)
                enddist=underpt.distanceTo(overpt) # Not the best plan - top and bottom surface points are not always in line, leading to screwey distances...
                endcap = adsk.core.ObjectCollection.create() 

                endcap.add(overpt)
                endcap.add(underpt)
                endline=sketch.sketchCurves.sketchLines
                endline.addByTwoPoints(underpt,overpt) # join the shape ends with a line, so the shape can be lofted

                if 1: # Testing slowness
                    minor=adsk.core.Point3D.create( 0.5*(overpt.x+underpt.x), 0.5*(overpt.y+underpt.y), endscale*enddist*tweak+joinzz) # ellipse minor axis
                    midpt=minor
                    endcap.add(midpt)
                    center=adsk.core.Point3D.create( 0.5*(overpt.x+underpt.x), 0.5*(overpt.y+underpt.y), joinzz) # ellipse center
                    major=overpt
                    try:
                        ellipse=sketch.sketchCurves.sketchEllipses.add(center, major, minor)
                    except:
                        eprint('Caution: sketchEllipses.add failed')

                elif 0:
                    midpt=adsk.core.Point3D.create( 0.5*(overpt.x+underpt.x), 0.5*(overpt.y+underpt.y), .5*enddist+joinzz) # 3D point
                    endcap.add(midpt)
                else:
                    for mida in ( 3, 14.4, 90, 158.4, 177 ): # 2%, 8%, 50%, 92%, 98% - so the spline ends are not at an angle
                        # nice ends require the fist spline points to be in line with the foil edge
                        if mida==3:      clamp=-3
                        elif mida==177:  clamp=3
                        else:            clamp=0
                        intpt=adsk.core.Point3D.create( 0.5*(1+math.cos(    (mida+clamp)/180*math.pi))*overpt.x+
                                                        0.5*(1+math.cos((180-mida-clamp)/180*math.pi))*underpt.x,
                                                        0.5*(1+math.cos(    (mida+clamp)/180*math.pi))*overpt.y+
                                                        0.5*(1+math.cos((180-mida-clamp)/180*math.pi))*underpt.y, 
                                                        0.5*math.sin(mida/180*math.pi)*enddist+joinzz) # 3D point
                        endcap.add(intpt)
                        if mida==90: midpt=intpt 
                        adsk.doEvents()

                if j==middle+1:                                                 # join the nose too
                    endline=sketch.sketchCurves.sketchLines
                    endline.addByTwoPoints(midpt, adsk.core.Point3D.create( splinex[middle]*scale+nosex+shifto, spliney[middle]*scale+nosey+shifto, joinzz) )
                if j==len(splinex)-2:                                           # join the tail too
                    endline=sketch.sketchCurves.sketchLines
                    endline.addByTwoPoints(midpt, adsk.core.Point3D.create( splinex[0]*scale+nosex+shifto, spliney[0]*scale+nosey+shifto, joinzz) )

        # Close it with a line between endpoints here
        if minthick>0:
            if 0: # lines do not patch
                tailine=sketch.sketchCurves.sketchLines
                mytail=tailine.addByTwoPoints(points[0],points[-1])  # join the shape ends with a line, so the shape is a closed loop
                sharejson['mytail']=mytail  # needed for patch
                sharejson['tail_entity']=mytail.entityToken
            else:
                mytail = adsk.core.ObjectCollection.create()
                mytail.add(points[0])
                mytail.add(points[-1])                               # join the shape ends with a (fake 2-point spline) line, so the shape is a closed loop
                mytailobj=sketch.sketchCurves.sketchFittedSplines.add(mytail)
                sharejson['mytail_entity']=mytailobj.entityToken

        if patch:
            eprint("wrapdeg={}".format(wrapdeg))
            try:
                patches = sketch.parentComponent.features.patchFeatures # patches = rootComp.features.patchFeatures
                patchInput = patches.createInput(foilobj, adsk.fusion.FeatureOperations.NewBodyFeatureOperation)
                # Need to add mytail...
                sharejson['patchInput']=patchInput
                sharejson['patch']=patches.add(patchInput)
                sharejson['patch_entity']=sharejson['patch'].entityToken
            except:
                pass # avoid the below...
                # RuntimeError: 5 : SurfacePatch12 / 1 Reference Failures   The operation failed.     Try adjusting the values or changing the input geometry. //   The operation failed.
                # Try adjusting the values or changing the input geometry. / Compute Failed // Modeling Error: Can't create patch toolbody, please check the inputs.


        sharejson['points']=points # So caller can add rails
        sharejson['foilobj']=foilobj # So caller can use this with patches etc
        sharejson['foil_entity']=foilobj.entityToken # so we can in future do parametric things with this

        return sharejson    # Everything about what we just did

    # Error handling
    except:
        return err(sharejson,program_name+' insertFoil failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox

    #insertFoil




def bestFoil(Re=None,
             N=None,
             strut=False,
             TextInfo=False,
             note='',
             sharejson=None,
             liftamt=1,          # pick a foil that has best (1) lift, or no lift (0 - e.g. strut) - e.g. for winglets.
             Mach=0,
             Cloud=True,
             WantS=None          # i.e. Convert so splinetype 0 if we got a 1
             ):
    foilmatch=-1    # How close does the foil Re match our target Re?
    usefoil=None    # Which is the foil to use?
    if liftamt<0.5: strut=True # override their strut selection if they specified a lift amount (future: can optimise sections specifically here)
    source='Cloud'


    if Cloud: # AirfoilTools.py looks up our cloud DB from here...
        strutno=0
        if strut: strutno=1
        cloud_key='Re:{},N:{},mach:{},strut:{}'.format( Re, N, Mach, strutno ) 

        if cloud_cache.get(cloud_key,0):
            usefoil=cloud_cache[cloud_key]
        else:
            cloud_foil=web_req('json={{"Re":{},"N":{},"mach":{},"strut":{}}}'.format( Re, N, Mach, strutno ) )              # https://foils.pro/Airfoil_Tools/share.asp?json={"N":12,"Re":6991794,"mach":0.097,"strut":1}

            m=re.search('(?:\r\n\r\n|\r\r|\n\n)[0-9a-fA-F]+(?:\r\n|\r|\n)(.*?)(?:\r\n|\r|\n)',cloud_foil,re.DOTALL);        # Split the headers, the python hex size addon, and the trailling zero off the input.
            if m:
                cloud_foil=m.group(1)
            try:
                cloud_foil=json.loads(cloud_foil); # Try to convert it...
                if isinstance(cloud_foil,list):
                    usefoil=cloud_foil[0]
                    usefoil['xy']=points_to_float(usefoil['xy'])
                    #usefoil['S']=1 # all cloud foils are S:1
                    if isinstance(usefoil['s'], str): usefoil['s']=float(usefoil['s']);
                    #usefoil['s']=usefoil['S']            # normalizeme
                    (scale,rot,tx,ty)=normalize(usefoil) # scale, rotate, translate (in that order).
                    eprint("cloud foil Normalized via scale:{} rotate:{} translate({},{})".format(scale,rot*180/math.pi,tx,ty))
                    if WantS is not None and WantS==0:    # All our cloud foils are Control-point, but caller wants a fit-point...
                        splinetype=1 # Input points
                        (splinex,spliney)=jspline(-9999,splinetype,splinetype,1,points_to_xy( usefoil['xy'])) # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)
                        splinetype=0 # Output points
                        (splinex,spliney) = equispace(60, 10, splinetype, xy_to_points(splinex,spliney) )  # Convert our input spline into 60 fitpoints, with concentration-at-nose factor of 10
                        usefoil['xy'] = xy_to_points(splinex,spliney)
                        usefoil['s']=splinetype

                        #usefoil['xy'] = xy_to_points( jspline(-9999,splinetype,splinetype,1, usefoil['xy'] ) )  # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)

                        #xy=xy_to_points(splinex,spliney) # Because we changed the splinetype
                        #usefoil['xy'] = xy_to_points( jspline(-9999,splinetype,splinetype,1, usefoil['xy'] ) )  # expand what we got into a very large number (thus the control-points approach being a fit-point in so doing)
                        #usefoil['s']=0
                        #usefoil['xy'] = equispace(60, 10, splinetype, usefoil['xy'] ) # Convert our input spline into 60 fitpoints, with concentration-at-nose factor of 10

                else:
                    sharejson={'cloud_foil':cloud_foil}
                    err(sharejson, "Unexpected foil data recieved from Cloud service. Reverting to inbuilt DB.  Please let us know about this error")        # signal error to caller and # pops up a messageBox
                
            except:
                sharejson={cloud_foil:cloud_foil}
                err(sharejson, "Invalid foil data recieved from Cloud service. Reverting to inbuilt DB.  Please let us know about this error")        # signal error to caller and # pops up a messageBox

            if usefoil is not None: cloud_cache[cloud_key]=usefoil

        if usefoil is None: vlog("Cloud foil lookup failed. Fallbak to inbiult DB");
        # vlog("usefoil"); vlog(usefoil); vlog("cloud_foil"); vlog(cloud_foil)

    if usefoil is None: # Not could foil found, or, not using cloud_foil at all
        source='db'
    
        foil=foildb2020.Foildb2020()    # Load the 1220 top candidates (filtered from 307M experiments) to search
    
        # See if we have data for the N we want - but if we do not find it, try either side until we do
        useN=0
        for noffs in range(0,11,1): # N goes from 3 to 14 (12 possible points) so the furthest from any we're on would be 11
            if useN: break
            for ntry in [N+noffs, N-noffs]: # does N+0 twice... boo hoo... should always be found first-go anyhow.
                if useN: break
                for tryfoil in foil:
                    if ( tryfoil.get('n',0)==ntry                               # found at least one foil with this critical angle computed.
                        and ( (tryfoil.get('sym',1) < 0.00125 and strut)        # If it's symmetrical, and we want symmetrical, 
                        or (tryfoil.get('sym',1) >= 0.00125 and not strut))):   # or it's not symmetrical, and we want a wing (not symmetrical)
                        useN=ntry
                        break
    
        # Which database foil Re most closely-matches the one we want?
        for tryfoil in foil:
            if (tryfoil.get('n',0)!=useN                                        # Skip ones that do not apply to us...
                or (tryfoil.get('sym',1) < 0.00125 and not strut)               # If it's symmetrical, and we want a foil...
                or (tryfoil.get('sym',1) >= 0.00125 and strut)):                # or it's not symmetrical, and we want a strut
                continue # Skip the wrong ones
            dif=abs(Re - tryfoil.get('Re',0))
            if usefoil is None or dif<foilmatch:
                usefoil=tryfoil
                foilmatch=dif
    
        if foilmatch==Re: # foilmatch is 0 if perfect, Re if worst possible
            if TextInfo:
                app = adsk.core.Application.get()
                ui = app.userInterface
                ui.messageBox('Failed to find a foil to match your criteria - sorry!\n\nPlease use our forum/feedback link to report this problem')
    
    # Tell them about what we found
    goodness='' # CDp  =  CD - CDf
    if strut:
        goodness='drag coefficient: {:.5g} (=friction drag {:.5g} + pressure drag {:.5g})'.format(usefoil.get('CD','?'),usefoil.get('CDf','?'),usefoil.get('CDpres','?'))
    else:
        goodness='Best L/D={:.5g} at Aoa={:.5g} lift coefficient: {:.5g}, drag coefficient: {:.5g} (friction drag {:.5g} + pressure drag {:.5g})'.format(
            usefoil.get('CL',0) / usefoil.get('CD',1),
            usefoil.get('alpha','?'),
            usefoil.get('CL','?'),
            usefoil.get('CD','?'),
            usefoil.get('CDf','?'),
            usefoil.get('CDpres','?'))
    
    fmsg="{}. Best {} foil is Re {:,.0f} n{} mach{} and named {}. {}".format(note,source,usefoil['Re'],usefoil.get('n',0),usefoil.get('mach',0),usefoil['Foil_fn'],goodness)

    if TextInfo:
        vlog(fmsg)
        #vlog(usefoil)   # {'r': 6, 'id': 289, 'cmt': 'OPT s(na) foildb2020_spline/spline_foil_be6568.dat-s17_n17s0.dat (opt29.4917047265402>29.4917047265402) l1 Foil spline 17pts Re125000 N12', 'good': 18.5635312399738, 'xver': 6.991, 'iver': 1.2020052, 'Cm': -0.150387676675614, 'CDp': 0.0128041714353068, 'CD': 0.0326595778938143, 'CL': 1.25227953535953, 'n': 4, 'Cpmin': -1.73613342431045, 'CDf': 0.01255, 'xy': [1, 0, 0.896866800105748, 0.0301632991781163, 0.724380844156728, 0.060917897648545, 0.510552344463168, 0.0826298785144619, 0.334472441945641, 0.0881611787641998, 0.190270468629251, 0.077284105995217, 0.0873478702982618, 0.0541033747278037, 0.0305344216970311, 0.0288773683684043, 0, -3.43565157668151e-09, 0.0220756519913723, -0.016159129053836, 0.135324553635068, 0.00294665648312284, 0.285751757739152, 0.0276101636800961, 0.44123537629784, 0.0416652008361107, 0.591955087760974, 0.0448076159702944, 0.726206007656939, 0.040248526777877, 0.854808845142057, 0.0262245756217473, 1, 0], 'Bot_Xtr': 1, 'sym': 5.7279194130313, 'alpha': 5.7, 'CDpres': 0.02011, 'clcd': 38.343408461403, 'Top_Xtr': 0.538408658005029, 'Foil_fn': 'optbs-spline_foil_be6568.dat-s17_n17s0-Re125000-N12.dat', 'diter': 3, 'src': 'retest_134.txt', 'Re': 35000}
    if sharejson: sharejson['foilinfo']=fmsg

    return usefoil



def insertTurbine( sketch=None,
                   nose=None,       # dlg['nose'+aft].selection(0).entity
                   tail=None,       # dlg['tail'+aft].selection(0).entity
                   join=None,
                   usefoil=None,    # from bestFoil()
                   #rotate=None,     # <360 means they want best cl/cd auto-rotate applied
                   flip=None,       # Lift or downforce?
                   #extrarotate=0,   # Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
                   wrapline=None,   # Center of a cylinder that the foil will be bent around
                   wrapstart=None,  # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                   endscale=None,   # end caps
                   textInfo=None,      # Tell the user what we did in the text info area (and, for errors, with UI popups)
                   constrain=None,
                   camber=None,
                   sharejson=None): # Return info
    global VERSION,foilinfo

    try:

        (nosex,nosey,nosez)=(nose.geometry.x, nose.geometry.y, nose.geometry.z)
        (tailx,taily,tailz)=(tail.geometry.x, tail.geometry.y, tail.geometry.z)
        scale=nose.geometry.distanceTo(tail.geometry) # OK

        '''

        # Perform the insert
        insertTurbine( sketch=sketch, 
                    nose=nose,           # dlg['nose'+aft].selection(0).entity
                    tail=tail,           # dlg['tail'+aft].selection(0).entity
                    join=join,
                    usefoil=usefoil,
                    rotate=rotate,    
                    flip=flip,    
#                    extrarotate=0, # Additional (on top of AoA, if any) rotation to apply (typically used for propeller pitch based on RPM, speed, and inflow)
#                    wrapline=None,  # Center of a cylinder that the foil will be bent around
#                    wrapstart=None, # The point where the wrap will commence from (e.g. usually nose or tail or some middle area - does not have to be part of the foil - only the X and distance-from-line is used.)
                    endscale=endscale,
                    textInfo=True,      # Tell the user what we did in the text info area (and, for errors, with UI popups)
                    constrain=constrain,
                    camber=camber,
                    sharejson=sharejson) 
        return sharejson    # Everything about what we just did


        '''

    except:
        pass


# put dimensions into user-firendly non-overlapping spots
def pointme(dimpt,scale,xy,topbot):
    amt=scale/400
    newx=dimpt.x
    newy=dimpt.y
    if xy=='x':
        if topbot: # bottom
            newy=newy-amt
            newx=newx+amt
        else:
            newy=newy+amt/2
            newx=newx+amt
    else:
        if topbot: # bottom
            newy=newy+amt/2
            newx=newx-amt
        else:
            newx=newx-amt
    return adsk.core.Point3D.create(newx, newy, 0)

def xy_to_points(x,y):                                                          # Take 2 lists (a X list and a Y list) and put them into a single list [x1,y1,x2,y2,...xn,yn]
    points=[]
    for i in range(0,len(x)):
        points.append(x[i])
        points.append(y[i])
    return points


def points_to_float(points):
    for i in range(0,len(points)):
        if isinstance(points[i], str): points[i]=float(points[i]);
    return points

def points_to_xy(points):                                                       # Opposite of xy_to_points
    (x,y)=([],[])
    for i in range(1,len(points)+1,2):
        x.append(points[i-1])
        y.append(points[i])
    return (x,y)

def points_to_1xy(points):                                                      # Convert an [x1,y1,x2,y2,...xn,yn] into a [ [x1,y1], [x2,y2], ...[xn,yn]]
    foil=[]
    for i in range(1,len(points)+1,2):
        foil.append([ points[i-1], points[i] ])
    return foil

def points_to_1xyz(points,zz):
    foil=[]
    for i in range(1,len(points)+1,2):
        foil.append([ points[i-1], points[i], zz ])
    return foil

# translate, scale, and rotate x,y coords so X ranges from 0 to 1 and, nose is at 1,0 and tail is at (or in line with) 0,0
def normalize(usefoil):
    (splinex,spliney)=points_to_xy( usefoil['xy'] )
    (orix,oriy)=points_to_xy( usefoil['xy'] )
    splinetype= usefoil['s'] 

    # Turn our working copy into a very high-res shape - now done in jspline
    #tgt=len(splinex)
    #fact=0
    #while tgt < 9999:
    #    fact=fact+1
    #    tgt=tgt*2   # Work out what factor to use to get the number of points we want
    #(splinex,spliney)=jspline(fact,splinetype,splinetype,1, [ splinex,spliney ])    # expand it

    # Turn our working copy into a very high-res shape
    (splinex,spliney)=jspline(-4000,splinetype,splinetype,1, [ splinex,spliney ])    # expand it to at least 4096 points

    # find the midpoint between the two ends (which might not be touching)
    tailx=(splinex[0]+splinex[-1])/2    # Should be 1
    taily=(spliney[0]+spliney[-1])/2    # Should be 0

    # find the point that is the furthest away from the tail - this will be the nose
    maxdist=-1
    maxi=-1
    for i in range(0,len(splinex)):
        dist=( (splinex[i]-tailx)**2 + (spliney[i]-taily)**2 ) ** 0.5
        if dist > maxdist:
            maxdist = dist
            maxi = i

    # grow/shrink this so the nose will lie on point 0,0 after any subsequent rotate
    scale=1/maxdist
    for i in range(0,len(splinex)):
        splinex[i] = 1+(splinex[i]-1) *scale
        spliney[i] = spliney[i] *scale
    for i in range(0,len(orix)):    # Works
        orix[i] = 1+(orix[i]-1) *scale
        oriy[i] = oriy[i] *scale

    # Rotate everything to lie the nose onto the origin line
    rot = math.atan(spliney[maxi] ) # using radians... /(math.pi/180)) ... nb: already scaled above -  90-degrees causes sin(rot) to ==1 and cos==0
    sina=math.sin(rot)
    cosa=math.cos(rot)
    for i in range(0,len(splinex)):
        tx=(splinex[i]-1) # rotate about the tail
        splinex[i] = 1+ tx * cosa - spliney[i] * sina
        spliney[i] = spliney[i] * cosa + tx * sina
    for i in range(0,len(orix)):
        tx=(orix[i]-1)
        orix[i] = 1+ tx * cosa - oriy[i] * sina
        oriy[i] = oriy[i] * cosa + tx * sina

    # find the new midpoint between the two end points (which might not be touching) - should be 1,0 - if we rotated above, this will not be the same as the first tailx,taily in some situations
    tailx=(splinex[0]+splinex[-1])/2
    taily=(spliney[0]+spliney[-1])/2

    # Translate this to the origin: note that if we rotate it in a moment, we might need another translate...
    for i in range(0,len(splinex)):
        splinex[i] = splinex[i] - (tailx-1)
        spliney[i] = spliney[i] - taily
    for i in range(0,len(orix)):
        orix[i] = orix[i] - (tailx-1)
        oriy[i] = oriy[i] - taily 
        # dust removal? 0.000000019310507
        if(abs(orix[i])<0.0000001): orix[i]=0
        if(abs(oriy[i])<0.0000001): oriy[i]=0
        if(abs(orix[i]-1)<0.0000001): orix[i]=1
        if(abs(oriy[i]-1)<0.0000001): oriy[i]=1

    #vlog("normalize: s={} scale={} rot={} tailx={} taily={} ori={} norm={}".format(splinetype, scale,rot,-tailx,-taily, usefoil['xy'], xy_to_points(orix,oriy)))

    usefoil['xy']=xy_to_points(orix,oriy)   # Return results to the caller

    return scale,rot,-tailx,-taily          # Also tell caller what we had to do to normlaize this



# Check to see if they changed something (e.g. chord len or parameter) that would require us to re-compute any foils...
class CommandHandler(adsk.core.ApplicationCommandEventHandler):
    def __init__(self, targetCommand):
        super().__init__()
        self.tgtCmd = targetCommand
    def notify(self, args):
        global _ui
        try:
            web_nbreq(None) # Service non-blocking websocket

            if 1: # Debugging
                eventArgs = adsk.core.ApplicationCommandEventArgs.cast(args)
                #eprint("CommandHnadler DEBUG: {}".format(eventArgs.commandId)) - can poll here for parametric things that might have changed...
            # Possible things to watch...: FusionDragSketchCommand SketchEditDimensionCmdDef SketchDimension ChangeParameterCommand
            pass
#            if _ui.activeCommand == self.tgtCmd:
#                # Here, the change of UserParameters is checked.
#                _ui.messageBox('call ' + self.tgtCmd)
#            if debug:
#                eprint(_ui.activeCommand)

        except:
            pass



# Trigger a co-process to see if any updates exist, and get them if yes
def check_update(forcechk):
    ui = None
    try:
        global debug, VERSION, userdata, updatedata

        updatedata=load_user_settings(update_file)                  # Check any old previously-loaded update info
        if updatedata.get('loaded',0):                              # e.g. /Users/cnd/Library/Application Support/Airfoil_Tools/latest_version_aft.json or  C:\Users\cnd\AppData\Roaming\Airfoil_Tools\latest_version_aft.json
            updatedata['old']=True                                  # So we know the difference between old update info, and new stuff
        updatedata['loaded']=False

        # If not checked for updates before, or, last check was 23+ hours ago...
        if forcechk or ( not updatedata.get('last_check_time',0) ) or updatedata['last_check_time']+(23*60*60) < (int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())):
            # Start an update check, in the background, so it doesn't slow the user down. See home_file for where this gets written to
            try: # Writes JSON data into file latest_version'+aft+'.json in appropriate user settings folder
                eprint("backgrounding update-check into " + home_file(program_name,'.'))
                #r=subprocess.Popen([get_exec(),os.path.join(prog_folder,"check_update.py"),str(VERSION) + foildb2020.VERSION(VERSION) ,aft,program_name.replace(" ", "_")])
                #r=subprocess.Popen([get_exec(),os.path.join(prog_folder,"check_update.py"),str(VERSION) + "&db=" + str(foildb2020.VERSION(0)) ,aft,program_name.replace(" ", "_"),str(forcechk)])
                r=subprocess.Popen([get_exec(),os.path.join(prog_folder,"check_update.py"), str(VERSION) , aft, program_name.replace(" ", "_"), str(foildb2020.VERSION(0)), str(forcechk+2)])
            except:
                eprint('subprocess failed {}'.format(traceback.format_exc()))
                try:    # see if the path magically finds it
                    #r=subprocess.Popen(['python',os.path.join("prog_folder","check_update.py"),str(VERSION) + foildb2020.VERSION(VERSION) ,aft,program_name.replace(" ", "_")])
                    #r=subprocess.Popen(['python',os.path.join("prog_folder","check_update.py"),str(VERSION) + "&db=" + str(foildb2020.VERSION(0)) ,aft,program_name.replace(" ", "_"),str(forcechk+2)])
                    r=subprocess.Popen(['python',os.path.join("prog_folder","check_update.py"), str(VERSION) , aft, program_name.replace(" ", "_"), str(foildb2020.VERSION(0)), str(forcechk+2)])
                except:
                    eprint('subprocess failed again {}'.format(traceback.format_exc()))

    except:
        eprint("update-check problem")



#################################################################################
####                                                                          ###
##                               Main() Entry Point                            ##
####                                                                          ###
#################################################################################

def run(context):
    ui = None
    try:
        global textpalette, debug, VERSION, userdata, updatedata, pnl, _ui
        app = adsk.core.Application.get()
        _ui = app.userInterface
        ui = app.userInterface

        textpalette=ui.palettes.itemById('TextCommands')
        #welcome="Welcome to " + program_name + " Add-in v"+str(VERSION)+' ' + foildb2020.VERSION(VERSION)+" on " + datetime.datetime.now().strftime("%Y-%m-%d %H:%M %Z%z") + " (" + str(int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())) + ")"
        #welcome='Welcome to {} Add-in v{:1.8f}{} on {} ({})'.format(program_name, VERSION, foildb2020.VERSION(VERSION), datetime.datetime.now().strftime("%Y-%m-%d %H:%M %Z%z"), int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds()))
        welcome='Welcome to {} Add-in v{:1.8f}{} on {}'.format(program_name, VERSION, foildb2020.VERSION(VERSION), datetime.datetime.now().strftime("%Y-%m-%d %H:%M %Z%z") )
        eprint(welcome) # eprint only prints in debug mode

        if  textpalette:
#            textpalette_ori_state=textpalette.isVisible
#            if debug:
#                textpalette.isVisible=True
            if textpalette.isVisible:
                textpalette.writeText(welcome) # So any logs we get from customers identify what and when

        check_update(0)

        if updatedata.get('critical',0):                            # Note that this flag was read from the result of an update that happened earlier (not the above in this run)
            ui.messageBox(updatedata['critical'])                   # Same as fatal, but allows our update code to run allowing the possible clearing of this problem
            # Need to backup user settings first...
            sys.exit(1)                                             # quit to prevent catastrophe (e.g. data corruption or security issue)


        #################################################################################
        ####                                                                          ###
        ##        Menu builder (for Button and DropDown (slide-out) UI elements)       ##
        ####                                                                          ###
        #################################################################################

        #ctrl => parent
        #  cmd => button inside ctrl
        #  def => assets for cmd (and, if promoted, ctrl buttons)
        def aft_make_ui(ui,onCommandCreated,aftid,mywhere,parentwhere,pctrl):                            # Add a command onto the Create menu in the Sketch workspace (toolbar panel)
            global uiel, handlers, userdata
            myui=uiel[aftid]    # this is a reference
            (myctrl,mycmd,mydef,mycb,myid)=( myui['ctrl'], myui['cmd'], myui['def'], myui['cb'], mywhere + '_' + myui['id'] ) # indexed by mywhere, so same commands can go on many menus
            newcmd=False # Is this a menu item that needs a new command handler for itself?
            web_nbreq(None) # process any non-blocking operations
            if not onCommandCreated:
                newcmd=True

            if (myui['type'] == 'Command') or (myui['type'] == 'DropDown'):  # SolidCreatePanel / SketchCreatePanel
                newcmd=True

            if newcmd:
                onCommandCreated = CommandCreatedEventHandlerPanel({'aftid':aftid,'mywhere':mywhere}) # Instantiate the handler new object - we use just one for everything
                handlers.append(onCommandCreated)


            if myui['type'] == 'Button':                                                # Add a button to their tools menu
                mycmd[mywhere] = ui.workspaces.itemById('FusionSolidEnvironment').toolbarPanels.add(myui['id'],  myui["name"], mywhere, False) # SelectPanel

                # Recursive call to aft_make_ui to build all the commands we want to add on to our button
                done_icon=False
                for menu_item in uiel_order:
                    if menu_item:                                                       # Skip not-yet-defined menu stuff
                        subui=uiel[menu_item]
                        if (subui['type'] == 'Command') or (subui['type'] == 'Separator'):
                            aft_make_ui(ui,onCommandCreated,menu_item,myui['id'],mywhere,None)           # Add a command onto the Create menu in the Soild workspace
                            if not done_icon:
                                cmdControl :adsk.core.CommandControl = subui['cmd'][myui['id']] # Put an icon onto the button
                                cmdControl.isPromotedByDefault = True
                                done_icon=True

            elif (myui['type'] == 'Command') or (myui['type'] == 'Separator') or (myui['type'] == 'DropDown'):  # SolidCreatePanel / SketchCreatePanel
                if not pctrl:                                                           # If not, it's a fusion360 control, so go find the object
                    myctrl[mywhere]=ui.workspaces.itemById('FusionSolidEnvironment').toolbarPanels.itemById(mywhere)
                else:
                    if not myctrl.get('mywhere',0):
                        myctrl[mywhere]=pctrl                                           # The control is our parent if we have one
                        newcmd=True

                if not myctrl[mywhere]:
                    myctrl[mywhere]=ui.workspaces.itemById('FusionSolidEnvironment').toolbarPanels.itemById(mywhere).controls.itemById(myui['id'])

                if myctrl[mywhere].controls.itemById(myui['id']):
                    ui.messageBox('Sorry!  We detected an unclean-shutdown from ' + program_name + '.  You may need to save your work, then quit and restart Fusion 360 to correct.\n\n'+
                                  'If this message appears again, disable "Run on Startup" for ' + program_name + ' from your Tools => Add-ins menu')
                    # We could in theory search all menus and panels for lal our IDs here if we want... but this should never happen anyway.

                if(myui['type'] == 'Separator'):                                        # Add a button to their tools menu
                    mycmd[mywhere]=myctrl[mywhere].controls.addSeparator(myui['id'])    # (id, positionID, isBefore)
                    myui['name']='-----------'                                          # So the debug messagebox works

                elif(myui['type'] == 'DropDown'):                                       # Add a right-sliding list of commands to their dropdown
                    mycmd[mywhere] = myctrl[mywhere].controls.addDropDown(myui['name'], myui['res'], myui['id'], insertIndex(aftid,mywhere), False) # False means after

                    # Recursive call to aft_make_ui to build all the commands we want to add on to our button
                    for menu_item in uiel_order:
                        if menu_item:                                                   # Skip not-yet-defined menu stuff
                            subui=uiel[menu_item]
                            if (subui['type'] == 'Command') or (subui['type'] == 'Separator'):
                                aft_make_ui(ui,onCommandCreated,menu_item,myui['id'],mywhere,mycmd[mywhere])     # Add a command onto the Create menu in the Soild workspace
                    mycmd[mywhere].isVisible = True

                else:   # Command
                    mydef[mywhere]=ui.commandDefinitions.itemById(myid)                 # Already done'id'
                    if not mydef[mywhere]:
                        altid={} # use alternate versions of this command's assets for this specific sub menu
                        for tryid in ('name','desc','html','res','abs_res'):            # Things that can look different when this command appears in different sub menus
                            for myparent in ('.'+mywhere, '.'+parentwhere, ''):
                                if myui.get(tryid+myparent,0):                          # Different name for items in this specific parent
                                    altid[tryid]=myui[tryid+myparent]
                                    break
                        mydef[mywhere] = ui.commandDefinitions.addButtonDefinition(myid, altid['name'], altid['desc'], altid['res'])
                        mydef[mywhere].toolClipFilename = make_html(myui['id'],altid,mywhere)   # Clips cannot be relative-positioned files        
                        if myui.get('tip',0):
                            mydef[mywhere].tooltip = myui['tip']
                        #mycb[mywhere]=CommandCreatedEventHandlerPanel(ui,aftid,mywhere,myparent,altid.copy()) # Instantiate the handler new object - and in myui(uiel).cb thus keeps the handler referenced beyond this function
                        mydef[mywhere].commandCreated.add(onCommandCreated)              # Link our handler to the new definition
                    else:
                        dome=False

                    mycmd[mywhere]=myctrl[mywhere].controls.addCommand(mydef[mywhere])  # Add the command referencing this definition to the menu
                    mycmd[mywhere].isVisible = True
            else:
                if userdata["debug"]:
                    ui.messageBox(program_name + ' unknown type:' + myui['type'])

            return myui # Give caller everything so they can use what they like




        #################################################################################

        class MyCommandValidateInputsHandler(adsk.core.ValidateInputsEventHandler): # Only called if all needed points are selected
            def __init__(self, pnlid):
                global userdata
                super().__init__() # need to store our dialogue id in here...
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
                #if userdata["debug"]:
                    #eprint("init MyCommandValidateInputsHandler()\n")
                    #eprint("init InputChangedHandler("+pnlid+")\n")
            def notify(self, args):
                # areInputsValid    Used during the AreInputsValid event to get or set if all inputs are valid and the OK button should be enabled.
                # firingEvent       The event that the firing is in response to.
                # inputsu           Returns the collection of command inputs that are associated with the command this event is being fired for.
                #    adsk.core.ValidateInputsEventArgs.cast(args)  .firingEvent.sender.commandInputs
                try:
                    eventArgs = adsk.core.ValidateInputsEventArgs.cast(args)
                    inputs = eventArgs.firingEvent.sender.commandInputs        

                    ret=MyValidate(self.pnlid,eventArgs,inputs,'validate') # Same code does Validate and Change events
                    args.areInputsValid=ret
                    web_nbreq(None) # proccess any non-blocking operations
                    return ret
                except:
                    eprint('MyCommandValidateInputsHandler failed {}'.format(traceback.format_exc()))




        #################################################################################

        class InputChangedHandler(adsk.core.InputChangedEventHandler):
            def __init__(self, pnlid):
                super().__init__()
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
            def notify(self, args):
                global pnl,debug
                #command = args.firingEvent.sender
                try:
                    eventArgs = adsk.core.InputChangedEventArgs.cast(args)
                    inputs=eventArgs.inputs

                    #ret=MyValidate(self.pnlid,None,None,self,args) # Same code does Validate and Change events
                    ret=MyValidate(self.pnlid,eventArgs,inputs,'changed') # Same code does Validate and Change and Command events
                    args.areInputsValid=ret
                    web_nbreq(None) # proccess any non-blocking operations
                    return ret
                except:
                    eprint('InputChangedHandler failed: {}'.format(traceback.format_exc()))


        #################################################################################

        class CommandExecuteHandler(adsk.core.CommandEventHandler):
            def __init__(self, pnlid):
                super().__init__()
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
            def notify(self, args):
                try:
                    eventArgs = adsk.core.CommandEventArgs.cast(args)
                    inputs = eventArgs.command.commandInputs        
                    eprint("exec:"+self.pnlid.get('aftid'))

                    ret=MyValidate(self.pnlid,eventArgs,inputs,'execute') # Same code does Validate and Change and Execute events

                    if ret: # A valid foil
                        if self.pnlid.get('aftid')=='create_turbine'+aft or \
                           self.pnlid.get('aftid')=='create_propeller'+aft:
                            insertTurbineUI(self.pnlid,args,inputs)
                        elif self.pnlid.get('aftid')=='import_foil'+aft or \
                             self.pnlid.get('aftid')=='export_foil'+aft:
                            importExport(self.pnlid,args,inputs)    # Returns True if something failed
                        else:
                            insert2DWing(self.pnlid,args,inputs)

                    web_nbreq(None) # proccess any non-blocking operations
                except:
                    sharejson={}
                    return err(sharejson,program_name+' cmd exec failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox


        #################################################################################

        class ExecutePreviewHandler(adsk.core.CommandEventHandler):
            def __init__(self, pnlid):
                super().__init__()
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
            def notify(self, args):
                try:
                    eventArgs = adsk.core.CommandEventArgs.cast(args)   # args.firingEvent.sender
                    inputs = eventArgs.command.commandInputs            # 

                    ret=MyValidate(self.pnlid,eventArgs,inputs,'preview') # Same code does Validate and Change and Execute and preview events

                    if ret: # A valid foil
                        pass
                    eprint('ExecutePreviewHandler called')

                    web_nbreq(None) # proccess any non-blocking operations
                except:
                    eprint('ExecutePreviewHandler failed: {}'.format(traceback.format_exc()))

        #################################################################################

        class MyMarkingMenuHandler(adsk.core.MarkingMenuEventHandler):
            def __init__(self, pnlid):
                super().__init__()
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
            def notify(self, args):
                try:                    
                    eprint('MarkingMenuEventHandler (right click) called')
                    #setLinearMarkingMenu(args)
                    #setRadialMarkingMenu(args)
                    
                    # selected entities
                    #global entities
                    #entities.clear()
                    #entities = args.selectedEntities
                except:
                    eprint('MarkingMenuEventHandler (right click) error')
                    
        #################################################################################

        class CommandCreatedEventHandlerPanel(adsk.core.CommandCreatedEventHandler):
            def __init__(self, pnlid):
                super().__init__() 
                self.pnlid=pnlid        # So we know which menu-item/button triggered us
            def notify(self, args):
                global userdata, updatedata, pnl
                try:
                    app = adsk.core.Application.get()
                    ui = app.userInterface
                    eventArgs = adsk.core.CommandCreatedEventArgs.cast(args)
                    cmd = eventArgs.command
                    #inputs = cmd.commandInputs
                    #cmd = args.command
                    #cmd = adsk.core.Command.cast(args.command)
                    nospacehelpfn='/tmp/help.html'
                    if os.path.isfile(nospacehelpfn):
                        cmd.helpFile = nospacehelpfn                            # Works! - everything else with spaces in the path always fails.


                    onExecute = CommandExecuteHandler(self.pnlid)
                    cmd.execute.add(onExecute)
                    handlers.append(onExecute)

                    onInputChanged = InputChangedHandler(self.pnlid)
                    cmd.inputChanged.add(onInputChanged)
                    handlers.append(onInputChanged)                             # keep the handler referenced beyond this function

                    onValidate = MyCommandValidateInputsHandler(self.pnlid) # (self.ui,self.pnlid)
                    cmd.validateInputs.add(onValidate)
                    handlers.append(onValidate)                                 # keep the handler referenced beyond this function

                    onPreview = ExecutePreviewHandler(self.pnlid)
                    cmd.executePreview.add(onPreview)
                    handlers.append(onPreview)

                    # rats - does nothing...
                    onMarking = MyMarkingMenuHandler(self.pnlid)                # Right-Click menu
                    ui.markingMenuDisplaying.add(onMarking)
                    handlers.append(onMarking)

                    #onDestroy = MyCommandDestroyHandler()                      # We are an add-in (we stick around) - see the stop() which always gets called when we re unloaded though

                    updatedata=load_user_settings(update_file)                  # Check results from our earlier update check now, so our panel knows if it needs to tell the user about updates

                    p=aft_make_dlg(cmd,self.pnlid) 
                    pnl['opened']=int((datetime.datetime.now() - datetime.datetime(1970, 1, 1)).total_seconds())
                    pnl['openc']=pnl['openc']+1
                    if updates(True): # check only
                        pnl[p]['update_now'+aft].isVisible=True

                    web_nbreq(None) # proccess any non-blocking operations
                except:
                    sharejson={}
                    return err(sharejson,program_name+' panel create failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox


        #################################################################################

        if userdata.get("firstrun",1):
            if userdata.get("release",1):
                ui.messageBox(program_name + ' is added in these 2 places:\n\n' +
                                            '*  The Create menu dropdown in your Sketch\n'+
                                            '     workspace\n' +
                                            '*  A button in your "Tools" tool bar panel\n\n' +
                                            '             This is a one-time message.\n' +
                                            '             We hope you love our add-in.','Welcome to '+program_name)
            else:
                ui.messageBox(program_name + ' is added in these 3 places:\n\n' +
                                            '*  The Create menu dropdown in your Sketch\n'+
                                            '     workspace\n' +
                                            '*  The Create menu dropdown in your Solid Design\n' +
                                            '     workspace\n' +
                                            '*  Your "Tools" tool bar panel\n\n' +
                                            '             This is a one-time message.\n' +
                                            '             We hope you love our add-in.','Welcome to '+program_name)
            userdata["firstrun"]=False
            userdata["debug"]=False
            userdata["unsaved"]=True
            userdata["release"]=True
            save_user_settings(ui,settings_file)                                # Bug them about how to run/use/find this just one time

        aft_make_ui(ui,None,'sketch_airfoil_command'+aft,'SketchCreatePanel',None,None)   # Add a command onto the Create menu in the Sketch workspace
        if userdata.get("debug",0): aft_make_ui(ui,None,'solid_airfoil_command'+aft,'SolidCreatePanel',None,None)     # Add a command onto the Create menu in the Soild workspace
        aft_make_ui(ui,None,'airfoil_toolbar_button'+aft,'SelectPanel',None,None)         # Add a button to their Tools menu - if this is last, it works.

        if userdata.get("debug",0):
            debug=True
            ui.messageBox(program_name + ' started.\n\nIf you see this message in error, edit the debug flag to false in this file:-\n\n' + home_file(program_name,settings_file))

        # link in to check if they change anyting that affects a foil we inserted
        pnl['onCommand'] = CommandHandler('ChangeParameterCommand')
        ui.commandTerminated.add(pnl['onCommand']) # Also services any non-blocking sockets.

    except:
        if ui:
            ui.messageBox(program_name+' AddIn start failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc()))
        eprint('AddIn Start Failed: {}'.format(traceback.format_exc())) # ?vlog?



#################################################################################
####                                                                          ###
##                                  The End                                    ##
####                                                                          ###
#################################################################################

def stop(context):                                                              # Runs when we get unloaded
    global userdata, pnl
    ui = None
    try:
        app = adsk.core.Application.get()
        ui = app.userInterface

#        if textpalette:                                                         # Put it back how it was, so the View=>TextCommands master window which says "Show" or "Hide" matches the state OK)
#            try:
#                textpalette.isVisible = textpalette_ori_state
#            except:
#                pass

        web_nbreq(None) # proccess any non-blocking operations

        if userdata.get('unsaved',1):
            save_user_settings(ui,settings_file)                                # Bug them about how to run/use/find this just one time

        # Go through all our handlers and run the deleteMe() on them now...
        for aftid in uiel:                                                      # Every menu item
            myui=uiel[aftid]
            if myui.get('cmd',0): 
                for mywhere in myui['cmd']:                                         # Every place it's used in a command
                    if myui['cmd'][mywhere]:
                        try:
                            myui['cmd'][mywhere].deleteMe()
                        except:
                            pass
                    myui['cmd'][mywhere]=None
            if myui.get('def',0): 
                for mywhere in myui['def']:                                         # Every definition too
                    if myui['def'][mywhere]:
                        try:
                            myui['def'][mywhere].deleteMe()
                        except:
                            pass
                    myui['def'][mywhere]=None

        if pnl.get('onCommand',0): 
            ui.commandTerminated.remove(pnl['onCommand'])
            pnl.clear()

        if debug:
            ui.messageBox(program_name + ' ended.\n\nIf you see this message in error, edit the debug flag to false in this file:-\n\n' + home_file(program_name,settings_file))

    except:
        sharejson={}
        return err(sharejson,program_name+' stop cmd failed: please send us a screenshot of this:\n{}'.format(traceback.format_exc())) # pops up a messageBox

    # stop



#################################################################################
####                                                                          ###
##                             Helper functions                                ##
####                                                                          ###
#################################################################################

def eprint(*args, **kwargs):                                                    # Print debug stuff to STDOUT
    global debug
    if debug:
        print(*args, file=sys.stderr, **kwargs)

def nextId():                                                                   # Add sequential order to my UI command dictionary
    global panelorder
    panelorder=panelorder+1
    return panelorder

# Find a cross-platform friendly and appropriate place to save user settings files (see also the main program which does this same thing)
def home_file(program_name,settings_filename):
    if sys.platform.startswith('linux'):
        home_folder='~/.local/share'                                            # ~/.local/share/<AppName>
    elif sys.platform.startswith('darwin'):
        home_folder='~/Library/Application Support/'                            # ~/Library/Application Support/<AppName>
    else:
        home_folder=os.path.join('~','AppData','Roaming')                       # C:\Users\cnd\AppData\Roaming

    home_folder=os.path.expanduser(os.path.join(home_folder,program_name.replace(" ", "_")))
    os.makedirs(home_folder, exist_ok=True)
    return os.path.join(home_folder,settings_filename)

def load_user_settings(settings_file_only):                                     # Read user settings, like firstrun flag, units, history, helpers, etc
    settings_file=home_file(program_name,settings_file_only)                    # Put this in the correct user home folder
    try:
        with open(settings_file,'r', newline='\n', encoding='utf-8' ) as json_file:
            udata = json.load(json_file)
        udata['loaded']=True
    except:
        udata={"firstrun":True}
    return udata

def default_user_settings(userdata):                                            # Populate stuff that might be missing
    app=adsk.core.Application.get()
    if userdata.get('announcements'+aft,True):
        if not userdata.get('email'+aft,False):
            try:
                userdata['share_errors'+aft]=True                               # Default to send us crash reports
                userdata['announcements'+aft]=True                              # Default to recieve product announcement emails
                userdata['email'+aft]=app.currentUser.userName
                userdata['displayName'+aft]=app.currentUser.displayName
                sharejson=userdata
                sharejson['First_Run']=True
                ret=shareIt(sharejson,Text='send_email_prefernce_update')
                eprint("shared...") # DEBUG
                eprint(ret) # DEBUG
            except: pass
    return userdata

def save_other_settings(settings_file_only,tosave):                             # Save the settings (update file usually)
    settings_file=home_file(program_name,settings_file_only)                    # Put this in the correct user home folder
    try:
        with open(settings_file,'w', newline='\n', encoding='utf-8') as json_file:
            json_file.write(json_dumps(tosave)+"\n")
            json_file.close()
    except:
        pass

def save_user_settings(ui,settings_file_only):                                  # Save the settings that we want to remember between sessions etc.
    global userdata
    settings_file=home_file(program_name,settings_file_only)                    # Put this in the correct user home folder
    try:
        with open(settings_file,'w', newline='\n', encoding='utf-8') as json_file:
            userdata["unsaved"]=False
            json_file.write(json_dumps(userdata)+"\n")
            json_file.close()
    except:
        userdata["unsaved"]=True;sharejson={}
        return err(sharejson,program_name + ' Problem writing user settings save file: {}\n{}'.format(settings_file,traceback.format_exc())) # pops up a messageBox


def foiltool(ui,settings_file,mywork):                                          # Call our core processor to perform specified work
    cmdname=prog_folder+'/airperl.pl'
    forperl = "For Perl! path='" + os.path.dirname(os.path.realpath(__file__)) + "'\n"

    try:
        r=subprocess.run([foiltool, '-l', '/dev/null'],
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         check=True,
                         universal_newlines=True,
                         input=forperl)
        ui.messageBox(r.stdout)

    except Exception as e:
        ui.messageBox(e)


def make_html(myid,altid,panel):
    global clipa, clipb
    if altid.get('html.'+panel,0):  html=altid['html.'+panel]                   # toolClip help text exists: put it into a file for Fusion360 to use right now:
    elif altid.get('html',0):       html=altid['html']
    else:                           html=''
    if html:                                                                    # toolClip help text exists: put it into a file for Fusion360 to use right now:
        clipfn=os.path.join(altid['abs_res'], myid + '.html')
        try:
            f=open(clipfn,'w')
            f.write(clipa+html+clipb)                                           # Add required template HTML around our toolClip help text.
            f.close()
            return clipfn
        except:
            eprint("Error writing toolClip output html file '"+clipfn+"':\n\n{}".format(traceback.format_exc()))
    return ''


def insertIndex(aftid,mywhere):                                                 # Return the string to place our command in the right place, if we asked for it
    myui=uiel[aftid] # this is a reference
    if myui.get('after',0): # If we have defined a lits of things, one of which we want to appear after...
        for pnlid in myui['after']: # Search for one
            insertAfter = myui['ctrl'][mywhere].controls.itemById(pnlid) # Does this exist?
            if insertAfter: # yes
                return pnlid # insert after this then
    return ''


def distance_between_two_points(frompt,topt):                                   # ... in 3D
    if frompt and topt:
        return (   ( frompt.point.x - topt.point.x )**2 + \
                   ( frompt.point.y - topt.point.y )**2 + \
                   ( frompt.point.z - topt.point.z )**2 \
                            ) ** 0.5
    return 0.0

def vlog(*args, **kwargs):                                                      # Print runtime stuff to the Fusion360 TextCommands palette area
    global textpalette, userdata
    if userdata.get('text_info_aft',0) != 'Off (None)':                         # Print nothing if they disabled this setting
        if debug:
            print(*args, file=sys.stderr, **kwargs)
        if textpalette:
            pstr = io.StringIO()
            print(*args, file=pstr, end = '', **kwargs)
            rmsg=pstr.getvalue()
            if(green): rmsg=rmsg.replace(green,"")  # remove ANSI
            if(norm): rmsg=rmsg.replace(norm,"")    # remove ANSI
            #textpalette.writeText(pstr.getvalue()) 
            textpalette.writeText(rmsg) 
            pstr.close()


def C_to_Rankine(C):                                                            # temperature converter; Rankine is the imperial Farenheight version of K
    #0^C * 9/5 + 491.67 = 491.67^R
    return C * 9/5 + 491.67

def Rankine_to_C(R):                                                            # temperature converter; Rankine is the imperial Farenheight version of K
    return (R-491.67) * 5/9

def Gas_Viscosity(gas,tempC,dlg):                                               # Return the dynamic viscosity in centipoise for a gas at the given temperature
    #https://www.lmnoeng.com/Flow/GasViscosity.php
    #Sutherland's constant C  /  To (oR)  /  ?o (centipoise)
    constant= { 'Air':{'C':120, 'T0':524.07, 'U0':0.01827}, #   standard air
                'NH3':{'C':370, 'T0':527.67, 'U0':0.00982}, #   ammonia, NH3
                'CO2':{'C':240, 'T0':527.67, 'U0':0.01480}, #   carbon dioxide, CO2
                'CO': {'C':118, 'T0':518.67, 'U0':0.01720}, #   carbon monoxide, CO
                'H2': {'C':72,  'T0':528.93, 'U0':0.00876}, #   hydrogen, H2
                'N2': {'C':111, 'T0':540.99, 'U0':0.01781}, #   nitrogen, N2
                'O2': {'C':127, 'T0':526.05, 'U0':0.02018}, #   oxygen, O2
                'SO2':{'C':416, 'T0':528.57, 'U0':0.01254}} #   sulfur dioxide, SO2
    # Gas viscosity is computed using Sutherland's formula (Crane, 1988):
    # ? = ?o*(a/b)*(T/To)3/2

    # a = 0.555To + C
    # b = 0.555T + C

    #   where
    # ?  = viscosity in centipoise at input temperature T�R
    # ?o = reference viscosity in centipoise at reference temperature To�R
    # T   = input temperature in degrees Rankine
    # To = reference temperature in degrees Rankine
    # C  = Sutherland's constant

    if gas=='Other':
        if dlg.get('dynamic_viscosity'+aft,0):
            vlog('Dynamic Viscosity is {}'.format(dlg['dynamic_viscosity'+aft].value))
            return float(dlg['dynamic_viscosity'+aft].value)
        else:
            gas='Air' # default to air
            vlog('Caution: using Air Viscosity')

    T=C_to_Rankine(tempC)
    a = 0.555 * constant[gas]['T0'] + constant[gas]['C'] 
    b = 0.555 * T + constant[gas]['C']
     
    u = constant[gas]['U0'] * (a/b) * ( T / constant[gas]['T0'] ) ** (3/2)
    
    return u/1000                                                               # dynamic viscosity in centipoise at input temperature T�C

def Water_Density(fluid,tempC):                                                 # Return the density in kg/m� of water at the given temp (C) 
    if tempC < 126.85:
        p =  (-9.204453627E-11 * tempC**4 +3.420742008672E-8* tempC**3 -7.08919807166417E-6* tempC**2 +4.375294545181970E-5* tempC +0.999888264405735)*1000
    else:
        p = (-6.7028E-15* tempC**6 +8.28885789E-12* tempC**5 -4.18617351813E-9* tempC**4 +1.0964248668453E-6* tempC**3 -1.58525167407245E-4* tempC**2 +0.0111966695465985* tempC +0.669696280822552)*1000

    if fluid == 'Fresh Water':
        return p
    elif fluid == 'Sea Water':
        return Water_Density('Fresh Water',tempC*1.0278+4)+27.69924894   # Scale sea based on fresh
    else:       # Any water
        return 0.5 * ( p + Water_Density('Fresh Water',tempC*1.0278+4)+27.69924894 )  # Midway between them both...


def Water_Viscosity(tempC):                                           # Return the viscosity of water in Pa.s
    T=273.15+tempC
    A=2.414E-5
    B=247.8
    C=140
    v = A * 10**(B/(T-C))
    return v


def Water_Vapour_Pressure(tempC): # Needed for cavitation math - returns kPa
    # Pressure(kPa) = 0.61121 exp(( 18.678 - ( T(celcius) / 234.5)) * (T/(257.14+T))    # Buck equation: https://en.wikipedia.org/wiki/Vapour_pressure_of_water. 
    # Testing 0C=0.6112kPa, 20C=2.3383kPa, 35=5.6268, 50=12.349, 75=38.595, 100=101.31
    # python -c 'import math,sys; tempC=20; print( 0.61121 * math.exp( (18.678-tempC/234.5) * ( tempC / (257.14+tempC) ) ) );'
    return 0.61121 * math.exp( (18.678-tempC/234.5) * ( tempC / (257.14+tempC) ) )


def Fluid_Viscosity(fluid): # Not temp dependent
    #http://www-mdp.eng.cam.ac.uk/web/library/enginfo/aerothermal_dvd_only/aero/fprops/propsoffluids/node5.html
    # The viscosity of air depends mostly on the temperature. At 15 �C, the viscosity of air is 1.81 � 10-5 kg/(m�s) , 18.1 ?Pa�s or 1.81 � 10-5 Pa�s .
    # The kinematic viscosity of air at 15 �C is 1.48 � 10-5 m2 /s or 14.8 cSt. At 25 �C, the viscosity is 18.6 ?Pa�s and the kinematic viscosity 15.7 cSt.
    # https://www.translatorscafe.com/unit-converter/en-US/calculator/altitude/#altidude-scheme-big
    pass

'''
def Density(gas,tempC,pressurePa):                                              # Return the density in kg/m� of air at the given temp (C) and pressure (Pa)
    if gas[-5:] == 'Water': # All 3 that end with "Water"
        dense0=None
        if gas == 'Fresh Water':
            dense0=1000.0
        else:
            dense0=1028.1685    # Sea at 0C from https://ittc.info/media/4048/75-02-01-03.pdf

        A=0.14395
        B=0.0112 
        C=649.727 
        D=0.05107 
        Tmin=273 
        Tmax=648

        p= A / ( B ** (1+ ( 1 - (tempC-273.15)/C )**D ) )    # http://ddbonline.ddbst.de/DIPPR105DensityCalculation/DIPPR105CalculationCGI.exe?component=Water
        return p

    elif gas != 'Air':
        # The molar mass of dry air is 28.9647 g/mol
        # The gas constant R is 8.314 J / mol�K. 
        # Rsp = 287.052 J�kg?��K?� 
        #   where
        # p is the absolute pressure in Pa,
        # T is the absolute temperature of air in K, and

        # P = p/(Rsp.T)
        P = pressurePa / 287.052 * (273.15+tempC)
        return P
    else:
        raise Exception("This is only for air at present")

    return None

'''

def Altitude_to_Temperature(h): # Alt in meters above sea level, returns tempC
    # See also https://www.grc.nasa.gov/www/k-12/airplane/jsfiles/mach.js or https://www.translatorscafe.com/unit-converter/en-US/calculator/altitude/

    # height asl: h(m)  mass density: p(kg/m^3) standard temp: T(K)   temp lapse rate: L(K/m)
    table=[{'b':0,   'h':0,     'hto':11000,  'P':101325,    'T':288.15,  'L':-0.0065   },
           {'b':1,   'h':11000, 'hto':20000,  'P':22632.06,  'T':216.65,  'L':0.0       },
           {'b':2,   'h':20000, 'hto':32000,  'P':5474.889,  'T':216.65,  'L':0.001     },
           {'b':3,   'h':32000, 'hto':47000,  'P':868.0187,  'T':228.65,  'L':0.0028    },
           {'b':4,   'h':47000, 'hto':51000,  'P':110.9063,  'T':270.65,  'L':0.0       },
           {'b':5,   'h':51000, 'hto':71000,  'P':66.93887,  'T':270.65,  'L':-0.0028   },
           {'b':6,   'h':71000, 'hto':999999, 'P':3.95642,   'T':214.65,  'L':-0.002    }]

    for layer in table: # Find the line we want
        if h >= layer['h'] and h < layer['hto']:
            break

    #vlog("Altitude_to_Temperature h={} L={} tempC={}".format(h, layer['L'], -273.15 + layer['T'] + layer['L'] * (h-layer['h']) ))
    return -273.15 + layer['T'] + layer['L'] * (h-layer['h'])


def Altitude_to_Pressure(h): # Alt in meters above sea level, returns Pa
    # See https://en.wikipedia.org/wiki/Barometric_formula
    # Tested on https://www.digitaldutch.com/atmoscalc/
    # Test; 0=101.32, 1=89.876, 7=41.105, 11=22.7, 15=12.112, 20=5.529, 25=2.549, 32=8.89m ,40=2.871, 47=1.158, 49=90.327Pa, 51=70.45, 60=21.955, 71=4.478, 75=2.387, 84=0.531, 85.5=0.408

    # height asl: h(m)  mass density: p(kg/m^3) standard temp: T(K)   temp lapse rate: L(K/m)
    table=[{'b':0,   'h':0,     'hto':11000,  'P':101325,    'T':288.15,  'L':-0.0065   },
           {'b':1,   'h':11000, 'hto':20000,  'P':22632.06,  'T':216.65,  'L':0.0       },
           {'b':2,   'h':20000, 'hto':32000,  'P':5474.889,  'T':216.65,  'L':0.001     },
           {'b':3,   'h':32000, 'hto':47000,  'P':868.0187,  'T':228.65,  'L':0.0028    },
           {'b':4,   'h':47000, 'hto':51000,  'P':110.9063,  'T':270.65,  'L':0.0       },
           {'b':5,   'h':51000, 'hto':71000,  'P':66.93887,  'T':270.65,  'L':-0.0028   },
           {'b':6,   'h':71000, 'hto':999999, 'P':3.95642,   'T':214.65,  'L':-0.002    }]

    for layer in table: # Find the line we want
        if h >= layer['h'] and h < layer['hto']:
            break

    R=8.3144598 # N�m/(mol�K) universal gas constant 
    g0=9.80665  # m/s2          gravity
    M=0.0289644 # kg/mol        molar mass of Earth's air

    if layer['L']: # Formula 1
        p = layer['P'] * ( layer['T'] / (layer['T']+layer['L']*(h-layer['h'])) )**( (g0*M)/(R*layer['L']) )
    else: # Formula 2
        p = layer['P'] * math.exp( ( -g0*M*(h-layer['h']) ) / ( R * layer['T']  ) )

    return p

def Altitude_to_Density(h): # Alt in meters about sea level, returns kg/m^3 
    # See https://en.wikipedia.org/wiki/Barometric_formula
    # Tested on https://www.digitaldutch.com/atmoscalc/

    # height asl: h(m)  mass density: p(kg/m^3) standard temp: T(K)   temp lapse rate: L(K/m)
    table=[{'b':0,   'h':0,     'hto':11000,  'p':1.2250,   'T':288.15,  'L':-0.0065   },
           {'b':1,   'h':11000, 'hto':20000,  'p':0.36391,  'T':216.65,  'L':0.0       },
           {'b':2,   'h':20000, 'hto':32000,  'p':0.08803,  'T':216.65,  'L':0.001     },
           {'b':3,   'h':32000, 'hto':47000,  'p':0.01322,  'T':228.65,  'L':0.0028    },
           {'b':4,   'h':47000, 'hto':51000,  'p':0.00143,  'T':270.65,  'L':0.0       },
           {'b':5,   'h':51000, 'hto':71000,  'p':0.00086,  'T':270.65,  'L':-0.0028   },
           {'b':6,   'h':71000, 'hto':999999, 'p':0.000064, 'T':214.65,  'L':-0.002    }]

    for layer in table: # Find the line we want
        if h >= layer['h'] and h < layer['hto']:
            break

    R=8.3144598 # N�m/(mol�K) universal gas constant 
    g0=9.80665  # m/s2          gravity
    M=0.0289644 # kg/mol        molar mass of Earth's air

    if layer['L']: # Formula 1
        p = layer['p'] * ( layer['T'] / (layer['T']+layer['L']*(h-layer['h'])) )**( 1 + (g0*M)/(R*layer['L']) )
    else: # Formula 2
        p = layer['p'] * math.exp( ( -g0*M*(h-layer['h']) ) / ( R * layer['T']  ) )

    return p


def getReV(dens,kvisc,L,s,h):    # Length in meters, s in meters per second, height in meters
    return dens * s * L / kvisc          # Reynolds number

def getRe(medium,L,s,h,tempC,dlg):    # Length in meters, s in meters per second, height in meters
    if medium[-5:] == 'Water': # All 3 that end with "Water"
        Re=Water_Density(medium,tempC) * s * L / Water_Viscosity(tempC)         # Reynolds number
    else:
        Re=Altitude_to_Density(h) * s * L / Gas_Viscosity(medium,tempC,dlg)     # Reynolds number
    return Re

def Kv(gas,L,s,h,tempC,dlg):    # Length in meters, s in meters per second, height in meters
    kv=Gas_Viscosity(gas,tempC,dlg) / Altitude_to_Density(h)                    # kinematic viscosity (in cSt?)
    return kv

# zero_aoa=math.atan((speed_loss_factor * in_speed) / section_speed ) / math.pi * 180 # Work out the needed pitch at this section for the RPM and speed (in_speed is less because of betz limit)
def ZeroAoA(in_speed,section_speed,thrustN,density,discarea):# returns degrees # What is the zero angle-of-attack, knowing the speeds and thrust info?
    eprint("ZeroAoA: in_speed={} NtoT={} section_speed={} thrustN={} density={} discarea={} zeroAoA={}".format(in_speed, NtoThrust(thrustN,density,discarea), section_speed, thrustN, density, discarea, math.atan(( in_speed + NtoThrust(thrustN,density,discarea) ) / section_speed ) / math.pi * 180 ))
    if section_speed:
        return math.atan(( in_speed + NtoThrust(thrustN,density,discarea) ) / section_speed ) / math.pi * 180 # Work out the needed pitch at this section for the RPM and speed (in_speed is less because of betz limit)  # ERROr - /0
    else:
        return 0

def NtoThrust(thrustN,density,discarea): # returns m/s^2                      # How much does the energy we're putting in change the speed of the medium?
    return ((thrustN ) / ( density * 0.5 * discarea )) ** 0.5
    #g=9.80665
    #return ((thrustKG * g) / ( density * 0.5 * discarea )) ** 0.5
    # static thrust kg = density kg/m^3 / 2 * disc area m^2 * (flowspeed m/s)^2 / gravity m/2^2
    # flow m/s = ( thrust Newtons             / ( density kg/m^3 * 0.5 * disc area m^2  )) ^ 0.5
    # flow m/s = ( thrust kg * gravity m/2^2  / ( density kg/m^3 * 0.5 * disc area m^2  )) ^ 0.5

def machno(vel,tempC): # a = ( rho . R . T ) ^ 0.5.  rho=1.4,  R = 286.9 J/(kg-K) for air, T = kelvin
    # http://flowlab.groups.et.byu.net/me412/hw/ReMmatching.pdf
    #eprint("v={} t={} soundspeed={}".format(vel,tempC , ((1.4 * 286.9 * (273.15 + tempC)) ** 0.5)))
    #vlog("machno v={} t={} soundspeed={}".format(vel,tempC , ((1.4 * 286.9 * (273.15 + tempC)) ** 0.5)))
    return vel / ((1.4 * 286.9 * (273.15 + tempC)) ** 0.5)

# Do a J-Spline, with facility to handle ending points properly as well (or do loops too)
def jspline(sl, a, b, link, pts):                                               # link=0 (join), 1 (simple clamp), 2 (tangent clamp), or 3 (loop)

    if sl<0: # they provided a target number opoints they want to get, not a subdivision level - convert provided number *into* a subdivision level first
        fact=0; tgt=int(len(pts[0]))
        while tgt < -sl: fact=fact+1; tgt=tgt*2                                 # Work out what factor to use to get AT LEAST the number of points we want
        sl=fact
        if sl<2: sl=2                                                           # minimum expansion - callers expect more always

    ret=[]                                                                      # sl usually ~ 5.  a=b=1 for b-spline, a=b=0 for 4-point subdiv, etc
    for px in pts:              
        x=px.copy()                                                             # Where the spline gets built
        k=0                                     
        while k < sl:
            k=k+1                               
            if link==1:                                                         # simple clamping   0Pn = 20Pn�1 � 0Pn�2 and 0Pn+1 = 20Pn�1 � 0Pn�3.
                x.append(x[-1]*2-x[-2])                                         # 0P�1 = 20P0 � 0P1   and   0P�2 = 20P0 � 0P2. 
                x.append(x[-2]*2-x[-4])                                         # Note that the above grew the list already, so -2,-4 was really -1,-3 formerly.
                px1=x[0]*2-x[1]                                                 # 0Pn = 20Pn�1 � 0Pn�2   and   0Pn+1 = 20Pn�1 � 0Pn�3.
                px2=x[0]*2-x[2]                 
                x.insert(0,px1)
                x.insert(0,px2)                 
            elif link==2:                                               
                px1=(9-a)/4  * x[0] + (a-3)/2 * x[1] + (1-a)/4 * x[2]
                px2=(12-a)/2 * x[0] + (a-8)   * x[1] + (6-a)/2 * x[2]   
                x.insert(0,px1)
                x.insert(0,px2)                 
                px1=(9-a)/4  * x[-1] + (a-3)/2 * x[-2] + (1-a)/4 * x[-3] 
                px2=(12-a)/2 * x[-1] + (a-8)   * x[-2] + (6-a)/2 * x[-3]  
                x.append(px1)
                x.append(px2)                    
            elif link==4:                                                       # join without dropping points ?
                x.append(x[-1]*2-x[-2])                 
                px1=x[0]*2-x[1]                         
                x.insert(0,px1)
            j = 0                                
            tx=[]
            ptx=0
            while j < len(x):                   
                if j==len(x)-1 and link != 3:       
                    break
                if j == 0 and link != 3:                                        # Anchor start of output line to the start point
                    tx.append(x[j])
        
                elif j + 1 < len(x) or link==3:  
                    if link==3: 
                        ptx = ( a * x[( j - 1 )%(len(x)+1)] + ( 8 - 2 * a ) * x[j] + a * x[( j + 1 )%(len(x)+1)] ) / 8 
                    else: 
                        ptx = ( a * x[ j - 1 ] + ( 8 - 2 * a ) * x[j] + a * x[ j + 1 ] ) / 8       
                    tx.append(ptx)                  
                if link==3 or ( j + 2 < len(x) and j > 0 ):  
                    if link==3:  
                        ptx = ( ( b - 1 ) * x[(j -1)%(len(x)+1)] + ( 9 - b ) * x[ j ] + ( 9 - b ) * x[( j + 1 )%(len(x)+1)] + ( b - 1 ) * x[( j + 2 )%(len(x)+1)] ) / 16  
                    else: 
                        ptx = ( ( b - 1 ) * x[j -1] + ( 9 - b ) * x[ j ] + ( 9 - b ) * x[ j + 1 ] + ( b - 1 ) * x[ j + 2 ] ) / 16      
                    tx.append(ptx) 
                j=j+1 
            if link==3:  
                pass # skip push
            elif link>0:
                if k<sl:
                    tx=tx[3:-2]; # was -3
                else:
                    tx=tx[3:-1]; # put the actual end point back on the last one
            else: 
                tx.append(x[-1]) 
            x=tx.copy() 
        if link==3:  
            x.append(x[0])                                                      #join end to start for drawing

        if link==1:  
            x.pop() # grr
            x[-1]=x[0] # Always is anyway, but let us be certain

        ret.append(x) # this is in a loop, happens first for X, then for Y

    if len(ret)==2:
        return undup(ret[0],ret[1])
    return ret


# Take an input foil, and move the ends apart so they're at least minthick away from each other
def enforceThickness(minthick, pts):
    (x,y)=points_to_xy(pts)
    if minthick > 0:
        midpt=int(len(x)/2)/2                                                       # half way between tail and nose (we limit the points we move to here)
        i=0
        while i < midpt and y[i] - y[-(i+1)] < minthick:
            cdist=minthick - (y[i] - y[-(i+1)]) # how much extra we needed
            y[i]+=cdist/2
            y[-(i+1)] -= cdist/2
            i=i+1
    return (x,y)



# Strip duplicated start/end points.
def undup(x,y):
    if x[0]==x[1] and y[0]==y[1] and x[-2]==x[-1] and y[-2]==y[-1]:
        return undup(x[1:-1],y[1:-1])
    return(x,y)


# Take a raggedy input spline, and output N fitpoints with S leading-edge concentration factor.  S=1.000001 means equal spacing.  S=10 is good for most of our foils when using N=60
def equispace(N, S, splinetype, pts):
    N=int(N/2)-1                                                                # The nose and both tails are fixed, so, re-compute N to mean "how many on the top and again on the bottom"
    (x,y)=points_to_xy(pts)                                                     # Caller must expand these for us first now.
    nose=int(len(x)/2)
    #tailN=4 # extra points for the tail
    tailN=int((N+1)/(S/4)) # extra points for the tail
    if tailN < 2:
        tailN=0 # skip this if too few points
    elif tailN < 4: tailN=4 # min 4, if we're going to do it

    #See equispace_formula_workingout.xlsx:
    #=(1-(($I$5^(1-(($H$5-H10)/$H$5)) -1 )/($I$5-1))*$J$5) * (H10<=$H$5) + (($G$5^(1-(H10/$F$5)) -1 )/($G$5-1)) * (H10>$H$5)
    #$I$5   - S (expansion factor for the tail)
    #$H$5   - tailN (number of compressed tail points)
    #H10    - i (from 0 to N - or from 1 to N-1 if skipping both ends, which always evaluate to 1 and 0)
    #$J$5   - 1 - What the spacing would be at point tailN-1 if we weren't compressing.  i.e. 1- ($G$5^(1-((tailN-1)/$F$5)) -1 )/($G$5-1)
    #$G$5   - S (expansion factor)
    #$F$5   - N (number of points)

    # Substituting $J$5:-
    #=(1-(($I$5^(1-(($H$5-H10)/$H$5)) -1 )/($I$5-1))*  (1- ($G$5^(1-((tailN-1)/$F$5)) -1 )/($G$5-1))  ) * (H10<=$H$5) + (($G$5^(1-(H10/$F$5)) -1 )/($G$5-1)) * (H10>$H$5)

    #=(1-((S^(1-((tailN-i)/tailN)) -1 )/(S-1))*  (1- (S^(1-((tailN-1)/N)) -1 )/(S-1))  ) * (i<=tailN) + ((S^(1-(i/N)) -1 )/(S-1)) * (i>tailN)

    #N=N-tailN

    #while True:
    #    (x,y)=jspline(2,splinetype,splinetype,1, [x, y])                                          # Keep expanding them until we have a very high resolution dataset to work from
    #    #nose=int(len(x)/2)
    #    #x[nose]=0.0
    #    #y[nose]=0.0 # Don't let it stray
    #    if len(x) >770: break                                                   # This will typically be 2x that.  NB: 330 and 770 had no discernable differences.
    #    #if len(x) >16384: break                                                 # This will typically be ~ 32769
    #nose=int(len(x)/2)

    if x[nose]!=0 or y[nose]!=0: eprint("nose {} ( {} , {} ) is whiffy - not at (0,0) lx={}".format(nose,x[nose],y[nose],len(x)))
    lentop=0                                                                    # The circumferential length of the top-surface from nose to tail
    lenbot=0                                                                    # ditto - for bottom

    # Work out the length of the top and bottom surfaces, so we can equi-space our points around them
    for i in range(0,nose):
        lentop = lentop + ((x[i]-x[i+1])**2 + (y[i]-y[i+1])**2)**0.5;           # Add up all the lengths of the tiny panels
    #eprint(lentop)
    for i in range(nose,len(x)-1):
        lenbot = lenbot + ((x[i]-x[i+1])**2 + (y[i]-y[i+1])**2)**0.5;           # Add up all the lengths of the tiny panels
    #eprint(lenbot)

    newpoints=[x[0],y[0]]                                                       #  Where we plan to put the results - start at the tail


    # Do the top
    for i in range(1,N):                                                        # Skip both the ends (skip 0 and skip N)
        # scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001) with a little bit of closerness at the trailing edge tip
        if tailN:
            xi = (1-((S**(1-((tailN-i)/tailN)) -1 )/(S-1))*  (1- (S**(1-((tailN-1)/N)) -1 )/(S-1))  ) * (i<=tailN) + ((S**(1-(i/N)) -1 )/(S-1)) * (i>tailN) # A number from 1 to 0, indicating the total proportion of the top-side length we want to have
        else:
            xi=( S**(1-(i/N)) -1 )/(S-1)                                        # scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001)

        slen=xi*lentop                                                          # how far away from the nose is the point that we want?
        j=nose                                                                  # Start at the nose
        while slen>0 and j>0:                                                   # walk along the surface, counting down to the place we want...
            #vlog("top{}: j={} slen={}".format(cnt, j, slen))
            slen = slen - ((x[j]-x[j-1])**2 + (y[j]-y[j-1])**2)**0.5
            j=j-1
        newpoints.append(x[j])  # Add the point that we got up to
        newpoints.append(y[j])
        #vlog("top{} [ {},{} ]  xi={} slen={}".format(cnt, x[j], y[j], xi, slen))
 
    newpoints.append(0) # Add the nose
    newpoints.append(0)
    #eprint(len(newpoints))

    # Do the Bottom
    for i in range(1,N):                                                        # Skip both the ends (skip 0 and skip N)
        if tailN:
            #xi= (1-((S**(  ((tailN-i)/tailN)) -1 )/(S-1))*  (1- (S**(  ((tailN-1)/N)) -1 )/(S-1))  ) * (i<=tailN) + ((S**(  (i/N)) -1 )/(S-1)) * (i>tailN) # A number from 1 to 0, indicating the total proportion of the top-side length we want to have
            #xi = (1-((S**(1-((tailN-i)/tailN)) -1 )/(S-1))*  (1- (S**(1-((tailN-1)/N)) -1 )/(S-1))  ) * (i<=tailN) + ((S**(1-(i/N)) -1 )/(S-1)) * (i>tailN) # A number from 1 to 0, indicating the total proportion of the top-side length we want to have
            xi = (1-((S**(1-((tailN-(N-i))/tailN)) -1 )/(S-1))*  (1- (S**(1-((tailN-1)/N)) -1 )/(S-1))  ) * ((N-i)<=tailN) + ((S**(1-((N-i)/N)) -1 )/(S-1)) * ((N-i)>tailN) # A number from 1 to 0, indicating the total proportion of the top-side length we want to have
        else:
            xi=( S** (i/N)  -1 )/(S-1)                                        # scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001)
            #xi=( S** ((N-i)/N)  -1 )/(S-1)                                      # NG - backwards, but combined with slen=(1-xi)*lenbot does opposite spacing
            #xi=( S**(1-(i/N)) -1 )/(S-1)                                        # NG as above scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001)
            #xi=( S** (i/N)  -1 )/(S-1)                                         # NG
            #xi=( S**(1-((N-i)/N)) -1 )/(S-1)                                   # NG - same as above  scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001)

        slen=xi*lenbot                                                          # how far away from the nose is the point that we want?
        j=nose                                                                  # Start at the nose
        while slen>0 and j<len(x)-1:                                            # walk along the surface, counting down to the place we want...
            #vlog("bot{}: j={} slen={}".format(cnb, j, slen))
            slen = slen - ((x[j]-x[j+1])**2 + (y[j]-y[j+1])**2)**0.5
            j=j+1
        newpoints.append(x[j])  # Add the point that we got up to
        newpoints.append(y[j])
        #vlog("bot{} [ {},{} ]  xi={} slen={}  tailN={}".format(cnb, x[j], y[j], xi, slen, tailN ))
 
    #vlog("cnt={} cnb={} lentop={} lenbot={}".format( cnt, cnb, lentop, lenbot )) # TEMP # lentop=1.034837823816967 lenbot=1.0348378238169669 nose=64 len=128  cnt=8 cnb=8 cnt=8 cnb=8 lentop=1.0346646784543105 lenbot=1.0346646784543103
    newpoints.append(x[-1]) # Join back to the tail at the end
    newpoints.append(y[-1])
    return points_to_xy(newpoints)


'''



# Take a raggedy input spline, and output N fitpoints with S leading-edge concentration factor.  S=1.000001 means equal spacing.  S=10 is good for most of our foils when using N=60
def equispaceOK(N, S, splinetype, pts):
    N=int(N/2)-1                                                                # The nose and both tails are fixed, so, re-compute N to mean "how many on the top and again on the bottom"
    (x,y)=points_to_xy(pts)                                                     # Caller must expand these for us first now.
    nose=int(len(x)/2)

    #while True:
    #    (x,y)=jspline(2,splinetype,splinetype,1, [x, y])                                          # Keep expanding them until we have a very high resolution dataset to work from
    #    #nose=int(len(x)/2)
    #    #x[nose]=0.0
    #    #y[nose]=0.0 # Don't let it stray
    #    if len(x) >770: break                                                   # This will typically be 2x that.  NB: 330 and 770 had no discernable differences.
    #    #if len(x) >16384: break                                                 # This will typically be ~ 32769
    #nose=int(len(x)/2)

    if x[nose]!=0 or y[nose]!=0: eprint("nose is whiffy")
    lentop=0                                                                    # The circumferential length of the top-surface from nose to tail
    lenbot=0                                                                    # ditto - for bottom

    # Work out the length of the top and bottom surfaces, so we can equi-space our points around them
    for i in range(0,nose):
        lentop = lentop + ((x[i]-x[i+1])**2 + (y[i]-y[i+1])**2)**0.5;           # Add up all the lengths of the tiny panels
    #eprint(lentop)
    for i in range(nose,len(x)-1):
        lenbot = lenbot + ((x[i]-x[i+1])**2 + (y[i]-y[i+1])**2)**0.5;           # Add up all the lengths of the tiny panels
    #eprint(lenbot)

    newpoints=[x[0],y[0]]                                                       #  Where we plan to put the results - start at the tail

    # Do the top
    for i in range(1,N):                                                        # Skip both the ends (skip 0 and skip N)
        xi=( S**(1-(i/N)) -1 )/(S-1)                                            # scrunch it - Yeilds 1..0 spaced more aggressively at the trailing edge (or equal if s=1.000001)
        slen=xi*lentop                                                          # how far away from the nose is the point that we want?
        j=nose                                                                  # Start at the nose
        while slen>0 and j>0:                                                   # walk along the surface, counting down to the place we want...
            slen = slen - ((x[j]-x[j+1])**2 + (y[j]-y[j+1])**2)**0.5
            j=j-1
        newpoints.append(x[j])  # Add the point that we got up to
        newpoints.append(y[j])
 
    newpoints.append(0) # Add the nose
    newpoints.append(0)
    #eprint(len(newpoints))

    # Do the Bottom
    for i in range(1,N):                                                        # Skip both the ends (skip 0 and skip N)
        xi=( S** (i/N) -1)/(S-1)
        slen=xi*lenbot                                                          # how far away from the nose is the point that we want?
        j=nose                                                                  # Start at the nose
        while slen>0 and j<len(x)-1:                                            # walk along the surface, counting down to the place we want...
            slen = slen - ((x[j]-x[j+1])**2 + (y[j]-y[j+1])**2)**0.5
            j=j+1
        newpoints.append(x[j])  # Add the point that we got up to
        newpoints.append(y[j])
 
    newpoints.append(x[-1]) # Join back to the tail at the end
    newpoints.append(y[-1])
    return points_to_xy(newpoints)
'''


# cross-platform 'find our python'
# C:\Users\Administrator\AppData\Local\Autodesk\webdeploy\production\8b802e7a3c3d3db523b64fd80db49d9f63efcaf3\Python\Lib\venv\scripts\nt\pythonw.exe
def get_exec():
    try:
        #for ppath in (sys.exec_prefix, sys.executable, *sys.path, sys.argv[0]):
        for ppath in (sys.exec_prefix, sys.executable, *sys.path, sys.argv[0]):
            for subpath in (['bin'],['Python'],['']):
                for pyname in ('python.exe','python'): # pythonw prevents the unwanted black box popup - but fails to run due to env stuff
                    mypython=os.path.join(ppath,*subpath,pyname)
                    #eprint(mypython)
                    if os.path.isfile( mypython ): return mypython
    except: pass

    try:
        if sys.platform.startswith('darwin'):
            return os.path.join(sys.exec_prefix,'bin','python')
        else:
            return os.path.join(sys.path[0],'Python','python')

    except: return 'python' # give up searching - hope that the OS knows how to fix it itself...

# Turn the data into a TCP packet
def web_pkt(req):
    strdat=str.encode(req)
    if len(strdat)<2000:
        return b'GET /Airfoil_Tools/share.asp?'+strdat+b' HTTP/1.1\x0d\x0aHost: foils.pro\x0d\x0aAccept-Encoding: identity\x0d\x0aUser-Agent: Python\x0d\x0aConnection: close\x0d\x0a\x0d\x0a' # was check_update.asp
    else:
        return b'POST /Airfoil_Tools/share.asp HTTP/1.1\x0d\x0aHost: foils.pro\x0d\x0aAccept-Encoding: identity\x0d\x0aUser-Agent: Python\x0d\x0aContent-Length: '+str(len(strdat)).encode("latin-1")+b'\x0d\x0aConnection: close\x0d\x0aContent-Type: application/x-www-form-urlencoded\x0d\x0a\x0d\x0a'+strdat

# Do a web operation without making the user wait for it (threads). Fails in fusion (breaks their threads) if called via _thread.start_new_thread(web_req, ('u=u1.20200506&v=1.20200503&a=_aft&p=Airfoil_Tools',))
def web_req(req):
    #Not used for updates (see nbreq) - used for cloud foil lookpus still now.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if not s:
        return 0
    s.connect(("foils.pro", 80))
    sent=s.send(web_pkt(req)) # b'GET /Airfoil_Tools/share.asp?'+str.encode(req)+b' HTTP/1.1\x0d\x0aHost: foils.pro\x0d\x0aAccept-Encoding: identity\x0d\x0aUser-Agent: Python\x0d\x0aConnection: close\x0d\x0a\x0d\x0a')
    if sent==0:
        eprint("socket send problem")
        s.close()
        return 0

    ret=''
    chunk = s.recv(1024000)
    #eprint('got chunk "{}".'.format(chunk))
    while chunk != b'':
        ret += chunk.decode()
        chunk = s.recv(1024000)
        #eprint('got chunk "{}".'.format(chunk))
    ret += chunk.decode()

    if ret == b'':
        eprint("socket receive problem")
        s.close()
        return 0

    s.close()
    return ret



# Do a web operation without making the user wait for it (nonblocking sockets). Call with parm to trigger.  call often with (None) to process
def web_nbreq(req):
    global pnl
    try: # suppress all errors - this gets run often from command handler
        if pnl.get('nbs',0) and pnl['nbs']: # already set up - do read/write when we get a chance
            #eprint("-")
            ready_to_read, ready_to_write, in_error = select.select( [pnl['nbs']], [pnl['nbs']], [pnl['nbs']], 0.1)
            if len(in_error):
                eprint("e")
                eprint("non-blocking socket problem")
                pnl['nbs'].close()
                pnl['nbs']=None
                pnl['nbd']=None
            if len(ready_to_write) and pnl.get('nbd',''):
                eprint("s")
                sent=pnl['nbs'].send(pnl['nbd'])
                pnl['nbd']=None # clear - we've sent it now
            if len(ready_to_read):
                eprint("r")
                pnl['nbr'] = pnl['nbs'].recv(1024000)
                if pnl['nbr'] == b'':
                    eprint("socket receive problem")
                else:
                    eprint('got {}.'.format(pnl['nbr']))
                pnl['nbs'].close()
                pnl['nbs']=None
            return pnl['nbs']
        elif req: # initial call - set it up
            #eprint("!")
            # stores socket 'nbs' and send data 'nbd' and response data 'nbr' in the global pnl[] structure
            pnl['nbs'] = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            if not pnl['nbs']:
                return 0
            #eprint("1")
            pnl['nbs'].setblocking(0)
            #eprint("2")
            pnl['nbs'].connect_ex(("foils.pro", 80))
            #eprint("3")
            pnl['nbd']=web_pkt(req) # 'GET /Airfoil_Tools/share.asp?'+str.encode(req)+b' HTTP/1.1\x0d\x0aHost: foils.pro\x0d\x0aAccept-Encoding: identity\x0d\x0aUser-Agent: Python\x0d\x0aConnection: close\x0d\x0a\x0d\x0a'
            return web_nbreq(None) # try immediately
        # Future; how to do errors:
        #>>> err = sock.connect_ex(('10.0.0.1', 12345))
        #>>> import errno
        #>>> print errno.errorcode[err]
        #EINPROGRESS
        #>>> print sock.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        #61
        #>>> print errno.errorcode[61]
        #ECONNREFUSED
    except:
        eprint("web_nbreq problem:\n{}\n\nreq:'{}'".format(traceback.format_exc()),req)
        pass



# Get any updates that are awaiting - see also check_updates.py, the other half of this code.  Does not install updates (user gets asked first)
# Returns True if this is check_only and updates await.  Returns True if not check_only and all updates are applied
def updates(check_only):
    global updatedata
    latest_ver=updatedata
    uready=not check_only                                                       # starts with False for check_only, True for doing update (so doing-update returns "True" when done)
    ugotall=True
    nfiles=0

    #os.makedirs(home_folder, exist_ok=True)
    if latest_ver.get('updates',0):
        for updatefn in latest_ver['updates']:
            updfn=home_file(program_name, os.path.join('updates', updatefn))   # Put update files temporarily in the users data folder - the main program will apply the updates later if the user approves
            # check that the existing files, if any, are not corrupted, and remove if yes
            if os.path.isfile( updfn ):                     # Already fetched
                try:
                    if os.stat(updfn).st_size == latest_ver['updates'][updatefn]['len']:
                        uready=True
                        if check_only:
                            nfiles=nfiles+1
                            #vlog("{} Update file {} awaits".format(program_name,updfn))
                        else:
                            fn=os.path.join(prog_folder,updatefn)               # Work out where this needs to go
                            path=re.split("/|\\\\",fn)                          # Split path from filename, separator-agnostic
                            if path:
                                if len(path[0]) and path[0][-1]==':':
                                   path[0]=path[0]+os.sep                       # Python defaults to relative, but nothing else ever does...
                                fn=os.path.join(*path)                          # convert incoming / to outgoing \
                                path.pop()                                      # discard the file name 
                            vlog("{} Updating {} from {}".format(program_name,fn,updfn))
                            try:
                                if path: vlog('making path:{}.'.format(os.path.join(*path)))
                                if path: os.makedirs( os.path.join(*path), exist_ok=True)# Ensure path exists
                                with open( updfn, 'rb') as in_file:             # Open the temp file that was stored in the user home folder
                                    with open( fn, 'wb') as out_file:           # Prep to write to the target file
                                        out_file.write(in_file.read())          # Copy the file contents over
                                        out_file.close()
                                    in_file.close()
                                os.unlink(updfn)                                # Get rid of the updated file now
                            except:
                                vlog('{} Failed to update {} from {} - reason: {}'.format(program_name, fn, updfn, traceback.format_exc()))
                                ugotall=False                               # So caller knows we failed

                    else: # Wrong size - might still be downloading.  check_update.py takes care of removing corrupted files
                        uready=False
                        ugotall=False
                        vlog("{} Update file {} not fully downloaded yet".format(program_name,updfn))

                except:
                    vlog('{} update failed - please let us know about this problem: {}'.format(program_name, traceback.format_exc()))
            else: # not fetched, or, already done
                if check_only:
                    uready=False
                    ugotall=False
                    vlog("{} Update file {} not downloaded".format(program_name,updfn))

    if check_only and nfiles:
        vlog("{} of {} {} update file(s) await".format(nfiles,len(latest_ver['updates']),program_name))

    return uready and ugotall


#if __name__ == '__main__':
main_aft() # run this last, so everything is defined before use


'''

#########################
##### Design Plans ######
#########################

promo idea - world record for longest hand-thrown flight (cowl+wings)

store run history in preferences, including user metrics.

load/save into the project?
json_value = json_dumps(catch_me, sort_keys=True)
adsk.fusion.Design.cast(adsk.core.Application.get().activeProduct).attributes.add(self.group_name, 'settings', json_value)
json_value = json.loads( self._attributes_collection.itemByName(self.group_name, 'settings').value)
die if design.unitsManager.defaultLengthUnits != 'cm' # TODO


#    commandDefinitionPanel = ui.commandDefinitions.addButtonDefinition(aft+aftid, uiel[aftid]["name"], uiel[aftid]["desc"], resource_folder)
deleteMe
name
resourceFolder          Gets or sets the directory that contains any additional files associated with this command. These are typically the image files that will be used for a button and the html files for a tool clip or helps and tips.
toolClipFilename        Gets or sets the full filename of the image file (png) used for the tool clip. The tooltip is always shown but as the user hovers over the control it will progressively display the tool clip along with the tooltip text.
tooltip
commandCreated

controlDefinition       You can use properties on the control definition to define the look and behaviour of the control.
#ida

CommandDefinitions Object
  addButtonDefinition
  addCheckBoxDefinition
  addListDefinition

StandardListType


#########################
##### FeatureCreep ######
#########################
min thickness
drawing outputs
polar perf outputs and charts
scan photo to airfoil
insert foil data into sketch (image PNGs of polar diagrams etc)
sqllite DB
RPM to V/Re
RE-Calculator can name and save results into settings file
username/password for web storage? (API to save in user account prolly better)

https://forums.autodesk.com/t5/fusion-360-api-and-scripts/add-in-announcement-hydrofoil-and-airfoil-tools-seeking-your/td-p/9453985


#########################
#####  Panel Plan  ######
#########################

About Airfoil Tools + help and videos

(solid): Create Wing or Strut
(solid): Create Cowling
(solid): Create Propellor or Turbine
(sketch): Insert " (same as above)

Import / Export airfoil data files
Airfoil data maintenance operations  [declutter] [normalize] [scale] [close] [spline/points-only] [compare] => can be sideways arrows in menus
Airfoil performance analysis
Airfoil CFD genetic optimiser
Reynolds (Re) Number Calcultor

Report Bug / Request Feature

File menu; import/export airfoil ?

Airfoil UX Design.
5 Tabs: Utilities, Wing/Strut, Cowling, Propellor/Turbine, Help/Support
Utilities: import DAT; [declutter] [normalize] [scale] [close] [spline/points-only]  (export?)
Quick-insert (medium;Air(alt+temp)/Water(sea/fresh/depth/temp)/other m2/s, chord, thick, speed (m,km,knots,mph,foot,etc) ,conditions(rough+turbulent, smooth+laminar...)
Re computer
Performance Analysis
Optimize...
Compare. Investigate (most alike)
* Air
* Water
* [From KiV] - [To KiV] (filled by above, but selectable) - Kinematic Viscosity (http://airfoiltools.com/calculator/reynoldsnumber)  [Re info box?]
* Cord [nn] Thicknes[nn]
* Shape [Sketch name] ?ground plane? - see https://www.youtube.com/watch?v=qGQgMADxoFU
* Best Lift for Lowest Drag
* Lowest drag ( no lift )
* Highest lift ( poor drag )
Speed range [low]..[high] * m/s, km/h, knots, miles, etc...
Cruise [speed] [importance %]

prop/turbine - rpm, torque, hp/kw, amps/volts/phase efficiency, etc etc



#########################
#####  Handy Stuff ######
#########################


# Handy stuff for later:
returnValue = command_var.setDialogInitialSize(width, height)   # Set panel size

command_var.cancelButtonText = "No" # Change the word "Cancel" on the cancel button to something else

command_var.helpFile = resources+"/somefile.html" - adds (i) thing (left of OK at panel bottom) that opens browser to this supposedly

Good overview of interfacE: https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-F31C76F0-8C74-4343-904C-68FDA9BB8B4C

Emulate 3-point-rectangle process:
  pops up panel
  cursor says "select first corner"...

preSelect: - so we can allow users to select points or lines
https://help.autodesk.com/view/fusion360/ENU/?guid=GUID-E3D24F04-8576-420E-B9AD-44746A6B12EE

Command.validateInputs 1Event

all 4 controls: command, drop-down, split-button, and separator.


Needs 4 x 3 = 12icons for every resource entry
    16x16.png - 16x16 pixel image showing the standard image.
    32x32.png - 32x32 pixel image showing the standard image.
    16x16@2x.png - This is a copy of the 32x32.png file that is used on a retina display for the equivalent of the 16x16 standard image.
    32x32@2x.png - 64x64 pixel image which is used on a retina display for the equivalent of the 32x32 standard image.

    All above, with -disabled : e.g. 32x32-disabled@2x.png  # for when not useable
    All above, with -dark : e.g. 32x32-disabled@2x.png      # for when actually being used - use dark blue background to appear depressed.

Preferences.defaultUnitsPreferences

When editing python in vim, use:-

:set tabstop=4
:set softtabstop=4
:set shiftwidth=4
:set expandtab

'''

'''
def big_test():
    # See big_test.py
    alldats=['foildb2020/nasasc2-0714.dat', 'foildb2020/kc135a.dat', 'old_foildb2020_spline_opt/optbs_spline_sym_naca0006.dat-s17_n17s0_nsopt-Re300-N11.dat', 'foildb2020/goe611.dat']

    sharejson={}
    cnt1=len(alldats)
    err=0
    for fn in alldats:
        cnt1=cnt1-1
        if cnt1<54669:
            filename="/Users/cnd/Airfoil/" + fn
            eprint("{} err={} f={}".format(cnt1,err,filename))
            #vlog(filename)
            usefoil={'imported':filename,'Cm':0,'s':0,'alpha':0,'xy':[]}                                  # Get it ready for loadDatFile/insertFoil to use
            if loadDatFile(filename,usefoil,sharejson,True):
                err=err+1
            adsk.doEvents()
            
    eprint("DONE_REPAIRS")
    vlog("DONE_REPAIRS")


    cnt0=len(alldats)
    err0=0
    for fn in alldats:
        filename="/Users/cnd/Airfoil/" + fn
        eprint("{} err={} err0={} f={}".format(cnt0,err,err0,filename))
        cnt0=cnt0-1
        usefoil={'imported':filename,'Cm':0,'s':1,'alpha':0,'xy':[]}                                  # Get it ready for loadDatFile/insertFoil to use
        if loadDatFile(filename,usefoil,sharejson,True):
            err0=err0+1
        adsk.doEvents()
        
    eprint("DONE_REPAIRS")
    vlog("DONE_REPAIRS")


    cnt2=len(alldats)
    err2=0
    for fn in alldats:
        filename="/Users/cnd/Airfoil/" + fn
        eprint("#2: {} err1={} err0={} err2={} f={}".format(cnt2,err,err0,err2,filename))
        cnt2=cnt2-1
        #vlog(filename)
        usefoil={'imported':filename,'Cm':0,'s':0,'alpha':0,'xy':[]}                                  # Get it ready for loadDatFile/insertFoil to use
        if loadDatFile(filename,usefoil,sharejson,False):
            err2=err2+1
        usefoil={'imported':filename,'Cm':0,'s':1,'alpha':0,'xy':[]}                                  # Get it ready for loadDatFile/insertFoil to use
        if loadDatFile(filename,usefoil,sharejson,False):
            err2=err2+1
        adsk.doEvents()
        
    eprint("OK")
    #vlog("OK")
    return

'''
